package ActiveState::DNS::Request;

use strict;
use base 'Exporter';

our @EXPORT_OK = qw(
    REQUEST_ERROR
    REQUEST_OK
    REQUEST_INPROGRESS
    REQUEST_TIMEOUT
);

our %EXPORT_TAGS = (
    all => \@EXPORT_OK,
    status => [qw( 
        REQUEST_ERROR
        REQUEST_OK
        REQUEST_INPROGRESS
        REQUEST_TIMEOUT
    )],
);

our $VERSION = '1.0';

use constant REQUEST_ERROR              => -1;
use constant REQUEST_OK                 =>  0;
use constant REQUEST_INPROGRESS         =>  1;
use constant REQUEST_TIMEOUT            =>  2;

use Socket qw(inet_aton sockaddr_in);
use ActiveState::DNS::Util qw(:types decode_dns_reply encode_dns_request);

sub new {
    my $cls = shift;
    my $self = bless {}, ref($cls) || $cls;
    my $rm = shift;

    $self->{'status'}             = REQUEST_INPROGRESS;
    $self->{'error'}              = undef;
    $self->{'try_count'}          = 0;
    $self->{'nameserver_index'}   = 0;
    $self->{'retry'}              = $rm->retry;
    $self->{'timeout'}            = $rm->timeout;
    $self->{'start_time'}         = 0;
    $self->{'end_time'}           = 0;
    $self->{'packet'}             = '';
    $self->{'reply'}              = '';
    $self->{'waiting_for_reply'}  = 0;
    $self->{'waiting_to_send'}    = 0;
    $self->{'request_type'}       = undef;
    $self->{'requested_ip'}       = undef;
    $self->{'requested_domain'}   = undef;
    $self->{'returned_ips'}       = undef;
    $self->{'returned_domains'}   = undef;
    $self->{'returned_txt'}       = undef;
    $self->{'decoded'}            = 0;
    $self->{'callbacks'}          = [];

    # NOTE: All time values come from $rm->time_now, not the
    # system clock!
 
    return $self;
}

sub add_callback {
    my $self = shift;
    my $cb   = shift;
    my $data = shift;

    push @{$self->{'callbacks'}}, [$cb, $data];
}

sub _error {
    my $self = shift;
    my $rm = shift;
    
    if (@_) {
        $self->{'status'} = REQUEST_ERROR;
        $self->{'error'} = "@_";
    } else {
        $self->{'error'} = undef;
    }
    $self->end_request($rm);
}

sub error {
    return $_[0]->{'error'};
}

sub status {
    return $_[0]->{'status'};
}

sub _mkpacket {

    my $self = shift;
    my $rm = shift;

    eval {
        $self->{'packet'} = encode_dns_request(
            $self->{'query_id'},
            $self->{'request_type'},
            $self->{'requested_domain'}
        );
    };

    $self->error($rm, $@) if $@;
}

sub set_query_id {
    $_[0]->{'query_id'} = $_[1];
}

sub query_id {
    return $_[0]->{'query_id'};
}

sub lookup {
    my $self = shift;
    my $dom = lc shift;

    $dom =~ s/\.$//;
    $self->{'request_type'}     = REQUEST_TYPE_LOOKUP;
    $self->{'requested_domain'} = $dom;
}

sub reverse_lookup {
    my $self = shift;
    my $ip = shift;

    $self->{'request_type'}     = REQUEST_TYPE_REVERSELOOKUP;
    $self->{'requested_ip'}     = $ip;
    $self->{'requested_domain'} = join('.', reverse(split(/\./, $ip)), 'in-addr.arpa');
}

sub mx_lookup {
    my $self = shift;
    my $dom = lc shift;
    $dom =~ s/\.$//;

    $self->{'request_type'}     = REQUEST_TYPE_MXLOOKUP;
    $self->{'requested_domain'} = $dom;
}

sub txt_lookup {
    my $self = shift;
    my $dom = lc shift;
    $dom =~ s/\.$//;

    $self->{'request_type'}     = REQUEST_TYPE_TXTLOOKUP;
    $self->{'requested_domain'} = $dom;
}

sub ns_lookup {
    my $self = shift;
    my $dom = lc shift;
    $dom =~ s/\.$//;

    $self->{'request_type'}     = REQUEST_TYPE_NAMESERVER;
    $self->{'requested_domain'} = $dom;
}

sub start_request {
    my $self = shift;
    my $rm = shift;

    $self->_mkpacket($rm);
    return if $self->{'error'};

    $self->{'waiting_to_send'} = 1;
    $self->animate_write($rm); # try to send right away

    return if $self->{'error'};

    $self->{'start_time'} = $rm->time_now;
}

sub end_request {
    my $self = shift;
    my $rm = shift;
    $self->{'waiting_for_reply'} = 0;
    $self->{'waiting_to_send'} = 0;
    $self->{'end_time'} = $rm->time_now;
    $self->{'callbacks'} = [];
}

sub elapsed_time {
    my $self = shift;
    return $self->{'end_time'} - $self->{'start_time'};
}

sub animate_write {
    my $self = shift;
    my $rm = shift;

    if ($self->{'waiting_to_send'}) {
    	# get the address of the nameserver, inc current nameserver
    	my $nameservers = $rm->nameservers;
    	my $hostname = $nameservers->[ $self->{'nameserver_index'} ];
    	$self->{'nameserver_index'} = 
	    ($self->{'nameserver_index'}+1) % scalar(@$nameservers);
        my $port = 53;
        if ($hostname =~ /^([^:]+):(\d+)$/) {
            ($hostname, $port) = ($1,$2);
        }
      
    	my $addr = sockaddr_in($port, inet_aton($hostname));

	if (defined(send($rm->{'socket'}, $self->{'packet'}, 0, $addr))) {
            $self->{'waiting_to_send'} = 0;
	    $self->{'waiting_for_reply'} = 1;
	}
    }
}

# read packet from the socket and put it in 'reply'
# clear waiting_for_reply
sub animate_read {
    my $self = shift;
    my $rm = shift;
    $self->{'reply'} = shift;
    $self->{'waiting_for_reply'} = 0;
    $self->{'status'} = REQUEST_OK;
    $self->_decode($rm);
    $self->end_request($rm);
}

# compute time left before next retry or TIMEOUT
sub timeleft {
    my $self = shift;
    my $rm = shift;
    my $timeout = $self->{'timeout'};
    if ($timeout && $self->{'waiting_for_reply'}) {
	my $time_since_start = $rm->time_now - $self->{'start_time'};
	my $try_number = int($time_since_start / $timeout);
	if ($try_number > ($self->{'retry'})) {
            return 0;
        }
        my $timeleft = ((($try_number+1) * $timeout) - $time_since_start);
        return $timeleft if $timeleft>0;
    }
    return 0;
}

# once every tick, after all other animates
# if waiting_for_reply, check timeout. if timeout has occured than compute new
# value for trys and if it is less than or equal to retrys send again.
# the new tries value is (time_now - time_start)/timeout. (eg, a "try" can
# timeout without every having been sent, this is a "dropped" attempt.) This
# will ensure than no more than timeout * (retry+1) seconds will elapse.
sub animate_tick {
    my $self = shift;
    my $rm = shift;

    my $timeout = $self->{'timeout'};
    if ($timeout) {
        my $time_since_start = $rm->time_now - $self->{'start_time'};
        my $try_number = int($time_since_start / $timeout);
        if ($try_number > ($self->{'retry'})) {
            # timeout
            $self->{'waiting_for_reply'} = 0;
            $self->{'waiting_to_send'} = 0;
            $self->{'status'} = REQUEST_TIMEOUT;
            $self->end_request($rm);
            return;
        } 

        if ($self->{'waiting_for_reply'}) {
            # check for timeouts and retries.
            if ($self->{'try_count'} < $try_number) {
                # retry
                $self->{'try_count'} = $try_number;
                $self->{'waiting_to_send'} = 1;
                $self->animate_write($rm);
            }
        }
    }
}

# return true if waiting to receive data
sub want_read {
    return $_[0]->{'waiting_for_reply'};
}

# return true if waiting to send data
sub want_write {
    return $_[0]->{'waiting_to_send'};
}

sub type {
    return $_[0]->{'request_type'};
}

sub requested_ip {
    return $_[0]->{'requested_ip'}
}

sub requested_domain {
    return $_[0]->{'requested_domain'}
}

sub _decode {
    my $self = shift;
    my $rm = shift;
    $self->{'decoded'} = 1;

    eval {
        @{$self}{qw(returned_ips returned_domains
                    returned_txt returned_glue)} =
            decode_dns_reply(@{$self}{qw(query_id request_type
                                         requested_domain)}, \$self->{reply});
    };
    if ($@) {
        (my $code = $@) =~ s/ at .+? line .*$//s;
        $self->_error($rm, $code);
    } else {
        for my $cb (@{$self->{'callbacks'}}) {
            $cb->[0]->($rm, $self, $cb->[1]);
        }
    }
}

sub returned_ips {
    my $self = shift;

    return [] if $self->{'error'};
    
    return $self->{'returned_ips'} if defined $self->{'returned_ips'};
    return [];
}

sub returned_domains {
    my $self = shift;

    return [] if $self->{'error'};
    return $self->{'returned_domains'} if defined $self->{'returned_domains'};
    return [];
}

sub returned_txt {
    my $self = shift;
    return [] if $self->{'error'};
    return $self->{'returned_txt'} if defined $self->{'returned_txt'};
    return [];
}

sub returned_glue {
    my $self = shift;
    return {} if $self->{'error'};
    return $self->{'returned_glue'} if defined $self->{'returned_glue'};
    return {};
}

1;

=head1 NAME

ActiveState::DNS::Request - Used by ActiveState::DNS::RequestManager

=head1 SYNOPSIS

use ActiveState::DNS::Request qw(:status);

see ActiveState::DNS::RequestManager

=head1 DESCRIPTION

This is an I<internal> API for the AntiSpam SDK.

=head2 API

=over 4

=item $req->error

Returns the last error to occur on this request.

=item $req->status

Returns the status of the request. Current status codes include:

  REQUEST_ERROR
  REQUEST_OK
  REQUEST_INPROGRESS
  REQUEST_TIMEOUT

=item $req->elapsed_time

Returns the total time taken to process this request.

=item $req->type

Returns the type of the request. See ActiveState::DNS::Util for a list of types.

=item $req->requested_ip

For reverse lookup request returns the ip of the request.

=item $req->requested_domin

For mx lookup or lookup requests returns the domain of the request.

=item $req->returned_ips

For lookup requests returns the ips returned by the dns server for the requested domain.

=item $req->returned_domains

For mx loopup and reverse lookup returns the domains returned by the dns server for the requested domain or ip.

=item $req->returned_glue

For NS lookup requests, returns the associated glue records (i.e. additional RRs).

=back

=cut

