package ActiveState::DNS::RequestManager;

use strict;
use base 'Exporter';

our @EXPORT_OK = qw();
our %EXPORT_TAGS = (
    'all' => \@EXPORT_OK,
);
our $VERSION = '1.0';

BEGIN {
    if($^O eq 'MSWin32') {
        use ActiveState::Message;
        srand(ActiveState::Message::ThreadId() * time());
    }
}


use Time::HiRes q(time);

use ActiveState::DNS::Request qw(REQUEST_INPROGRESS);
use Socket qw(PF_INET SOCK_DGRAM sockaddr_in inet_aton);
use ActiveState::DNS::Util qw(nonblock);

# since every response to a request contains the query
# we can map responses to queries based only on the data
# in the response. This means we only need one socket
# for sending requests and receiving replies.

sub _open_socket {
    my $self = shift;
    my $proto = getprotobyname('udp');
    my $iaddr = inet_aton('0.0.0.0');
    socket($self->{'socket'}, PF_INET, SOCK_DGRAM, $proto);
    if (!defined $self->{'socket'}) {
        die "Could not open socket: $!";
    }
    my $count;
    for ($count = 0 ; $count < 50 ; $count++) {
        my $port = int(rand() * 64512) + 1024;
        my $paddr = sockaddr_in($port, $iaddr);
        bind($self->{'socket'}, $paddr) && last;
    }

    nonblock($self->{'socket'});
}

sub _close_socket {
    my $self = shift;
    if (defined $self->{'socket'}) {
        close($self->{'socket'});
        $self->{'socket'} = undef;
    }
}

sub new {
    my $cls = shift;
    my $self = bless {}, ref($cls) || $cls;

    # arrayref of requests being managed
    $self->{'requests'} = {};

    # arrayref of nameservers to rotate through
    $self->{'nameservers'} = [];

    # retry times
    $self->{'retry'} = 0;
    $self->{'timeout'} = 60;

    # request signatures used to prevent doing the
    # same request twice
    $self->{'request_sigs'} = {};

    $self->{'max_simultaneous_requests'} = 64;
    $self->{'queued_requests'} = [];
    $self->{'active_requests'} = [];

    $self->_tick;

    return $self;
}

sub set_timeout {
    $_[0]->{'timeout'} = $_[1];
}

sub timeout {
    return $_[0]->{'timeout'};
}

sub set_retry {
    $_[0]->{'retry'} = $_[1];
}

sub retry {
    return $_[0]->{'retry'};
}

sub _tick {
    $_[0]->{'time_now'} = time;
}

sub time_now {
    return $_[0]->{'time_now'};
}

sub set_nameservers {
    my $self = shift;
    # make sure no undefined values are used as nameservers.
    $self->{'nameservers'} = [ map { $_ ? $_ : ()} @_ ];
}

sub nameservers {
    return $_[0]->{'nameservers'};
}

sub available {
    return scalar (@{$_[0]->{'nameservers'}});
}

sub _check_for_sig {
    my $self = shift;
    my $sig = shift;

    # are we alreadying doing a request that looks like this?
    if (exists $self->{'request_sigs'}{$sig}) {
	# yep, so alias it.
	return $self->{'request_sigs'}{$sig};
    }

    return undef;
}

sub _next_request {
    my $self = shift;
    my $sig = shift;
    my $id;

    die "No nameservers defined" unless @{$self->{nameservers}};

    my $requests = $self->{'requests'};

    do {
        $id = int(rand() * int(0xFFFF)); #id must fit in 16 bits
    } while (exists $requests->{$id});

    $self->{'request_sigs'}{$sig} = $id;
    
    my $request = ActiveState::DNS::Request->new($self);
    $requests->{$id} = $request;
    $request->set_query_id($id);
    push @{$self->{'queued_requests'}}, $request;

    return $request;
}

sub lookup {
    my $self = shift;
    my $dom = shift;

    my $sig = "lookup:$dom";
    my $id = $self->_check_for_sig($sig);
    return $id if defined $id;

    my $request = $self->_next_request($sig);

    $self->_tick();
    $request->lookup($dom);

    return $request->query_id;
}

sub reverse_lookup {
    my $self = shift;
    my $ip = shift;

    my $sig = "reverse_lookup:$ip";
    my $id = $self->_check_for_sig($sig);
    return $id if defined $id;

    my $request = $self->_next_request($sig);

    $self->_tick();
    $request->reverse_lookup($ip);

    return $request->query_id;
}

sub mx_lookup {
    my $self = shift;
    my $dom = shift;

    my $sig = "mx_lookup:$dom";
    my $id = $self->_check_for_sig($sig);
    return $id if defined $id;

    my $request = $self->_next_request($sig);

    $self->_tick();
    $request->mx_lookup($dom);

    return $request->query_id;
}

sub add_request_callback {
    my $self = shift;
    my $request_id = shift;
    my $cb = shift;
    my $data = shift;

    my $request = $self->request($request_id);
    if (defined $request) {
        $request->add_callback($cb, $data);
    }
}

sub txt_lookup {
    my $self = shift;
    my $dom = shift;

    my $sig = "txt_lookup:$dom";
    my $id = $self->_check_for_sig($sig);
    return $id if defined $id;

    my $request = $self->_next_request($sig);

    $self->_tick();
    $request->txt_lookup($dom);

    return $request->query_id;
}

sub ns_lookup {
    my $self = shift;
    my $dom = shift;

    my $sig = "ns_lookup:$dom";
    my $id = $self->_check_for_sig($sig);
    return $id if defined $id;

    my $request = $self->_next_request($sig);

    $self->_tick();
    $request->ns_lookup($dom);

    return $request->query_id;
}

sub animate {
    my $self = shift;
    my $timeout = shift || 0;
    my $max_wait = shift;

    my $requests = $self->{'requests'};

    # anything to do?
    return 0 if (! scalar(keys %$requests));

    $self->_open_socket() unless $self->{'socket'};

    my $active_requests = $self->{'active_requests'};
    my $queued_requests = $self->{'queued_requests'};

    @$active_requests =
       map { $_->status == REQUEST_INPROGRESS ? $_ : () } @$active_requests;

    if (@$queued_requests) {
        while (scalar(@$active_requests) < $self->{'max_simultaneous_requests'}) {
            my $req = shift(@$queued_requests) || last;
            $req->start_request($self);
            push @$active_requests, $req;
        }
    }

    return 0 if (! scalar(@$active_requests));

    my ($want_read, $want_write) = ('','');

    if ($max_wait) {
        $timeout = $self->{'timeout'};
        for my $req (@$active_requests) {
            my $timeleft = $req->timeleft($self);
            if ($timeleft < $timeout) {
                $timeout = $timeleft;
            }
        }
    }

    my $socket_fileno = fileno($self->{'socket'});

    for my $req (@$active_requests) {
        if ($req->want_read) {
            vec($want_read, $socket_fileno, 1)=1;
            last;
        }
    }

    for my $req(@$active_requests) {
        if ($req->want_write) {
            vec($want_write, $socket_fileno, 1)=1;
            last;
        }
    }

    select(
	$want_read,
	$want_write,
	undef,
	$timeout
    );

    $self->_tick();

    if (vec($want_write, $socket_fileno, 1) == 1) {
        for my $req (@$active_requests) {
            $req->animate_write($self);
        }
    }

    if (vec($want_read, $socket_fileno, 1) == 1) {
        my ($data, $servaddr);
        $servaddr = recv($self->{'socket'}, $data, 0xffff, 0);
        if ($servaddr) {
            my $id = unpack('n', $data);
            my $req = $self->request($id);
            $req->animate_read($self, $data) if $req;
        }
    }

    for my $req (@$active_requests) {
	if ($req->status == REQUEST_INPROGRESS) {
	    $req->animate_tick($self);
	}
    }

    return scalar(@$active_requests);
}

sub join {
    while($_[0]->animate(
            0, # timeout
            1  # ignore timeout passed in and compute the timeout based on requests remaining
        )
    ) {
        # wait for requests to finish or timeout
    }
    $_[0]->_close_socket();
}

sub request {
    my $self = shift;
    my $id = shift;

    return $self->{'requests'}{$id};
}

sub enumerate_requests {
    my $self = shift;
    my $cb = shift;
    my $data = shift;

    for my $request (values %{$self->{'requests'}}) {
        last unless $cb->($data, $self, $request);
    }
}

sub reset {
    my $self = shift;

    $self->{'requests'} = {};
    $self->{'request_sigs'} = {};
    $self->{'queued_requests'} = [];
    $self->{'active_requests'} = [];
    $self->_close_socket() if $self->{'socket'};
}

1;

=head1 NAME

ActiveState::DNS::RequestManager - Manages ActiveState::DNS::Request objects

=head1 SYNOPSIS

    use ActiveState::DNS::RequestManager;
    use ActiveState::DNS::Request qw(:status);

    my $rm = ActiveState::DNS::RequestManager->new();

    # set timeout to one second
    $rm->set_timeout(1);

    # set retry to one
    $rm->set_retry(1);

    # set dns name servers to use, port can be included
    $rm->set_nameservers('127.0.0.1', '127.0.0.1:53');

    # lookup ips for a domain
    my $lookup_id = $rm->lookup("example.com");

    # lookup domains for an ip
    my $reverse_lookup_id = $rm->reverse_lookup("127.0.0.1");

    # lookup mx domains for a domain (returned in preference order)
    my $mx_lookup_id = $rm->mx_lookup("example.com");

    # wait for the requests to finish
    $rm->join;
   
    my $lookup_request = $rm->request($lookup_id);
    if (REQUEST_OK == $lookup_request->status) {
        print $lookup_request->requested_domain , " -> " ,
           join(',', @{$lookup_request->returned_ips}) , "\n";
    }

    my $reverse_lookup_request = $rm->request($reverse_lookup_id);
    if (REQUEST_OK == $reverse_lookup_request->status) {
        print $reverse_lookup_request->requested_ip , " -> " ,
           join(',', @{$reverse_lookup_request->returned_domains}) , "\n";
    }
    
    my $mx_lookup_request = $rm->request($mx_lookup_id);
    if (REQUEST_OK == $mx_lookup_request->status) {
        print $mx_lookup_request->requested_domain , " -> " ,
            join(',', @{$mx_lookkup_request->returned_domains}) , "\n";
    }
    
=head1 DESCRIPTION

This is an I<internal> ActiveState API for the AntiSpam SDK.

The ActiveState::DNS::RequestManger manages multiple DNS requests in a IO parallelized fashion.

The general usage pattern should go as follows:

- make a bunch of requests
- go do something else, calling animate every now and then
- join
- scan requests for results

The animate call gives the manager a chance to deal with pending IO and preforms retries.

=head2 API

=over 4

=item $rm = ActiveState::DNS::RequestManager->new()

Creates a new request manager.

=item $rm->set_timeout(timeout):

Set the timeout, see $rm->timeout

=item $rm->timeout

Returns the amount of time in seconds the manager will wait for a response from the dns server.

=item $rm->set_retry(retries)

Set the retry, see $rm->retry

=item $rm->retry

Returns the number of times the manager will retry a request. The maximum amount of time the manager will spend on a request is (timeout * (retry+1)). In order for retries to occur you must call $rm->animate or $rm->join. If more than timeout seconds pass between calls to animate some retries will be skipped.

=item $rm->time_now

Returns the Time::HiRes::time of the last $rm->animate

=item $rm->set_nameservers(nameservers)

Sets the nameservers, see nameservers

=item $rm->nameserver

Returns the nameservers the manager will send requests too. Each request will pick a random starting offset into the nameserver list and then cycles around the list for retries.

=item $rm->error

Returns a string describing the last error the manager encountered, or undef if no error has been encountered.

=item $rm->available

Returns true if the manager is properly configured to handle requests.

=item $rm->lookup(domain)

Returns the id of a request that will provide the ips found for the domain. Looking up the same domain multiple times will return the same request id. If the request timed out you have to reset the entire request manager and retry the lookup.

=item $rm->reverse_lookup(ip)

The reverse of the lookup call.

=item $rm->mx_lookup(domain)

Like the lookup call except returns domains sorted by mx preference rather than ips.

=item $rm->animate

Preform pending IO operations and retries.

=item $rm->join

Wait for all retries to timeout and then return.

=item $rm->request(request_id)

Return the request object for a given request id. Used to get the results of a request.

=item $rm->enumerate_requests(callback, dataref)

Iterate over all the requests calling the callback passing in the dataref, request manager, and the request object.

=item $rm->reset

Flush all the requests from the manager in preperation for preforming more requests.

=back

=cut

