package ActiveState::DNS::Util;

use strict;
use warnings;
use base 'Exporter';

our @EXPORT_OK = qw(
    REQUEST_TYPE_LOOKUP
    REQUEST_TYPE_NAMESERVER
    REQUEST_TYPE_MAILDESTINATION
    REQUEST_TYPE_MAILFORWARDER
    REQUEST_TYPE_CNAME
    REQUEST_TYPE_SOA
    REQUEST_TYPE_MAILBOX
    REQUEST_TYPE_MAILGROUP
    REQUEST_TYPE_MAILRENAME
    REQUEST_TYPE_NULL
    REQUEST_TYPE_WKS
    REQUEST_TYPE_REVERSELOOKUP
    REQUEST_TYPE_HOSTINFO
    REQUEST_TYPE_MAILINFO
    REQUEST_TYPE_MXLOOKUP
    REQUEST_TYPE_TXTLOOKUP
    &system_nameservers
    &nonblock
    &reuseaddr
    &decode_dns_request
    &encode_dns_request
    &decode_dns_reply
    &encode_dns_reply
);

our %EXPORT_TAGS = (
    all => \@EXPORT_OK,
    types => [qw(
	REQUEST_TYPE_LOOKUP
	REQUEST_TYPE_NAMESERVER
	REQUEST_TYPE_MAILDESTINATION
	REQUEST_TYPE_MAILFORWARDER
	REQUEST_TYPE_CNAME
	REQUEST_TYPE_SOA
	REQUEST_TYPE_MAILBOX
	REQUEST_TYPE_MAILGROUP
	REQUEST_TYPE_MAILRENAME
	REQUEST_TYPE_NULL
	REQUEST_TYPE_WKS
	REQUEST_TYPE_REVERSELOOKUP
	REQUEST_TYPE_HOSTINFO
	REQUEST_TYPE_MAILINFO
	REQUEST_TYPE_MXLOOKUP
	REQUEST_TYPE_TXTLOOKUP
    )],
);

use constant REQUEST_TYPE_LOOKUP          =>  1;
use constant REQUEST_TYPE_NAMESERVER      =>  2;
use constant REQUEST_TYPE_MAILDESTINATION =>  3;
use constant REQUEST_TYPE_MAILFORWARDER   =>  4;
use constant REQUEST_TYPE_CNAME           =>  5;
use constant REQUEST_TYPE_SOA             =>  6;
use constant REQUEST_TYPE_MAILBOX         =>  7;
use constant REQUEST_TYPE_MAILGROUP       =>  8;
use constant REQUEST_TYPE_MAILRENAME      =>  9;
use constant REQUEST_TYPE_NULL            => 10;
use constant REQUEST_TYPE_WKS             => 11;
use constant REQUEST_TYPE_REVERSELOOKUP   => 12;
use constant REQUEST_TYPE_HOSTINFO        => 13;
use constant REQUEST_TYPE_MAILINFO        => 14;
use constant REQUEST_TYPE_MXLOOKUP        => 15;
use constant REQUEST_TYPE_TXTLOOKUP       => 16;

use constant RCODE_NO_ERROR        => 0;
use constant RCODE_FORMAT_ERROR    => 1;
use constant RCODE_SERVFAIL        => 2;
use constant RCODE_NAME_ERROR      => 3; # NXDOMAIN
use constant RCODE_NOT_IMPLEMENTED => 4;
use constant RCODE_REFUSED         => 5;
use constant RCODE_YXDOMAIN        => 6;
use constant RCODE_YXRRSET         => 7;
use constant RCODE_NXRRSET         => 8;
use constant RCODE_NOTAUTH         => 9;
use constant RCODE_NOTZONE         => 10;

my %rcode_map = (
    &RCODE_NO_ERROR => "RCODE_NO_ERROR",
    &RCODE_FORMAT_ERROR => "RCODE_FORMAT_ERROR",
    &RCODE_SERVFAIL => "RCODE_SERVFAIL",
    &RCODE_NAME_ERROR => "RCODE_NXDOMAIN",
    &RCODE_NOT_IMPLEMENTED => "RCODE_NOT_IMPLEMENTED",
    &RCODE_REFUSED => "RCODE_REFUSED",
    &RCODE_YXDOMAIN => "RCODE_YXDOMAIN",
    &RCODE_YXRRSET => "RCODE_YXRRSET",
    &RCODE_NXRRSET => "RCODE_NXRRSET",
    &RCODE_NOTAUTH => "RCODE_NOTAUTH",
    &RCODE_NOTZONE => "RCODE_NOTZONE",
);

if ($^O eq 'MSWin32') {
    eval 'use Win32::Registry';
} else {
    require Fcntl;
    Fcntl->import(qw(F_SETFL F_GETFL O_NONBLOCK));
}

use Socket qw(SOL_SOCKET SO_REUSEADDR);

our $VERSION = '1.0';

sub system_nameservers {
    if ($^O eq 'MSWin32') {
	return win_nameservers();
    } else {
	return unix_nameservers();
    }
}

sub unix_nameservers {
    my @servers;
    for my $filename (
	"/etc/resolv.conf",
	"$ENV{HOME}/.resolv.conf",
	"$ENV{HOME}/.resolve.conf", # typo; kept for backwards compat --NeilW
    ) {
	if (open (my $FH, $filename)) {
            my $line;
            while ($line = <$FH>) {
		$line =~ s/#.*$//m;
                if ($line =~ /\s*nameserver\s+((?:\d{1,3}\.){3}\d{1,3})/i) {
                    push @servers, $1;
                }
            }
	    close($FH);
	}
    }
    return @servers;
}

# Adapted from the 1.0 release Net-DNS code
sub win_nameservers {
	my ($resobj, %keys);

	my $root = 'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters';
	unless ($main::HKEY_LOCAL_MACHINE->Open($root, $resobj)) {
		# Didn't work, maybe we are on 95/98/Me?
		$root = 'SYSTEM\CurrentControlSet\Services\VxD\MSTCP';
		$main::HKEY_LOCAL_MACHINE->Open($root, $resobj)
			or die "can't read registry: $!";
	}

	$resobj->GetValues(\%keys)
		or die "can't read registry values: $!";

	# This is (probably) adequate on NT4
	# NameServer overrides DhcpNameServer if both exist
	my $nt4nameservers = $keys{'NameServer'}->[2] || $keys{'DhcpNameServer'}->[2];
	my $nameservers = "";

        # See Net-DNS, ticket 1193.  This change came from Hanno Stock, May 19 2006
        # This should work on Win2K and XP and looks for the DNS services
        # using the BIND key
        #

        my $bind_linkage;
        my @sorted_interfaces;
        $main::HKEY_LOCAL_MACHINE->Open('SYSTEM\CurrentControlSet\Services\Tcpip\Linkage', $bind_linkage);
        if($bind_linkage){
            my $bind_linkage_list;
            my $type;
            $bind_linkage->QueryValueEx('Bind', $type, $bind_linkage_list);
            if($bind_linkage_list){
                @sorted_interfaces = split(m/[^\w{}\\-]+/s, $bind_linkage_list);
            }
            foreach my $interface (@sorted_interfaces){
                $interface =~ s/^\\device\\//i;
            }
        }


        my $interfaces;
        $resobj->Open("Interfaces", $interfaces);
        if ($interfaces) {
            my @ifacelist;
            if(@sorted_interfaces){
                @ifacelist = @sorted_interfaces;
            }else{
                $interfaces->GetKeys(\@ifacelist);
            }
            foreach my $iface (@ifacelist) {
                my $regiface;
                $interfaces->Open($iface, $regiface);
                if ($regiface) {
                    my $ns;
                    my $type;
                    my $ip;
                    my $ipdhcp;
                    $regiface->QueryValueEx("IPAddress", $type, $ip);
                    $regiface->QueryValueEx("DhcpIPAddress", $type, $ipdhcp);
                    if (($ip && !($ip =~ /0\.0\.0\.0/)) || ($ipdhcp && !($ipdhcp =~ /0\.0\.0\.0/))) {
                        # NameServer overrides DhcpNameServer if both exist
                        $regiface->QueryValueEx("NameServer", $type, $ns);
                        $regiface->QueryValueEx("DhcpNameServer", $type, $ns) unless $ns;
                        $nameservers .= " $ns" if $ns;
                    }
                }
            }
        }
        if (!$nameservers) {
            $nameservers = $nt4nameservers;
        }

        if ($nameservers) {
            # just in case dups were introduced...
            my @a;
            my %h;
            foreach my $ns (split(/[^\d\.]+/, $nameservers)) {
                push @a, $ns unless $h{$ns};
                $h{$ns} = 1;
            }
            return @a;
        }
        return ();
}


sub nonblock {
    if ($^O eq 'MSWin32') {
        # see winsock.h for derivation of the magic number 0x8004667e (FIONBIO)
        my ($f, $v) = (0x8004667e, pack("L",1));
        ioctl($_[0], $f, $v)
            or die "nonblock failed\n";
    } else {
        fcntl($_[0], &F_SETFL, fcntl($_[0], &F_GETFL, 0) | &O_NONBLOCK)
            or die "nonblock failed\n";
    }
    return $_[0];
}

sub reuseaddr {
    setsockopt($_[0], SOL_SOCKET, SO_REUSEADDR, pack("l",1));
}


sub _decode_dns_name {
    my $reply_ref   = shift;
    my $offset_ref  = shift;
    my $name_ref    = shift;

    # track offsets to previous 'pointers' so we can avoid cycles!
    my $offset_href = shift;

    $offset_href = {} if (!defined $offset_href);

    my $n;
    my $nhb;

    for(;;) {
        $n = unpack('C', substr($$reply_ref, $$offset_ref, 1) );
        $$offset_ref += 1;

        # Have we reached the end of the name?
        if ($n == 0) {
            # Is this a valid but empty name? If so, make sure buffer was initialized.
            if (!$$name_ref) {
                $$name_ref = '';
            }

            return 1;
        }

        $nhb = $n & 0xC0;
        if ($nhb) { # either of the top two high bits set
            if ($nhb != 0xC0) {
                die "Server returned unknown encoding for a name";
            }
            $n &= 0x3F; # (~ 0xC0) & 0xFF
            $n <<= 8;
            $n += unpack ('C', substr($$reply_ref, $$offset_ref, 1) );

            if (exists $offset_href->{$$offset_ref}) {
                # oops, cycle detected, that's bad.
                die "Server returned a name containing a 'pointer' cycle";
            }

            $offset_href->{$$offset_ref} = 1;
            $$offset_ref += 1;

            _decode_dns_name($reply_ref, \$n, $name_ref, $offset_href); # recurse
            return 1; # pointers are always the end of a name.
        }

        $$name_ref .= '.' if ($$name_ref || (defined $$name_ref and $$name_ref eq '0'));
        $$name_ref .= substr($$reply_ref, $$offset_ref, $n);
        $$offset_ref += $n;
    }
}

sub _encode_dns_name {
    my $hostname = shift;
    die "Empty hostname" unless $hostname;
    $hostname =~ s/\s*//g; # remove whitespace that might have been left around the hostname
    my @parts = split(/\./, $hostname);
    my $ret = '';
    for my $part (@parts) {
        my $len = length $part;

        if ($len > 63) {
            die "Name parts cannot exceed 63 characters.";
        }
        $ret .= chr($len) . $part;
    }
    $ret .= chr(0);
    return $ret;
}

sub decode_dns_request {
    my $rref = shift;

    my ($id, $flags, $qdcount) = unpack('nnn', $$rref);

    my $offset = 12;
    my $query;
    _decode_dns_name($rref, \$offset, \$query);

    my $type = unpack('n', substr($$rref, $offset, 2));

    return ($id, $type, $query);
}

sub decode_dns_reply {
    my $request_id = shift;
    my $request_type = shift;
    my $requested_domain = shift;
    my $rref = shift;

    # decode header
    my ($id, $flags, $qdcount, $ancount, $nscount, $arcount) =
        unpack('nnnnnn', $$rref);

    my $rcode = $flags & 0xf;
    if ($rcode != RCODE_NO_ERROR) {
        die $rcode_map{$rcode} if exists $rcode_map{$rcode};
        die "RCODE_UNKNOWN_$rcode";
    }

    # check query id is right
    if ($request_id != $id) {
        die "Server returned results for a different query (expected $request_id, got $id, domain: $requested_domain)";
    }

    # we only ever send a single query
    if ($qdcount != 1) {
        die "Server returned results for $qdcount queries! We only send one request";
    }

    my $offset = 12; # sizeof(header)
    my $query;

    # decode query section, this must come right after the header
    # format is 'C' count, 'a' x n ascii's, repeat until 'C' == 0
    # followed by 'n' QTYPE and 'n' QCLASS

    _decode_dns_name($rref, \$offset, \$query);

    my $qtype = unpack('n', substr($$rref, $offset, 2) );
    $offset +=2;

    if ($request_type != $qtype) {
        die "Server returned results for a different query type (expected $request_type, got $qtype, domain: $requested_domain)";
    }

    my $qclass = unpack('n', substr($$rref, $offset, 2) );
    $offset +=2;

    if ($qclass != 1) {
        # this should _never_ happen
        die "Server seems to think we are not on the internet";
    }

    $query = lc $query; # dom's are case insensitive
    if ($requested_domain ne $query) {
        die "Server returned results for a different domain (expected $requested_domain, got $query)";
    }

    # XXX need to abstract parsing the different return types
    my $returned_ips = [];
    my $returned_domains = [];
    my $returned_txt = [];
    my $returned_glue = {};

    # decode Answer RR's
    for (my $anrr = 0; $anrr < $ancount; ++$anrr) {
        my $name;
        _decode_dns_name($rref, \$offset, \$name);
        my ($type, $class, $ttl, $rdlength) = unpack ("nnNn", substr($$rref, $offset, 10) );
        $offset += 10;

        if ($class == 1 and $type == $request_type) {
            if ($type == REQUEST_TYPE_LOOKUP and $rdlength == 4) {
                # looks that an A IN with an IPv4!
                push @$returned_ips, join('.', unpack('CCCC', substr($$rref, $offset, 4)));
            } elsif ($type == REQUEST_TYPE_REVERSELOOKUP and $request_type == REQUEST_TYPE_REVERSELOOKUP) {
                # PTR
                my $name;
                my $temp = $offset;
                _decode_dns_name($rref, \$temp, \$name);
                push @$returned_domains, $name;
            } elsif ($type == REQUEST_TYPE_MXLOOKUP) {
                # MX
                my $mx = [];
                $mx->[0] = unpack('n', substr($$rref, $offset, 2));
                my $temp = $offset + 2;
                $mx->[1] = "";
                _decode_dns_name($rref, \$temp, \$mx->[1]);
                push @$returned_domains, $mx;
            } elsif ($type == REQUEST_TYPE_TXTLOOKUP) {
                #TXT
                my $temp = substr($$rref,$offset,$rdlength);
                my $temp_offset = 0;
                while($temp_offset<$rdlength) {
                    my $len = unpack('C', substr($temp, $temp_offset, 1));
                    # if $len <= 0 assume end of text record
                    last unless $len>0;
                    $temp_offset += 1;
                    push @$returned_txt, substr($temp, $temp_offset, $len);
                    $temp_offset += $len;
                }
            } elsif ($type == REQUEST_TYPE_NAMESERVER) {
                # NS
                my $name;
                my $temp = $offset;
                _decode_dns_name($rref, \$temp, \$name);
                push @$returned_domains, $name;
            }
        }
        $offset += $rdlength;
    }

    if ($request_type == REQUEST_TYPE_NAMESERVER) {

        # Skip authority records

        for (my $nsrr = 0; $nsrr < $nscount; ++$nsrr) {
            _decode_dns_name($rref, \$offset, \(my $name));
            my (undef, undef, undef, $rdlength) = unpack "n2Nn", substr($$rref, $offset, 10);
            $offset += 10 + $rdlength;
        }

        # Decode glue records

        for (my $arrr = 0; $arrr < $arcount; ++$arrr) {
            _decode_dns_name($rref, \$offset, \(my $name));
            my ($type, $class, $ttl, $rdlength) = unpack "n2Nn", substr($$rref, $offset, 10);
            $offset += 10;
            if ($class == 1 && $type == REQUEST_TYPE_LOOKUP && $rdlength == 4) {
                $returned_glue->{$name} = join ".", unpack "C4", substr($$rref, $offset, 4);
            }
            $offset += $rdlength;
        }
    }

    if ($request_type == REQUEST_TYPE_MXLOOKUP) {
        # sort by preference and leave behind just the domain name
        @$returned_domains = map { $_ = $_->[1] } sort { $a->[0] <=> $b->[0] } @$returned_domains;
    }

    return ($returned_ips, $returned_domains, $returned_txt, $returned_glue);
}

sub encode_dns_request {
    my ($request_id, $request_type, $requested_domain) = @_;

    my $packet = pack("nnnnnn", $request_id,  0x0100, 1, 0, 0, 0);
    #   ID       (16 bits)  id
    #   QR       ( 1 bit)   0      (query)
    #   Opcode   ( 4 bits)  0      (standard query)
    #   AA       ( 1 bit)   0      (not used in query)
    #   TC       ( 1 bit)   0      (not truncated)
    #   RD       ( 1 bit)   1      (recurse)
    #   RA       ( 1 bit)   0      (not used in query)
    #   Z        ( 3 bit)   0      (reserved, must be zero)
    #   RCODE    ( 4 bits)  0      (not used in query)
    #   QDCount  (16 bits)  1      (one request)
    #   ANCount  (16 bits)  0      (no answers, we are querying)
    #   NSCount  (16 bits)  0      ( '' )
    #   ARCount  (16 bits)  0      ( '' )

    $packet .= _encode_dns_name($requested_domain);
    #   hostname (varies)   $dom   (in dns query format)

    $packet .= pack('nn', $request_type, 1);
    #   QType    (16 bits)  x
    #   Class    (16 bits)  1      (IN internet)

    return $packet;
}

sub _encode_dns_lookup {
    my $reply = shift;
    return pack('CCnnNnCCCC', 0xC0, 12, REQUEST_TYPE_LOOKUP, 1,
        2*60*60, 4, @$reply);
}

sub _encode_dns_txtlookup {
    my $reply = shift;
    my $temp = length $reply;
    return pack('CCnnNnC', 0xC0, 12, REQUEST_TYPE_TXTLOOKUP, 1,
        2*60*60, $temp+1, $temp) . $reply;
}

sub _encode_dns_mxlookup {
    my $reply = shift;
    my $mx_domain = _encode_dns_name($reply->[1]);
    return pack('CCnnNnn', 0xC0, 12, REQUEST_TYPE_MXLOOKUP, 1,
        2*60*60, 2+length($mx_domain), $reply->[0]) . $mx_domain;
}

sub _encode_dns_reverselookup {
    my $reply = shift;
    my $ptr_domain = _encode_dns_name($reply);
    return pack('CCnnNn', 0xC0, 12, REQUEST_TYPE_REVERSELOOKUP, 1,
        2*60*60, length($ptr_domain)) . $ptr_domain;
}

sub _encode_dns_nslookup {
    my $reply = shift;
    my $ns_domain = _encode_dns_name($reply);
    return pack('CCnnNn', 0xC0, 12, REQUEST_TYPE_NAMESERVER, 1,
        2*60*60, length($ns_domain)) . $ns_domain;
}

sub encode_dns_reply {
    my ($request_id, $request_type, $request_domain, $response_data, $rcode) = @_;
    $response_data = [] unless defined $response_data;

    $rcode ||= 0;

    my $packet = pack('nnnnnn', $request_id, 0x8000 | $rcode, 1, scalar(@$response_data), 0, 0);
    $packet .= _encode_dns_name($request_domain);
    $packet .= pack('nn', $request_type, 1);
    
    for my $reply (@$response_data) {
        if ($request_type eq REQUEST_TYPE_LOOKUP) {
            $packet .= _encode_dns_lookup($reply);
        } elsif ($request_type eq REQUEST_TYPE_MXLOOKUP) {
            $packet .= _encode_dns_mxlookup($reply);
        } elsif ($request_type eq REQUEST_TYPE_REVERSELOOKUP) {
            $packet .= _encode_dns_reverselookup($reply);
        } elsif ($request_type eq REQUEST_TYPE_TXTLOOKUP) {
            $packet .= _encode_dns_txtlookup($reply);
        } elsif ($request_type eq REQUEST_TYPE_NAMESERVER) {
            $packet .= _encode_dns_nslookup($reply);
        }
    }

    return $packet;
}

1;

=head1 NAME

ActiveState::DNS::Util - Functions shared by Request and RequestManager

=head1 SYNOPSIS

use ActiveState::DNS::Util qw(:types);

=head1 DESCRIPTION

This is an I<internal> API for the AntiSpam SDK.

=head2 API

=over 4

=item system_nameservers()

Makes a valiant attempt at figuring out what the systems current nameservers are
and returns an arrayref to them. Doesn't work on windows.

=item nonblock($filehandle)

Sets a filehandle to nonblocking io mode

=item reuseaddr($socket)

Sets the SO_REUSEADDR flag on a socket. This lets the socket be bound to a port already bound by another socket.

=item decode_dns_request($ref_request_data)

Takes a reference to a dns request data packet and returns the id, type, and query for the request. Currently only requests with a single query are docoded.

=item decode_dns_reply($request_id, $request_type, $request_domain, $ref_reply_data)

Takes a reference to a dns reply data packet and the information about the original query. If the reply packet matches the query information it returns two values, the first being an arrayref of ips and the second being an arrayref of domains.

=item encode_dns_request($request_id, $request_type, $request_domain)

Takes query information and returns the data packet to send to a dns server.

=item encode_dns_reply($request_id, $request_type, $request_domain, $response_data)

Takes query information and response_data for the query and returns a data packet to send back to the client that made the query. The response data varies by query type. Lookups are an arrayref to arrayrefs contains a 4 element ip address. Reverse lookups are an arrayref to domain names. MX lookups are an arrayref to arraysrefs with the first element as a preference value (0 to 65535) and the second is a domain name.

=back

=cut

