package ActiveState::MIME::Header;

use strict;
use base 'Exporter';

our @EXPORT_OK = qw(header_utf8 _header_utf8);

sub header_utf8 {
    my $h = shift;
    _header_utf8($h);
    return $h;
}

sub _header_utf8 {
    for ($_[0]) {
	return unless /=\?/;  # quick exit

	# find the =?...?= chunks
	# note that we're allowing unencoded spaces in the text part,
	# even though RFC 2047 forbids them, because they've been seen
	# in spam
	my @x = split(/=\?((?:[^?\s]+\?){2}[^?]+\?)=/, $_);

	# decode the =?...?= chunks
	for (my $i = 1; $i < @x; $i += 2) {
	    my($charset, $encoding, $text) = split(/\?/, $x[$i]);
	    $_ = lc($_) for $charset, $encoding;
	    if ($encoding eq "b") {
		require MIME::Base64;
		local $SIG{__WARN__} = sub {};
		$text = MIME::Base64::decode($text);
	    }
	    elsif ($encoding eq "q") {
		$text =~ s/_/ /g;
		$text =~ s/=([\da-fA-F]{2})/chr(hex($1))/ge;
	    }
	    else {
		warn "Unknown rfc2047 encoding '$encoding'";
	    }

	    $x[$i] = _encode_utf8($charset, $text);
	}

	# any space between =?...?= chunks should collapse
	for (my $i = 2; $i < @x; $i += 2) {
	    $x[$i] =~ s/^\s+\z//;
	}

	# update it
	$_ = join("", @x); 
    }
}

sub _encode_utf8 {
    my $charset = shift;
    # deal with those that are trivial
    return $_[0] if $charset eq "us-ascii" || $charset eq "utf8";

    # deal with the almost trivial
    if ($charset eq "iso-8859-1" || $charset eq "iso-8859-15") {
	my $text = shift;
	$text =~ s/([\x80-\xFF])/chr(0b1100_0000 + (ord($1) >> 6)) .
                                 chr(0b1000_0000 + (ord($1) & 0b0011_1111))/ge;
	if ($charset eq "iso-8859-15") {
	    # this is almost latin1 but with these few differences
	    # http://www.cs.tut.fi/~jkorpela/latin9.html
	    for ($text) {
		s/\xc2\xa4/\xe2\x82\xac/g;  # Euro sign
		s/\xc2\xa6/\xc5\xa0/g;      # S with caron
		s/\xc2\xa8/\xc5\xa1/g;      # s with caron
		s/\xc2\xb4/\xc5\xbd/g;      # Z with caron
		s/\xc2\xb8/\xc5\xbe/g;      # z with caron
		s/\xc2\xbc/\xc5\x92/g;      # AE
		s/\xc2\xbd/\xc5\x93/g;      # ae
		s/\xc2\xbe/\xc5\xb8/g;      # Y with diaeresis
	    }
	}
	return $text;
    }

    # :-(  Invoke Encode when we have 5.8.x or perhaps Unicode::Map8
    return $_[0]; # XXX unknown charset
}

1;

__END__

=head1 NAME

ActiveState::MIME::Header - Decode RFC 2047 MIME headers

=head1 SYNOPSIS

 use ActiveState::MIME::Header qw(header_utf8);
 print header_utf8($header_field);

=head1 DESCRIPTION

This module exports the following functions:

=over

=item $utf8 = header_utf8( $field_value )

The input is a header field value as specified in RFC 2047.  The
output is a UTF8 string where all the =?...?= parts has been decoded.
Unencoded spaces are not allowed in an RFC 2047 encoding, however, we're
allowing them in the text part, as they've been seen in spam.

=item _header_utf8( $field_value )

Same as header_utf8(), but will change the value in-place.

=back

=head1 SEE ALSO

RFC 2047

L<Encode::MIME::Header>
