package ActiveState::Message;

use strict;
use XSLoader ();

our $VERSION = '1.0.0.0';

XSLoader::load(__PACKAGE__, $VERSION);

# Backwards-compatibility code
require PureMessage::Policy::MessageObject;
if (PureMessage::Policy::MessageObject->can("headers_ordered")) {
    *enumerate_headers = \&enumerate_headers_v45;
}
else {
    *enumerate_headers = \&enumerate_headers_v4;
}

sub new {
    my ($class, $msg) = @_;
    bless \$msg, ref($class) || $class;
}

sub get_message_size {
    my $self = shift;
    $$self->message_size;
}

sub get_message {
    my $self = shift;
    $$self->message_raw(my $txt, @_);
    $txt;
}

sub get_envelope_sender {
    my $self = shift;
    $$self->minfo_value("envelope-from");
}

sub enumerate_envelope_recipients {
    my ($self, $cb, $data) = @_;
    my $n = 0;
    for my $rcpt ($$self->minfo_value("envelope-to")) {
	++$n;
	if ($cb) {
	    &$cb($data, $rcpt) or last;
	}
    }
    $n;
}

sub get_header_size {
    my $self = shift;
    $$self->header_size;
}

sub get_header {
    my ($self, $offset, $len) = @_;
    my $hdr = $$self->header;
    if (defined $offset and defined $len) {
	return substr($hdr, $offset, $len);
    }
    elsif (defined $offset) {
	return substr($hdr, $offset);
    }
    else {
	return $hdr;
    }
}

sub enumerate_headers_v45 {
    my ($self, $hdr, $cb, $data) = @_;
    my $lc_hdr = lc($hdr||"");
    my $n = 0;
    my %seen;
  HEADER:
    for my $h ($$self->headers_ordered) {
	my $lc_h = lc $h;
	if (defined $hdr) {
	    next unless $lc_h eq $lc_hdr;
	}
	if ($cb) {
	    my $v = $$self->header($h, $seen{$lc_h}++);
	    &$cb($data, $h, $v) or last HEADER;
	}
	else {
	    $n++;
	}
    }
    $n;
}

sub enumerate_headers_v4 {
    my ($self, $hdr, $cb, $data) = @_;
    my $lc_hdr = lc($hdr||"");
    my $n = 0;
  HEADER:
    for my $h ($$self->headers) {
	my $lc_h = lc $h;
	if (defined $hdr) {
	    next unless $lc_h eq $lc_hdr;
	}
	my @v = $$self->header($h);
	if ($cb) {
	    for my $v (@v) {
		++$n;
		&$cb($data, $h, $v) or last HEADER;
	    }
	}
	else {
	    $n += @v;
	}
    }
    $n;
}

sub get_body_size {
    my $self = shift;
    $$self->body_raw_size;
}

sub get_body {
    my $self = shift;
    $$self->body_raw(my $txt, @_);
    $txt;
}

sub get_decoded_body {
    my $self = shift;
    $$self->body(my $txt, @_);
    $txt;
}

sub enumerate_parts {
    my ($self, $cb, $data) = @_;

    my @parts = map { $_->[1] }
		sort { $a->[0] <=> $b->[0] }
		map { (my $id = $_) =~ s/^1\.//; [$id, $_] }
		$$self->parts;
    return int(@parts) unless $cb;

    my $n = 0;
  PART:
    for my $id (@parts) {
	$n++;
	&$cb($data, $self->new($$self->part($id))) or last PART;
    }
    $n;
}

sub get_relay {
    my $self = shift;

    # If PureMessage is configured to talk SMTP, the relay IP information is
    # already in the Received headers. This means that the SpamEngine should
    # not be told about PureMessage's concept of the relay, or else it will
    # see two instances of the real relay. This would lead to much higher risk
    # of false positives, since multiple rules could fire on the same IP
    # address.

    # NOTE: PerlMx < 4.5 didn't have this functionality. PerlMx 4.5.0 did, but
    # in 4.5.0 smtp_mode() could not be called as a class method, and the
    # PerlMx object is not available to Policy plugins. PureMessage 4.5.1 and
    # later make smtp_mode() a class method.
    if (defined $PerlMx::VERSION
	    and $PerlMx::VERSION > 4.500
	    and PerlMx->smtp_mode)
    {
	return undef;
    }

    my $r = $$self->minfo_value("relay");

    # Extract the relay IP address
    my $ip = (split ' ', $r)[-1];
    $ip =~ s/[\[\]]//g;
    $ip;
}

sub get_relay_hostname {
    my $self = shift;
    if (defined $PerlMx::VERSION
	    and $PerlMx::VERSION > 4.500
	    and PerlMx->smtp_mode)
    {
	return undef;
    }
    my $r = $$self->minfo_value("relay");

    # Extract the relay hostname
    (split ' ', $r)[0];
}

sub get_helo {
    my $self = shift;
    $$self->minfo_value("helo");
}

1;
