package ActiveState::Sig;

use strict;

our $VERSION = '1.0';

# class structure:
#
# {
#   save_match => bool,
#
#   plugins => { <plugin1> => 1,
#                <plugin2> => 1,
#                "",
#              },
#
#   This field is so we can prioritize the calling order of the algorithms
#   algos_ordered => [ <algo1>, <algo2>, ... ],
#
#   algos => { <algo1> => \&algo1_calculate_cb,
#              <algo2> => \&algo2_calculate_cb,
#              "",
#            },
#
#   asdb_cache => ref to database,
#
#   max_bytes => max bytes to scan (in the message)
#
#   info => { <algo1> => [ { sig => <hex>,
#                            hit => ref to database entry,
#                            match => the thing matched against,
#                            extra => extra stuff the algo wants to store,
#                          },
#                          { "" },
#                        ],
#             <algo2> => "",
#           },
# }

# To do:
# - $eng being passed around breaks abstraction (d=42157)
#   - we only need this for the log method
#   - plugin attributes should be passed through %options and parsed into
#   plugins => { <plugin-name> => { attr => { <attr-name> => <value> , ... },
#                                   loaded => <bool>
#                                 },
#                ...
#              }
#   and then this hash sent down into the plugin where the options can be
#   parsed further into <algo>.<attr>
#
# - Need a .pm.pod document defining the public methods


sub new {
    my ($class, %options) = @_;
    my $self = bless {
	asdb_cache    => delete $options{asdb_cache},
	save_match    => delete $options{save_match},
        max_bytes     => delete $options{max_bytes},
	info          => {},
	plugins       => {},
	algos         => {},
	algos_ordered => [],
    }, ref($class) || $class;
    $self;
}

sub load_plugin {
    my ($self, $eng, $plugin) = @_;
    return 1 if $self->{plugins}{$plugin};

    $eng->log("DEBUG", "Sig::load_plugin: loading signature plugin: $plugin");

    my $cls = "ActiveState::Sig::Plugin::${plugin}";
    eval "require $cls";
    if ($@) {
	$eng->log("ERR", "Error loading Sig Plugin '$plugin': $@");
    } else {
	_load($self, $eng, $plugin, $cls);
	return 1;
    }
}

sub _load {
    my ($self, $eng, $plugin, $cls) = @_;
    my $method = "${cls}::plugin";
    no strict 'refs';
    $method->($self, $eng, "sig." . lc $plugin);
    $eng->log("DEBUG", "Sig plugin $plugin loaded");
    $self->{plugins}{$plugin} = 1;
}

sub declare_algo {
    my $self = shift;
    while (my $algo = shift) {
	# $algo->[0] is the name
	# $algo->[1] is the callback
	push @{$self->{algos_ordered}}, $algo->[0];
	$self->{algos}->{$algo->[0]} = $algo->[1];
    }
    # XXX return value???
}

# this pattern was repeated everywhere, and not consistently, so I pulled it
# into a single shared sub...
sub record_sig {
    my ($self, $algo, $hexdigest, $match, $algo_extra) = @_;
    push @{$self->{info}{$algo}}, {
	sig => $hexdigest,
	hit => $self->{asdb_cache}->get("$algo:$hexdigest"),
	$self->{save_match} ? (match => $match) : (),
	$algo_extra ? (extra => $algo_extra) : (),
    };
}

sub calculate {
    my ($self, $eng, $xtra, $msg) = @_;

    for my $algo (@{$self->{algos_ordered}}) {
	$self->{algos}{$algo}($self, $eng, $xtra, $msg);
    }

    return $self->{info};
}

sub sigs {
    my $self = shift;
    my $count = grep defined $self->{info}{$_}, keys %{$self->{info}};

    my %sigs = ();
    while(my ($key, $val) = each %{$self->{info}}) {
        for my $item (@$val) {
            push @{$sigs{$key}}, $item->{sig} if exists $item->{sig};
        }
    }

    return wantarray ? %sigs : $count;
};

sub info {
    my $self = shift;
    my $info = $self->{info};
    return wantarray ? %$info : $info;
}

sub reset {
    my $self = shift;
    $self->{info} = {};
}

1;
