use strict;
package ActiveState::Sig::Plugin::NW;

use ActiveState::SpamEngine qw(PMX_TYPE_INTEGER PMX_TYPE_DOUBLE _get1 _getall DEBUG);
use base 'ActiveState::Sig::Plugin';
use Digest::MD5 2.27;

our $VERSION = '0.02';

# attributes to be exposed via SpamEngine
# XXX - these should be declared in the SE sig plugin, not here, so we can
# abstract away the engine from this module (perhaps via a NW/Compile.pm
# module?)
our @ATTRIB = (
    # name           , type            ,  value,
    "body_skip_ratio", PMX_TYPE_DOUBLE ,    0.2,    # nw_body
    "body_stop_ratio", PMX_TYPE_DOUBLE ,    0.8,    # nw_body
    "chunk_size"     , PMX_TYPE_INTEGER,   8192,    # nw_body, nw_html
    "minimum_size"   , PMX_TYPE_INTEGER,     64,    # nw_body, nw_html, nw_para
);

sub plugin {
    my ($sig, $eng, $name) = @_;
    $sig->declare_algo(
    #   algo     , sub_ref
	["nw_para", \&sig_nw_para],
	["nw_body", \&sig_nw_body],
	["nw_html", \&sig_nw_html],
    );

    $eng->declare_attributes( \@ATTRIB , $name );
}

# Gets the attributes listed in @$which
sub sig_nw_attrs {
    my ($sig, $eng, $which) = @_;
    my $attrs = {};
    if (!$which) {
	my @a = @ATTRIB;
	while (@a) {
	    my $name = shift @a;
	    shift @a;
	    shift @a;
	    push @$which, $name;
	}
    }
    for my $attr (@$which) {
	my $temp;
	$eng->enumerate_attributes("plugin.sig.nw.$attr", \&_get1, \$temp);
	$attrs->{$attr} = $temp;
    }
    return $attrs;
}

# Algorithm callback declared to ActiveState::Sig
sub sig_nw_para {
    my ($sig, $eng, $xtra, $msg) = @_;

    my $attrs = sig_nw_attrs($sig, $eng, ['minimum_size']);

    my $para_index = 0;
    my $para_lengths = 0;
    for my $para (@{$xtra->{P}{part}{b_paras}}) {
        my $length = length $para;

        # Note that this length check is only approximate - it will not be
        # *exactly* the same as a comparison against the body length so far,
        # due the different ways we parse b_paras
        if (($para_lengths += $length) > $sig->{max_bytes}) {
            $eng->log(
                'DEBUG',
                "sig: message paragraphs overflow max bytes ($sig->{max_bytes}); aborting early"
            ) if DEBUG;
            last;
        }

	if ($length > $attrs->{minimum_size}) {
	    $sig->record_sig('nw_para', Digest::MD5::md5_hex($para), $para,
		{para_index => $para_index});
	}
	$para_index += 1;
    }
}

# this call is a bootstrap for the recursive call below and sets up the
# parameters in the same way enumerate_parts is going to pass them along. It
# also inits a new MD5 to use.
sub sig_nw_body {
    my ($sig, $eng, $xtra, $msg) = @_;

    # bail if body size is over max_bytes
    my $size = $msg->get_body_size;
    if ($size > $sig->{max_bytes}) {
        $eng->log(
            'DEBUG',
            "sig: message size ($size) exceeds max bytes ($sig->{max_bytes})"
        ) if DEBUG;
        return 0;
    }

    recursive_sig_nw_body(
	[$sig, sig_nw_attrs($sig, $eng), Digest::MD5->new(), 0, [], $eng],
	$msg
    );
}

{
    package normalizer;
    sub new {
        my ($this, $msg) = @_;
        return bless {
            'offset' => 0,
            total => $msg->get_body_size,
            msg => $msg,
            ispace => 0,
        }, ref $this || $this;
    }

    sub reset {
        my ($self) = @_;
        $self->{'offset'} = 0;
        $self->{'ispace'} = '';
        return;
    }

    sub read {
        my ($self, $chunksize) = @_;
        $chunksize ||= 8192;

        while ($self->{'offset'} < $self->{'total'}) {
            my $chunk = $self->{msg}->get_decoded_body(
                    $self->{'offset'},
                    $chunksize);
            my $cs = length($chunk);
            return undef if $cs == 0;

            if ($self->{ispace} && $chunk !~ m/^\s/) {
                # if previous chunk ended in a whitespace, and this chunk didn't,
                # then we need to prepend a space.
                $chunk = ' ' . $chunk;
            }

            $chunk =~ s/^\s+// unless $self->{'offset'};
            $self->{'offset'} += $cs;

            $self->{ispace} = ($chunk =~ s/\s+$//) ? 1 : 0 ;
            next if $chunk eq '';

            $chunk =~ s/\s+/ /g;
            return $chunk;
        }
        return undef;
    }
}

# this call is structured this way because we recursively descend into the
# message object, which only provides a single $data variable for us
# to pass around.
sub recursive_sig_nw_body {
    my ($data, $msg) = @_;
    my ($sig, $attrs, $md5, $level, $parent_ct, $eng) = @$data;

    # process this message part passed in ($msg)
    my $normalizer = normalizer->new($msg);

    my $offset = 0;
    my $max = 0;
    my $chunk;
    while (defined ($chunk = $normalizer->read($attrs->{chunk_size}))) {
        $max += length($chunk);
    }

    $normalizer->reset;

    my @ct;
    $msg->enumerate_headers("Content-Type", \&_getall, \@ct);

    # convert ct-header to mimetypes only.
    @ct = map { (/^\s*([^;\s]+)/)[0] } @ct;

    my $md5_original = $md5->clone;

    if (grep /^text\b/i, @ct) {

        # Determine how much to discard off the front and the back.
	$offset = int($attrs->{body_skip_ratio} * $max);
        my $remaining = int($attrs->{body_stop_ratio} * $max) - $offset;

	if ($remaining > $attrs->{minimum_size}) {
	    $md5->add(join "\x00", (@ct, $level, "\x00"));

            my @txt = ();
            while (defined ($chunk = $normalizer->read($attrs->{chunk_size}))) {
                if ($offset) {
                    if (length($chunk) > $offset) {
                        substr($chunk, 0, $offset, '');
                        $offset = 0;
                    } else {
                        $offset -= length($chunk);
                        next;
                    }
                }
                last unless $remaining;

                if ($remaining > length($chunk)) {
                    $remaining -= length($chunk);
                } else {
                    substr($chunk, $remaining) = '';
                    $remaining = 0;
                }
                push @txt, $chunk if $sig->{save_match};
                $md5->add($chunk);
            }

            $sig->record_sig('nw_body', $md5->hexdigest, join('', @txt),
                { parent_ct => $parent_ct, ct => \@ct, level => $level }
            );
	}

        if ($ENV{'SOPHOSLABS_DEBUG'}) {

            $offset = int($attrs->{body_skip_ratio} * $msg->get_body_size);
            $max    = int($attrs->{body_stop_ratio} * $msg->get_body_size);

            if ($max - $offset > $attrs->{minimum_size}) {
                $md5_original->add(join "\x00", (@ct, $level, "\x00"));
                my $length = 0;
                my @txt;
              LOOP_BODY:
                while ($offset < $max) {
                    for my $chunk ($msg->get_decoded_body($offset, $attrs->{chunk_size})) {
                        push @txt, $chunk if $sig->{save_match};
                        $md5_original->add($chunk);
                        $offset += length($chunk);
                        $length += length($chunk);
                        last LOOP_BODY if length($chunk) < $attrs->{chunk_size};
                    }
                }

                # The _decoded_ body is possibly shorter than the encoded body, so
                # repeat the test here.
                if ($length > $attrs->{minimum_size}) {
                    $eng->log('DEBUG', "sig: old body signature: " . $md5_original->hexdigest);
                }
            }
        }

    }

    # process the _children_ of $msg

    $msg->enumerate_parts(\&recursive_sig_nw_body,
	[ $sig, $attrs, $md5, $level + 1, \@ct, $eng ]
    );

    return 1;
}

# see sig_nw_body above for the reasons behind the struct of the calls below.
sub sig_nw_html {
    my ($sig, $eng, $xtra, $msg) = @_;

    # bail if body size is over max_bytes
    my $size = $msg->get_body_size;
    if ($size > $sig->{max_bytes}) {
        $eng->log(
            'DEBUG',
            "sig: message size ($size) exceeds max bytes ($sig->{max_bytes})"
        ) if DEBUG;
        return 0;
    }

    recursive_sig_nw_html(
	[$sig, sig_nw_attrs($sig, $eng), Digest::MD5->new(), 0, []],
	$msg
    );
}

sub recursive_sig_nw_html {
    my ($data, $msg) = @_;
    my ($sig, $attrs, $md5, $level, $parent_ct) = @$data;

    my @ct;
    $msg->enumerate_headers("Content-Type", \&_getall, \@ct);
    # convert ct-header to mimetypes only.
    @ct = map { (/^\s*([^;\s]+)/)[0] } @ct;

    # If any @ct looks matches "\bhtml\b", calculate a signature.
    my ($offset, $max) = (0, $msg->get_body_size);
    if ($max > $attrs->{minimum_size} && (grep /\bhtml\b/i, @ct)) {
	my ($l, @t) = 0;
	my ($l2, @t2) = 0;
	$md5->add(join "\x00", (@ct, $level, "\x00"));
	my $md52 = $md5->clone;
	my $p = HTML::Parser->new(
	    api_version => 3,
	    start_h => [sub {
			    # For simplicity, always add the extra space after
			    # the tag: <a ><br ><a name>
			    for ("<$_[0] @{$_[1]}>") {
				push @t, $_ if $sig->{save_match};
				$md5->add($_);
				$l += length;
			    }
			}, "tagname,attrseq"],
	    text_h  => [sub {
			    return unless shift; # is_cdata
			    for (shift) {
				push @t2, $_ if $sig->{save_match};
				$md52->add($_);
				$l2 += length;
			    }
			}, "is_cdata,dtext"],
	    end_h   => [sub {
			    for ("</$_[0]>") {
				push @t, $_ if $sig->{save_match};
				$md5->add($_);
				$l += length;
			    }
			}, "tagname"],
	);

      LOOP_HTML:
	while ($offset < $max) {
	    for my $chunk ($msg->get_decoded_body($offset, $attrs->{chunk_size})) {
		$p->parse($chunk);
		$offset += length($chunk);
		last LOOP_HTML if length($chunk) < $attrs->{chunk_size};
	    }
	}
	$p->eof;
	if ($l > $attrs->{minimum_size}) {
	    $sig->record_sig('nw_html', $md5->hexdigest, join('', @t),
		{parent_ct => $parent_ct, ct => \@ct, level => $level }
	    );
	}
	if ($l2 > $attrs->{minimum_size}) {
	    $sig->record_sig('nw_scpt', $md52->hexdigest, join('', @t2),
		{parent_ct => $parent_ct, ct => \@ct, level => $level }
	    );
	}
    }
    $msg->enumerate_parts(\&recursive_sig_nw_html, 
	[ $sig, $attrs, $md5, $level + 1, \@ct ]
    );
    return 1;
}

1;
