package ActiveState::SpamEngine;

use strict;
use base 'Exporter';

use ActiveState::DB;
use Sophos::Statistics;
use Sophos::SXL;
use Sophos::SXL::Limit;
use ActiveState::PCRE;
use ActiveState::Genotype;
use Sophos::Debug qw(_debug_with_caller DEBUG);

our @EXPORT_OK = qw(PMX_TYPE_NULL PMX_TYPE_CALLBACK PMX_TYPE_DOUBLE
                    PMX_TYPE_INTEGER PMX_TYPE_STRING PMX_TYPE_MESSAGE
                    PMX_TYPE_HASH PMX_TYPE_WEIGHTS
                    PMX_EV_FOUND_FEATURE PMX_EV_SPAMPROB PMX_EV_FUR
                    DEBUG

                    _get1 _getall
                );

# Increment for every engine release.
# Don't forget to update SpamEngine/Compiler.pm also, as well as
# package/sdk/make in Sophos-AntiSpam-DataPackager-SDK!
# Note that if the version is three digits only, then the build number will be
# automatically appended
our $VERSION = '2.7.2';
our $SCHEMA_VERSION = '17';

sub PMX_TYPE_NULL     () {0}
sub PMX_TYPE_CALLBACK () {1}
sub PMX_TYPE_DOUBLE   () {2}
sub PMX_TYPE_INTEGER  () {3}
sub PMX_TYPE_STRING   () {4}
sub PMX_TYPE_MESSAGE  () {5}
sub PMX_TYPE_HASH     () {6}
sub PMX_TYPE_WEIGHTS  () {7}

sub PMX_EV_FOUND_FEATURE () {1}
sub PMX_EV_SPAMPROB      () {2}
sub PMX_EV_FUR           () {3}

sub STOP_SCAN_MAX_THRESHOLD_DEFAULT () { 1e6  }
sub STOP_SCAN_MIN_THRESHOLD_DEFAULT ()  { -1e6 }

our @SP_ATTRIB = (
    # KEY                           TYPE                DEFAULT
    "core.version",                 PMX_TYPE_STRING,    $VERSION,
    "core.plugin-dir",              PMX_TYPE_STRING,    [],
    "core.max-bytes-scanned",       PMX_TYPE_INTEGER,   10 * 1024,
    "core.stop-scan",               PMX_TYPE_INTEGER,   0,
    "core.local-tests-only",        PMX_TYPE_INTEGER,   0,
    
    # Message opts are stored in a hash so they can be easily changed per
    # message by the application.  e.g. 'authed' for SMTP authenticated.
    #
    "core.message-opts",            PMX_TYPE_HASH,      undef,

    "core.train",                   PMX_TYPE_INTEGER,   undef,
    "core.is-spam",                 PMX_TYPE_INTEGER,   undef,
    "core.optimal-plugins",         PMX_TYPE_STRING,    [qw(_compatv1)],
    "core.compile",                 PMX_TYPE_INTEGER,   1,
    "core.datadir",                 PMX_TYPE_STRING,    undef,
    "core.userdatadir",             PMX_TYPE_STRING,    undef,
    "core.datafile",                PMX_TYPE_STRING,    "asdb.antispam",
    "core.schema-version",          PMX_TYPE_INTEGER,   $SCHEMA_VERSION,
    "core.data-version",            PMX_TYPE_STRING,    undef,
    "core.profile-level",           PMX_TYPE_INTEGER,   0,
    "core.event.fur",               PMX_TYPE_INTEGER,   0,
    "core.statistics-directory",    PMX_TYPE_STRING,    undef,

    # Allow overriding the stop-scan (a.k.a. early-out) scores.
    "core.stop-scan-max-threshold", PMX_TYPE_DOUBLE,    undef,
    "core.stop-scan-min-threshold", PMX_TYPE_DOUBLE,    undef,

    # This attribute isn't actually used yet, but it's required for SXL
    # forward compatibility. The network services require both a vendor id and
    # client id; the vendor id comes from the data, and the client id must be
    # set by the application.
    "core.client-id",               PMX_TYPE_STRING,    undef,

    # Buggy applications [which?] try to set these legacies before loading
    # "re" -- it would be nice if we could get rid of these
    "plugin.net.skip-rbl-checks",   PMX_TYPE_INTEGER,   undef,
    "plugin.re.skip-rbl-checks",    PMX_TYPE_INTEGER,   undef,
    "plugin.re.skip-dns-checks",    PMX_TYPE_INTEGER,   undef,
    "plugin.re.dns-timeout",        PMX_TYPE_INTEGER,   undef,
    "plugin.re.dns-retry",          PMX_TYPE_INTEGER,   undef,
    "plugin.re.dns-retrans",        PMX_TYPE_INTEGER,   undef,
    "plugin.re.ok-locales",         PMX_TYPE_STRING,    undef,
);

# The wrapper code injects the following functions for invoking callbacks:
#
# - attribute_cb($cb, $data, $attr, $type, $value);
# - plugin_cb($cb, $data, $plugin)
# - scan_cb($cb, $data, $event, $t1, $p1, $t2, $p2);
# - error_cb($engine, $level, $error);

sub create_message {
    my $engine = shift;
    require ActiveState::SpamEngine::Message;
    return ActiveState::SpamEngine::Message->new;
}

sub create_hash {
    my $engine = shift;
    require ActiveState::SpamEngine::Hash;
    return ActiveState::SpamEngine::Hash->new(@_);
}

#=============================================================================
# SpamEngine API
#=============================================================================

sub _get1   { ${$_[0]} = $_[2]; 0 }
sub _getall { push @{$_[0]}, $_[2] if defined $_[2]; 1 }

sub create_engine { return new(__PACKAGE__) }

sub new {
    my ($class, $logger) = @_;
    my $engine = bless {}, ref($class) || $class;
    $engine->{logger} = $logger;
    $engine->{loaded_plugin} = {};
    $engine->{profile} = 0; # non-zero when profiling is enabled
    $engine->declare_attributes(\@SP_ATTRIB);
    return $engine;
}

sub DESTROY {
    my ($engine) = @_;
    $engine->broadcast(undef, "core.unload");

    # Arrange to call del_attribute() for each attribute whose type is
    # qw(MESSAGE HASH WEIGHTS):
    $engine->enumerate_attributes(undef, sub {
        my ($engine, $key) = @_;
        if (grep { $engine->{type}{$key} eq $_ } (
                PMX_TYPE_MESSAGE, PMX_TYPE_HASH, PMX_TYPE_WEIGHTS
            ))
        {
            $engine->del_attribute($key, 1);
        }
        1;
    }, $engine);
}

sub declare_attributes {
    my ($engine, $list, $plugin) = @_;
    my @copy = @$list;
    while (@copy) {
        my ($attr, $type, $default) = splice(@copy, 0, 3);
        $attr = "plugin.$plugin.$attr" if $attr !~ /\./;
        $engine->{type}{$attr} = $type;
        $engine->log("DEBUG", "declare_attributes: $attr t=$type") if DEBUG;
        if (defined $default) {
            $default = [$default] unless ref($default) eq 'ARRAY';
            $engine->{attr}{$attr} = $default;
        }
    }
}

sub add_attribute {
    my ($engine, $attr, $type, $data, $repl) = @_;

    my $exists = exists $engine->{type}{$attr};

    if (DEBUG) {
        local($^W);
        $engine->log("DEBUG", "add_attribute: $attr t=$type d=$data r=$repl x=$exists");
    }

    die "error: attribute $attr is type $engine->{type}{$attr} and *not* $type\n"
        if defined($engine->{type}{$attr}) && $engine->{type}{$attr} != $type;
    $engine->{type}{$attr} = $type;

    if ($repl) {
        $engine->{attr}{$attr} = [$data];
    }
    else {
        push @{$engine->{attr}{$attr}}, $data;
    }

    $engine->log("WARNING", "add_attribute: $attr was not declared") unless $exists;

    # XXX Normally we want to return an error code if a developer spells an
    # attribute wrong; however, our current customers have all made mistakes,
    # so in DEBUG mode, we'll ignore errors so that we find as many mistakes
    # as possible.
    return 1 if DEBUG;

    return $exists;
}

sub del_attribute {
    my ($engine, $attr, $all) = @_;
    $engine->log("DEBUG", "del_attribute: $attr all=$all") if DEBUG;
    my $deleted;
    if ($all) {
        $deleted = delete $engine->{attr}{$attr};
    }
    else {
        $deleted = shift @{$engine->{attr}{$attr}};
    }
    return unless $deleted;

    my $type = $engine->{type}{$attr} or return; # was never defined
    return unless $type == PMX_TYPE_MESSAGE || $type == PMX_TYPE_HASH ||
                  $type == PMX_TYPE_WEIGHTS;
    $deleted = [$deleted] unless ref $deleted;

    foreach my $obj (@$deleted) {
        $obj->_destroy if UNIVERSAL::can($obj, "_destroy");
    }
}

sub enumerate_attributes {
    my ($engine, $attr, $cb, $data) = @_;
    if (DEBUG) {
        $engine->log(
            "DEBUG",
            "enumerate_attributes: " . (defined $attr ? $attr : "undef")
        );
    }
    my $n = 0;
  ATTR:
    for my $attr (defined $attr ? $attr : keys %{$engine->{attr}}) {
        unless ($cb) {
            $n += $engine->{attr}{$attr} ? @{$engine->{attr}{$attr}} : 0;
            next;
        }
        for my $a (@{$engine->{attr}{$attr}}) {
            ++$n;
            if (ref($cb) eq 'CODE') {
                $cb->($data, $attr, $a) or last ATTR;
            }
            else {
                my $t = $engine->{type}{$attr};
                attribute_cb($cb, $data, $attr, $t, $a) or last ATTR;
            }
        }
    }
    return $n;
}

my @builtin_plugins = qw(
    _compatv1 bayes dnsbl genotype net sigv2
    part re redb sig spamassassin sxl uri word_obfu
    spf fur freeweb rip uri_dns
);

my %builtin_plugins;
@builtin_plugins{@builtin_plugins} = (1) x scalar(@builtin_plugins);

sub load_plugin {
    my ($engine, $plugin) = @_;
    return 1 if $engine->{loaded_plugin}{$plugin};

    $engine->log("DEBUG", "load_plugin: loading plugin '$plugin'") if DEBUG;

    # Check to see if we can just require it:
    my $class = "ActiveState::SpamEngine::Plugin::${plugin}";

    if ($builtin_plugins{$plugin}) {
        eval "require $class";
        unless ($@) {
            _load($engine, $plugin, $class);
            return 1;
        }
        $engine->log("ERR", "error: loading built-in plugin '$plugin': $@");
        return 0;
    }

    my @paths;
    $engine->enumerate_attributes('core.plugin-dir', sub {
        my ($data, undef, $p) = @_;
        push @paths, $p;
        my $base = "$p/$data->[1]";
        my $f = "$base.pm"; # perl module -- C module would be .so
        my $m = "$base.md5";
        my $s = "$base.md5.sig";
        return 1 unless -f $f;

        # In C, this would do a dl_open (or nothing, for builtin plugins),
        # then load the "as_plugin_$plugin" symbol and call it. That symbol
        # returns a struct which contains the input requirements, what
        # scanners and preprocessors the plugin provides, what other plugins
        # it requires, and the function pointer to call when a message is
        # scanned.
        #
        # In Perl, we just 'require' the file, and then call the
        # AS::SpamEngine::Plugin::$plugin::plugin() method, which returns a
        # hash reference with the same information.
        {
            eval { require $f };
            if ($@) {
                $engine->log("WARNING", "Error loading plugin '$plugin': $@");
                return 1; # continue enumerating
            }
            else {
                _load(@$data);
            }
        }

        0; # stop enumerating
    }, [$engine, $plugin, $class]);

    unless ($engine->{loaded_plugin}{$plugin}) {
        $engine->log("ERR", "error: plugin '$plugin' not loaded [@paths].");
        return 0;
    }
    1;
}

sub _load {
    my ($engine, $plugin, $class) = @_;
    my $method = "${class}::plugin";
    no strict 'refs';
    defined &$method
        or die "error: plugin '$plugin' does not export plugin()";
    $method->($engine, $plugin)
        or die "error: plugin '$plugin' did not initialise";
    $engine->{loaded_plugin}{$plugin} = 1;
    $engine->log("DEBUG", "_load: plugin '$plugin' loaded") if DEBUG;
}

sub enumerate_loaded_plugins {
    my ($engine, $cb, $data) = @_;
    if (DEBUG) {
        my @p = keys %{$engine->{loaded_plugin}};
        $engine->log("DEBUG", "enumerate_loaded_plugins: [@p]");
    }
    return int(keys %{$engine->{loaded_plugin}}) unless $cb;

    my $n = 0;
    for my $k (keys %{$engine->{loaded_plugin}}) {
        ++$n;
        if (ref($cb) eq 'CODE') {
            $cb->($data, $k) or last;
        }
        else {
            plugin_cb($cb, $data, $k) or last;
        }
    }
    return $n;
}

sub enumerate_plugins {
    my ($engine, $cb, $data) = @_;
    $engine->log("DEBUG", "enumerate_plugins: ") if DEBUG;
    my @plugins = keys %builtin_plugins;
    $engine->enumerate_attributes('core.plugin-dir', sub {
        my ($data, undef, $p) = @_;
        return 1 unless -d $p;
        opendir(my $DIR, $p) or die "can't opendir $p: $!";
        while (local $_ = readdir($DIR)) {
            next unless /^([a-zA-Z]\w*)(?:\.pm)$/;
            my $pl = $1;
            my $f = "$p/$_";
            next unless -f $f; # and -f "$p/$_.md5" and -f "$p/$_.md5.sig";
            push @$data, $pl;
        }
        closedir($DIR);
        1;
    }, \@plugins);
    return int(@plugins) unless $cb;

    my $n = 0;
    my %seen;
    for my $p (@plugins) {
        next if $seen{$p}++;
        ++$n;
        if (ref($cb) eq 'CODE') {
            $cb->($data, $p) or last;
        }
        else {
            plugin_cb($cb, $data, $p) or last;
        }
    }
    return $n;
}

sub initialised {
    my ($engine, $plugin) = @_;
    $engine->{initialised}{$plugin} = 1;
}

sub load_antispam_data {
    my $engine = shift;
    my $cdd;
    $engine->enumerate_attributes("core.datadir", \&_get1, \$cdd);

    $engine->log("DEBUG", "load_antispam_data: loading") if DEBUG;
    my $dbfile;
    $engine->enumerate_attributes("core.datafile", \&_get1, \$dbfile);
    if ($dbfile and -f "$cdd/$dbfile") {
        my $db = ActiveState::DB->new("$cdd/$dbfile");
        $engine->{_dbman} = ActiveState::DB::Manager->new($db);
        $engine->{_dbversion} = $engine->{_dbman}->nest_manager($SCHEMA_VERSION);

        # special case
        delete $engine->{pcre};
    }
    else {
        die "load_antispam_data(): $dbfile not found in core.datadir.\n";
    }

    # FIXME: The SXL object will probably have to read configuration from DFTs,
    #        and to do this it will need access to $engine->{_dbman}.  However,
    #        we don't need this right now.
    $engine->{_sxl} = Sophos::SXL->new($engine);
    $engine->{_sxldft} = {};

    # Set the core.data-version attribute.
    $engine->add_attribute(
        "core.data-version",
        PMX_TYPE_STRING,
        $engine->db->get("data_version"),
        1,
    );

    # Load whatever plugins are required
    $engine->log("DEBUG", "load_antispam_data: auto-loading plugins") if DEBUG;
    my @load_order;
    $engine->nested_db("plugins")->enumerate(sub {
        my ($load, $plugin, $pri) = @_;
        # Older compilers always produced priority = 1, but this is more
        # forgiving.
        $pri ||= 1;
        push @{$load->[$pri]}, $plugin;
        1;
    }, \@load_order);
    for my $entry (grep defined, @load_order) {
        $engine->load_plugin($_) for @$entry;
    }
}

sub compile {
    my $engine = shift;

    # Clear out all events and subscribers except core.compile and core.unload
    $engine->log("DEBUG", "compile: clearing events") if DEBUG;
    $engine->reset_events;

    # Refresh data and load/unload plugins
    $engine->log("DEBUG", "compile: loading antispam data") if DEBUG;
    $engine->load_antispam_data();

    # Clear out the hash of initialised plugins. We use initialised plugins,
    # not loaded plugins, to satisfy meta rule dependencies.
    $engine->{initialised} = {};

    $engine->enumerate_attributes("core.profile-level", \&_get1, \$engine->{profile});

    $engine->broadcast(undef, "core.compile");

    # Set up the meta-rule engine
    init_meta($engine);

    $engine->{w} = $engine->nested_db("w");

    $engine->enumerate_attributes("core.stop-scan", \&_get1, \$engine->{stop_scan});

    # Calculate some stuff from the weights.
    {
        my ($stop_scan_max, $stop_scan_min, $bias);

        $bias = $engine->{w}->get("??bias??");
        $bias = 0 unless defined $bias;

        # If min-threshold > bias, everything breaks. Prevent that.
        $engine->enumerate_attributes("core.stop-scan-min-threshold", \&_get1, \$stop_scan_min);

        $stop_scan_min = $engine->{w}->get("??stop_scan_min_threshold??") unless defined $stop_scan_min;
        $stop_scan_min = STOP_SCAN_MIN_THRESHOLD_DEFAULT if
            not defined $stop_scan_min or $stop_scan_min > $bias;

        $engine->enumerate_attributes("core.stop-scan-max-threshold", \&_get1, \$stop_scan_max);

        $stop_scan_max = $engine->{w}->get("??stop_scan_max_threshold??") unless defined $stop_scan_max;
        $stop_scan_max = STOP_SCAN_MAX_THRESHOLD_DEFAULT unless defined $stop_scan_max;

        $engine->log("DEBUG", "bias=$bias, stop_scan_max=$stop_scan_max, stop_scan_min=$stop_scan_min") if DEBUG;
        @$engine{qw(__bias__ __stop_scan_max__ __stop_scan_min__)} = ($bias, $stop_scan_max, $stop_scan_min);
    }

    $engine->add_attribute("core.compile", PMX_TYPE_INTEGER, 0, 1);
}

sub scan_message {
    my ($engine, $msg, undef, $cb, $data) = @_;
    $engine->log('DEBUG', 'scan_message: starting up') if DEBUG;
    my ($start_time, undef, undef, undef) = times;

    # Call compile if 'core.compile' is true. Plugins which need to query
    # the attributes can then cache attributes internally.
    my $compile;
    $engine->enumerate_attributes('core.compile', \&_get1, \$compile);
    $engine->compile if $compile;

    my $weights = $engine->{w};

    # If profiling is enabled, reset the timer
    $engine->profile_reset if $engine->{profile};

    my %state = (
        feats   => {},
        w       => $weights,
        cb      => $cb,
        data    => $data,
        sum     => $engine->{__bias__},
        max_hi  => $engine->{__stop_scan_max__},
        max_lo  => $engine->{__stop_scan_min__},
        prob    => 0,

        # plugin-specific state data goes here. Plugins are free to use this
        # area for temporary data that can go away after this message is
        # scanned. To avoid clobbering, use $s{P}{$plugin-name}{...}
        P       => {},
    );

    $engine->{'_sxl'}->reset_count() if $engine->{'_sxl'};
    $engine->{'_sxldft'}{$_}->reset_count() foreach keys %{$engine->{'_sxldft'}};

    eval {
        $engine->broadcast(\%state, 'core.scan_message', $msg);
        eval_meta($engine, \%state, $msg);
        $engine->broadcast(\%state, 'core.meta_scan', $msg);
    };
    if ($@) {
        die if $@ !~ /STOP_SCAN/;
        $engine->log('DEBUG', 'found spam/not-spam decision point: stopping early') if DEBUG;
    }
    $engine->log('DEBUG', "biased score: $state{sum}") if DEBUG;

    # fetch correction factor
    my $corr = $state{w}->get('??corr??') || 0.5;

    # $sum is the total weight for this message. Now we have to convert to a
    # probability:
    my $p = $state{prob} = 1 / (1 + exp(-$state{sum} * $corr));

    $engine->broadcast(\%state, 'core.autolearn', $msg);

    my $stats = $engine->statistics();
    my $is_spam = ($p >= 0.5) ? 1 : 0;
    for my $rule (keys %{$state{feats}}) {
        $stats->increment('rules', $rule, "total");
        #adds a 0 for ham so that the in_spam=0 label is guaranteed to be there
        $stats->add('rules', $rule, "in_spam", $is_spam);
    }
    $stats->add('timing', 'cpu', ((times)[0] - $start_time) * 1000);
    $stats->increment('message-count');
    $stats->increment('spam-count') if $is_spam eq 1;
    $stats->store_stats();

    if (ref($cb) eq 'CODE') {
        $cb->($data, PMX_EV_SPAMPROB, $p);
    }
    else {
        scan_cb($cb, $data, PMX_EV_SPAMPROB, PMX_TYPE_DOUBLE, $p,
                PMX_TYPE_NULL, undef);
    }
    my $enable_fur_event;
    $engine->enumerate_attributes('core.event.fur', \&_get1, \$enable_fur_event);

    if (exists($state{fur}) && $enable_fur_event) {
        if (ref($cb) eq 'CODE') {
            $cb->($data, PMX_EV_FUR, $state{fur});
        }
        else {
            scan_cb($cb, $data, PMX_EV_FUR, PMX_TYPE_STRING, $state{fur},
                    PMX_TYPE_NULL, undef);
        }
    }
}

sub addfeature {
    my ($engine, $state, $hit, $info) = @_;
    return if $state->{feats}{$hit}++;
    my $weight = $state->{w}->get($hit) || 0;
    $state->{sum} += $weight;
    $engine->log("DEBUG", "feature: $hit $weight" . ($info ? " [$info]" : ""))
        if DEBUG;
    if (ref($state->{cb}) eq 'CODE') {
        $state->{cb}->($state->{data}, PMX_EV_FOUND_FEATURE, $hit, $info);
    }
    else {
        my $t2 = PMX_TYPE_NULL;
        my $p2 = undef;
        if (defined $info) {
            $t2 = PMX_TYPE_STRING;
            $p2 = $info;
        }
        scan_cb($state->{cb}, $state->{data}, PMX_EV_FOUND_FEATURE,
                PMX_TYPE_STRING, $hit, $t2, $p2);
    }
    $engine->log("DEBUG", "max_lo=$state->{max_lo} sum=$state->{sum} max_hi=$state->{max_hi}") if DEBUG;

    if ($hit ne '__STOP_SCAN' and $weight > 0 and $engine->{stop_scan} and
            ($state->{sum} >= $state->{max_hi} or
            $state->{sum} <  $state->{max_lo})) {
        $engine->addfeature($state, '__STOP_SCAN');
        die "STOP_SCAN";
    }
}

sub log {
    my ($engine, $level, $msg) = @_;
    chomp $msg;

    _debug_with_caller([caller(1)], "$level $msg") if DEBUG;

    if (defined &error_cb) {
        error_cb($engine, $level, $msg);
    }
    elsif ($engine->{logger}) {
        $engine->{logger}->($level, $msg);
    }
    else {
        print "log: $level $msg\n";
    }
}


# Profiling support
sub profile_reset {
    my $engine = shift;
    eval {
        require ActiveState::StopWatch;
        $engine->{timer} = ActiveState::StopWatch::start_watch();
        $engine->log("DEBUG", "profile [0s] starting profiler");
    };
    $engine->log("DEBUG", "profiler died: $@") if $@;
}

sub profile_now {
    my $engine = shift;
    my $result = eval { ActiveState::StopWatch::real_time($engine->{timer}); };
    $engine->log("DEBUG", "profiler died: $@") if $@;
    return $result;
}

sub profile {
    my ($engine, $level, $msg) = @_;
    return if $level > $engine->{profile};
    my $t = eval { ActiveState::StopWatch::real_time($engine->{timer}); };
    $engine->log("DEBUG", "profiler died: $@") if $@;
    $engine->log("DEBUG", sprintf("profile [%.6fs] $msg", $t));
}

sub profile_this_time {
    my ($engine, $level, $msg, $time) = @_;
    return if $level > $engine->{profile};
    $msg =~ s/\%\%TIME\%\%/sprintf("%.6fs", $time)/eg;
    my $t = eval { ActiveState::StopWatch::real_time(\$engine->{timer}); };
    $engine->log("DEBUG", "profiler died: $@") if $@;
    $engine->log("DEBUG", sprintf("profile [%.6fs] $msg", $t));
}

# Database
sub db {
    my $engine = shift;
    return $engine->{_dbversion}->master;
}

sub nested_db {
    my ($engine, $key) = @_;
    return $engine->{_dbversion}->nest($key);
}

sub nested_manager {
    my ($engine, $key) = @_;
    return $engine->{_dbversion}->nest_manager($key);
}

sub nested_object {
    my ($engine, $key, $class) = @_;
    return $engine->{_dbversion}->nest_object($key, $class);
}

sub pcre {
    my $engine = shift;
    $engine->{pcre} ||= ActiveState::PCRE->new($engine->nested_db("pcre"));
}

sub statistics {
    my ($engine) = @_;
    if(!defined $engine->{_statistics}) {
        my $stats_dir;
        $engine->enumerate_attributes("core.statistics-directory", \&_get1, \$stats_dir);
        if (! $stats_dir) {
            # If the statistics directory was not specified, then we'll put the
            # engine statistics into a 'stats' subdirectory of core.datadir.
            $engine->enumerate_attributes("core.datadir", \&_get1, \$stats_dir);
            $stats_dir .= '/stats' if $stats_dir;
        }
        $engine->{_statistics} = Sophos::Statistics->new($stats_dir);
    }
    return $engine->{_statistics};
}

# SXL
sub sxl {
    my ($engine, $type) = @_;

    # The main SXL object isn't used directly by plugins
    return $engine->{_sxl} unless $type;

    if (!exists $engine->{_sxldft}{$type})
    {
        $engine->{_sxldft}{$type} = Sophos::SXL::Limit::select_limit($engine->{_sxl}, $type);
    }
    return $engine->{_sxldft}{$type};
}

# Meta rules
sub init_meta {
    my $engine = shift;

    # Reset
    $engine->{_mr_ids} = [];
    $engine->{_mr_blobdb} = undef;
    $engine->{_mr_meta} = undef;

    my $dep = $engine->nested_db("meta.deps");
    $dep->enumerate(sub {
        my ($data, $id, $reqs) = @_;
        for my $plugin (split ' ', $reqs) {
            unless ($data->[1]{$plugin}) {
                # XXX log it?
                return 1;
            }
        }
        push @{$data->[0]}, $id;
        return 1;
    }, [$engine->{_mr_ids}, $engine->{initialised}]);
    return unless @{$engine->{_mr_ids}}; # no meta rules are needed

    $engine->{_mr_blobdb} = $engine->nested_db("meta.blob");
    $engine->{_mr_meta}   = $engine->nested_object(
        "meta.code",
        "ActiveState::Genotype::DB"
    );
    $engine->{_mr_meta}->init(1); # use sequential access
}
sub eval_meta {
    my ($engine, $state, $msg) = @_;
    return unless $engine->{_mr_blobdb} and $engine->{_mr_meta};
    $engine->{_mr_meta}->run_seq(
        $engine->{_mr_ids},
        $msg,
        0,
        $engine,
        $state,
        $engine->{_mr_blobdb}->_asdb,
    );
}

# Internal events
sub reset_events {
    my ($engine) = @_;
    for my $plugin (keys %{$engine->{_ev_SUBSCRIPTION}}) {
        for my $event (keys %{$engine->{_ev_SUBSCRIPTION}{$plugin}}) {
            next if grep { $event eq $_ } qw(core.compile core.unload);
            delete $engine->{_ev_EVENT}{$event};
            delete $engine->{_ev_SUBSCRIPTION}{$plugin}{$event};
        }
    }
}

sub subscribe {
    my ($engine, $plugin, $event, $callback) = @_;
    # Not all events are known ahead of time, so we can't try to detect
    # misspellings here.
    $engine->log("DEBUG", "subscribe $plugin to $event") if DEBUG;
    push @{$engine->{_ev_PA}}, $plugin unless $engine->{_ev_PH}{$plugin}++;
    $engine->{_ev_EVENT}{$event} = 0 unless exists $engine->{_ev_EVENT}{$event};
    push @{$engine->{_ev_SUBSCRIPTION}{$plugin}{$event}}, $callback;
    ++$engine->{_ev_EVENT}{$event};
}

sub unsubscribe {
    my ($engine, $plugin, $event) = @_;
    if (defined $event) {
        my $cblist = delete $engine->{_ev_SUBSCRIPTION}{$plugin}{$event};
        $engine->{_ev_EVENT}{$event} -= @$cblist if $cblist;
    }
    else {
        for my $ev (keys %{$engine->{_ev_SUBSCRIPTION}{$plugin}}) {
            my $cblist = delete $engine->{_ev_SUBSCRIPTION}{$plugin}{$ev};
            $engine->{_ev_EVENT}{$ev} -= @$cblist if $cblist;
        }
    }
}

sub listeners {
    my ($engine, $event) = @_;
    return $engine->{_ev_EVENT}{$event} || 0;
}

sub broadcast {
    my ($engine, $extra, $event) = splice(@_, 0, 3);
    return unless $engine->{_ev_EVENT}{$event};
    $extra ||= {}; # so we can report the EVENT
    local $extra->{EVENT} = $event;
    $engine->log("DEBUG", "broadcast: $event") if DEBUG;
    for my $plugin (@{$engine->{_ev_PA}}) {

        my $cblist = $engine->{_ev_SUBSCRIPTION}{$plugin}{$event} or next;
        for my $cb (@$cblist) {
            &$cb($engine, $extra, @_);
        }
    }
}

sub broadcastm {
    my ($engine, $extra, $event, $ARRAY) = @_;
    return unless $engine->{_ev_EVENT}{$event};
    $extra ||= {}; # so we can report the EVENT
    local $extra->{EVENT} = $event;
    $engine->log("DEBUG", "broadcastm: $event") if DEBUG;
    for my $plugin (keys %{$engine->{_ev_SUBSCRIPTION}}) {
        my $cblist = $engine->{_ev_SUBSCRIPTION}{$plugin}{$event} or next;
        for my $cb (@$cblist) {
            &$cb($engine, $extra, @$ARRAY);
        }
    }
}

1;

