package ActiveState::SpamEngine::Plugin::bayes;

use strict;
use constant ALGORITHM => "InverseChiSquare";
use constant VERSION   => 2;

use ActiveState::SpamEngine::Plugin::bayes::DB qw(spamprob addmsg recalc);
use ActiveState::SpamEngine qw(PMX_TYPE_INTEGER PMX_TYPE_HASH PMX_TYPE_STRING
                               _get1);
use base 'ActiveState::SpamEngine::Plugin';

#=============================================================================
# Support the SpamEngine plugin API
#=============================================================================
our @SP_REQUIRE = qw(part);
our @SP_ATTRIB  = (
    # KEY                   TYPE                DEFAULT
    "db",                   PMX_TYPE_HASH,      undef,
    "count",                PMX_TYPE_HASH,      undef,
    "datadir",              PMX_TYPE_STRING,    undef,
    "min-word-length",      PMX_TYPE_INTEGER,   3,
    "max-word-length",      PMX_TYPE_INTEGER,   12,
    "max-parsed-words",     PMX_TYPE_INTEGER,   50_000,
    "max-scoring-words",    PMX_TYPE_INTEGER,   20,
    "min-word-occurrences", PMX_TYPE_INTEGER,   5,
);

sub plugin {
    my ($eng, $name) = @_;

    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);

    # Add a callback to load the config file
    $eng->subscribe($name, 'core.compile', \&compile);
    $eng->subscribe($name, 'core.unload', \&update);

    1;
}

# special versioning fields that are saved in the weights file
use constant FLAG_VERSION     => "\1\1VERSION\1\1";
use constant FLAG_ALGORITHM   => "\1\1ALGORITHM\1\1";

# state keeping variables to control display of error messages.
my $Showed_errmsg;
my $update_needed;

sub get_datadir {
    my $eng = shift;
    my $datadir;
    my @attr = qw(core.userdatadir plugin.bayes.datadir);
    for my $attr (@attr) {
        $eng->enumerate_attributes($attr, \&_get1, \$datadir);
        return $datadir if defined $datadir;
    }
    $eng->log("ERR", "Can't find bayes database: [@attr] undefined");
    undef;
}

sub get_db {
    my ($eng, $readonly) = @_;
    my $db;
    $eng->enumerate_attributes('plugin.bayes.db', \&_get1, \$db);
    if (!$db) {
        my $datadir = get_datadir($eng) or return;
        $db = eval { $eng->create_hash("$datadir/bayes.dat", $readonly) };
        if ($db) {
            $eng->add_attribute('plugin.bayes.db', PMX_TYPE_HASH, $db, 1);
        }
        elsif ($readonly) {
            return undef;
        }
        else {
            $eng->log("ERR", "Unable to open or create $datadir/bayes.dat");
        }
    }
    return $db;
}

sub get_count {
    my ($eng, $readonly) = @_;
    my $count;
    $eng->enumerate_attributes('plugin.bayes.count', \&_get1, \$count);
    if (!$count) {
        my $datadir = get_datadir($eng) or return;
        my $filename = "$datadir/bayes.cnt";
        $count = eval { $eng->create_hash($filename, $readonly) };
        if ($count) {
            $eng->add_attribute('plugin.bayes.count', PMX_TYPE_HASH, $count, 1);
            return $count;
        }
        elsif ($readonly) {
            return undef;
        }

        if ($@) {
            # unlink the file and try again
            unlink $filename;
            $count = eval { $eng->create_hash($filename) };
        }
        if ($@) {
            $eng->log("ERR", "Unable to open or create $filename");
            return undef;
        }
    }
    return $count;
}

sub update {
    my $eng = shift;
    return unless $update_needed;

    my ($db, $count);
    $db = get_db($eng);
    $count = get_count($eng);
    unless ($count and $db) {
        $eng->log("ERR", "bayes::update: 'plugin.bayes.count' not set")
            unless $count;
        $eng->log("ERR", "bayes::update: 'plugin.bayes.db' not set")
            unless $db;
        return;
    }

    my $minoccur;
    $eng->enumerate_attributes('plugin.bayes.min-word-occurrences',
                               \&_get1, \$minoccur);

    my $ok = recalc($db, $count, $minoccur);

    if ( $ok ) {
        $db->set(FLAG_ALGORITHM, ALGORITHM);
        $db->set(FLAG_VERSION, VERSION);
        $db->save;
    }
}

sub train {
    my ($eng, $xtra, $msg) = @_;

    my $count = get_count($eng);
    unless ($count) {
        $eng->log("ERR", "bayes::train: 'plugin.bayes.count' not set");
        return;
    }

    my $isspam;
    $eng->enumerate_attributes('core.is-spam', \&_get1, \$isspam);

    _get_tokens($eng, $xtra, $msg);
    addmsg($count, $xtra->{A}, $isspam);

    $update_needed = 1;
}

sub db_has_data {
    my $db = shift;
    return unless $db;
    return $db->enumerate(sub {return 0;});
}

# Returns 1 if db is ok, 0 if not.
# Also prints error messages when first time called
sub check_db {
    my ($eng, $db) = @_;

    if (!db_has_data($db)) {
        $eng->log("ERROR", "The adaptive classifier was not trained.  Please train the adaptive classifier.")
            unless $Showed_errmsg;
        $Showed_errmsg = 1;
        return 0;
    }
    # post: there exists a bayes.dat file

    my $algorithm = $db->get(FLAG_ALGORITHM) || "NaiveBayes";
    my $version   = $db->get(FLAG_VERSION) || 1;

    if (!isvalid_algorithm_and_version( $algorithm, $version )) {
       $eng->log("ERROR",  "Previous training is not compatible with the current training algorithm. Please re-train the adaptive classifier.")
           unless $Showed_errmsg;
       $Showed_errmsg = 1;
       return 0;
    }

    return 1;
}

my $rules;
sub compile {
    my $eng = shift;
    $rules = $eng->db->get("bayes.rules");
    $eng->subscribe("bayes", 'core.meta_scan', \&scan);
    1;
}

sub scan {
    my ($eng, $xtra, $msg) = @_;

    my $train = 0;
    $eng->enumerate_attributes("core.train", \&_get1, \$train);
    return train($eng, $xtra, $msg) if $train;

    my ($db, $count, $maxtokens);
    $db = get_db($eng, 1);
    return unless check_db($eng, $db);

    $count = get_count($eng, 1);
    unless ($count) {
        $eng->log("ERR", "bayes::scan: 'plugin.bayes.count' not set");
        return;
    }

    $eng->enumerate_attributes('plugin.bayes.max-scoring-words',
                               \&_get1, \$maxtokens);
    _get_tokens($eng, $xtra, $msg);
    my $p = spamprob($db, $xtra->{A}, $maxtokens, $count);
    $eng->log("DEBUG", "bayesian probability: p=$p");

    # If we're in training mode, we won't have any rules defined because the
    # compiler won't have included them.
    return unless defined $rules;

    # Figure out which rule to fire based on the threshold. Only one
    # rule will fire.
    $p *= 100;
    for my $entry (split ' ', $rules) {
        my ($thresh, $rule) = split ':', $entry;
        if ($p >= $thresh) {
            $eng->addfeature($xtra, $rule, $p);
            last;
        }
    }
}

# Tokens. This is backwards compatible with the pre-focus bayes plugin, and
# will be left here for now.

my $TOKENS = qr/[^-\w'\x80-\x9f\xa1-\xff]+/;

sub _get_tokens {
    my ($eng, $xtra, $msg) = @_;

    $xtra->{A} = {};

    my ($minlen, $maxlen, $maxwords, $maxbytes);
    my $prefix = "plugin.bayes";
    $eng->enumerate_attributes("$prefix.min-word-length", \&_get1, \$minlen);
    $eng->enumerate_attributes("$prefix.max-word-length", \&_get1, \$maxlen);
    $eng->enumerate_attributes("$prefix.max-parsed-words", \&_get1, \$maxwords);
    $eng->enumerate_attributes("core.max-bytes-scanned", \&_get1, \$maxbytes);

    $minlen = 1 if $minlen < 1;

    # Each token in the subject should be prepended with "s:" so it can be
    # distinguished from tokens in the body.
    $msg->enumerate_headers("Subject", sub {
        my $data = shift;
        $data->{$_}++ for map { "s:$_" }
                          grep {
                              my $l = length($_);
                              $l >= $minlen && $l <= $maxlen && !skip_token($_)
                          }
                          split $TOKENS, $_[1];
        1;
    }, $xtra->{A});

    _tokenize_part(
        [$xtra->{A}, 1, $minlen, $maxlen, $maxbytes, $maxwords],
        $msg,
    );
}

sub _tokenize_part {
    my ($data, $part) = @_;
    my ($tokens, $first, $minlen, $maxlen, $maxbytes, $maxwords) = @$data;

    my $ctype = 'text/plain';
    $part->enumerate_headers("Content-Type", sub {
        return 1 unless $_[2] =~ /^\s*([^\s;]+)/;
        ${$_[0]} = lc $1;
        0;
    }, \$ctype);

    # Take the first `core.max-bytes-scanned` only. If it's larger than that,
    # most of the harmful content is probably at the top.
    if ($ctype =~ m#^text/#i or $first) {
        my $body = $part->get_decoded_body(0, $maxbytes) or return 1;
        $tokens->{$_}++ for grep {
                                my $l = length($_);
                                $l >= $minlen && $l <= $maxlen && !skip_token($_)
                            }
                            split $TOKENS, $body;
    }
    return 0 if keys(%$tokens) > $maxwords;

    $data->[1] = 0; # the next part is a subpart
    $part->enumerate_parts(\&_tokenize_part, $data);
    1;
}

## version control
sub isvalid_algorithm_and_version {  # returns true if ok, false otherwise
    my ($algorithm, $version) = @_;
    return $algorithm eq ALGORITHM and $version eq VERSION;
}
sub skip_token { # shortcoming of embedding extra info in the weights file
   return $_[0] eq FLAG_VERSION or $_[0] eq FLAG_ALGORITHM;
}
1;
