package ActiveState::SpamEngine::Plugin::bayes::DB;

use strict;
use base 'Exporter';

our @EXPORT_OK = qw(spamprob addmsg recalc);

#use constant S_CONSTANT => 0.373;
use constant S_CONSTANT => 0.00464;
use constant X_CONSTANT => 0.538;

use constant S_TIMES_X_CONSTANT => S_CONSTANT * X_CONSTANT;

#use constant MIN_DEVIATION => 0.05;
use constant MIN_DEVIATION => 0.44;
use constant MAX_DEVIATION => 0.5;



#=============================================================================
# The weights are stored in a simple pmx_weights table.
#=============================================================================

sub spamprob {
    my ($hash, $rh_tokens, $top_x) = @_;
    my $tokens = [keys %$rh_tokens];

    my $total_H = 0;
    my $total_S = 0;

    my @goodtokens = sort {
                my $pa = $hash->get($a) || 0.5;
                my $pb = $hash->get($b) || 0.5;

                abs($pb - 0.5) <=> abs($pa - 0.5);
            }
            @$tokens;

    # now, combine probabilities using Fisher method
    my $num_pvalues_used = 0;
    foreach my $token (@goodtokens) {
        my $pvalue = $hash->get($token);
        next unless $pvalue;
        # throw away non-significant features
        my $deviation = abs($pvalue - 0.5);
        next if $deviation < MIN_DEVIATION || $deviation > MAX_DEVIATION;

        $total_H += log $pvalue;
        $total_S += log (1.0 - $pvalue);
        $num_pvalues_used++;
        last if $num_pvalues_used > $top_x;
    }
   $total_H *= -2;
   $total_S *= -2;

   if ( $num_pvalues_used == 0 ) { # no observed token is was seen during training
        return 0.5;
   }

   # now, do the inverse chi-square test
   my $twice_degrees_of_freedom = 2 * $num_pvalues_used;
   my $S = 1 - chi2q($total_S, $twice_degrees_of_freedom);
   my $H = 1 - chi2q($total_H, $twice_degrees_of_freedom);

   my $prob = (($S - $H) + 1.0) / 2.0;
   $prob;
}

# inverse chi-squared function
# VERBATIM FROM SpamAssasin/Bayes.pm
sub chi2q {
  my ($x2, $v) = @_;

  die "v must be even in chi2q(x2, v)" if $v & 1;
  my $m = $x2 / 2.0;
  my ($sum, $term);
  $sum = $term = exp(0 - $m);
  for my $i (1 .. (($v/2)-1)) {
    $term *= $m / $i;
    $sum += $term;
  }
  return $sum < 1.0 ? $sum : 1.0;
}
# END OF SpamAssasin verbatim code

#=============================================================================
# The weights are calculated using a second "count" database, which tracks how
# many messages occurred in spam/ham, and how many times each token appeared.
#=============================================================================

sub IX_FREQ_S() { 0 }
sub IX_FREQ_H() { 1 }

sub _get {
    my ($hash, $key, $ix) = @_;
    my $val = $hash->get($key);
    $val = '0 0' unless defined $val;
    (split ' ', $val)[$ix];
}

sub _set {
    my ($hash, $key, $ix, $new) = @_;
    my $val = $hash->get($key);
    $val = '0 0' unless $val; # $initialized $val is a string 'x x' so it is never false
    my @val = split ' ', $val;
    $val[$ix] = $new;
    $hash->set($key, join(' ', @val));
}

# Use a reserved token for storing size of each corpus
sub K_CORPUS() { "\1\1CORPUS\1\1" }

sub addmsg {
    my ($h_count, $rh_tokens, $isspam) = @_;
    my $tokens = [keys %$rh_tokens];
    my $ix = $isspam ? IX_FREQ_S : IX_FREQ_H;

    # Update the count of each token
    for my $toke (@$tokens) {
        my $count = _get($h_count, $toke, $ix);
        $count++;
        _set($h_count, $toke, $ix, $count);
    }

    # Update the number of messages:
    my $count = _get($h_count, K_CORPUS, $ix);
    $count++;
    _set($h_count, K_CORPUS, $ix, $count);
}

sub ALPHA()  { 1 }      # smoothing factor

sub recalc {
    my ($h_weights, $h_count, $minoccur) = @_;

    # Flush uncommitted changes
    $h_count->save;

    # The base score starts at zero, and is updated for each token.
    my $base = 0;
    my $nspam = _get($h_count, K_CORPUS, IX_FREQ_S);
    my $nham = _get($h_count, K_CORPUS, IX_FREQ_H);
    if ( $nspam == 0 or $nham == 0 ) {
        warn qq|Before training, you must specify the training message set.\n|
            if ( $nspam == 0 or $nham == 0 );
        return;
    }

    # Delete all tokens which didn't occur at least
    $h_count->enumerate(sub {
        my ($data, $key, $val) = @_;
        return 1 if $key eq K_CORPUS; # skip reserved key

        my ($h_weights,  $minoccur, $nspam, $nham) = @$data;

        $val = '0 0' unless defined $val;
        my ($freq_s, $freq_h) = split ' ', $val;
        return 1 unless $freq_s + $freq_h >= $minoccur;

        my $pr_f_given_s = $freq_s / $nspam;
        my $pr_f_given_h = $freq_h / $nham;

        my $pr_s_given_f = $pr_f_given_s / ( $pr_f_given_s + $pr_f_given_h );

        # now, incorporate the beta-distrubution prior, and calculate
        # the expected value of the posterior distribution
        my $total = $nspam + $nham;
        my $expected_value =
          (S_TIMES_X_CONSTANT + $pr_s_given_f * $total )
          /
          (S_CONSTANT + $total);

        my $weight = $expected_value;
        $h_weights->set($key, $weight);

        1;
    }, [$h_weights,  $minoccur, $nspam, $nham]);

    $h_weights->save;
    1;
}

1;
