package ActiveState::SpamEngine::Plugin::dnsbl;

use strict;

use ActiveState::DNS::Request qw(:status);
use ActiveState::SpamEngine qw(PMX_TYPE_CALLBACK PMX_TYPE_STRING
                               PMX_TYPE_INTEGER
                               _get1 _getall);

use base qw(ActiveState::SpamEngine::Plugin);

#=============================================================================
# Support the SpamEngine plugin API
#=============================================================================
our @SP_REQUIRE = qw(net sxl);
our @SP_ATTRIB = (
    # KEY                           TYPE                DEFAULT
    "skip-rbl-checks",              PMX_TYPE_INTEGER,   undef,
    "relay-tests-only",             PMX_TYPE_INTEGER,   undef,
);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);
    $eng->subscribe($name, "core.compile", \&compile);
}

my %timeout;
my $rules;
my $HAS_RELAY;
my $RELAY_ONLY;
sub compile {
    my $eng = shift;

    for my $attr (qw(
        plugin.dnsbl.skip-rbl-checks
        plugin.net.skip-rbl-checks
        plugin.re.skip-rbl-checks
    )) {
        my $skip;
        $eng->enumerate_attributes($attr, \&_get1, \$skip);
        return if $skip;
    }

    # Initialise the timeouts
    $rules = eval { $eng->nested_db("dnsbl") } or return;
    %timeout = ();
    $rules->enumerate(sub {
        my ($timeout, undef, $def) = @_;
        my ($kind, $class, $zone) = split ' ', $def;
        if (($kind eq 'w' or $kind eq 'b') and defined $zone) {
            $timeout->{$zone}{count} = 0;
            $timeout->{$zone}{lasttry} = time;
        }
        1;
    }, \%timeout);

    # If somebody specified a trusted-relays list, enable the new behaviour.
    {
        my $tmp;
        $eng->enumerate_attributes("plugin.net.trusted-relays", \&_get1, \$tmp);
        if ($tmp) {
            $HAS_RELAY = 1;
        }
        $eng->enumerate_attributes("plugin.dnsbl.relay-tests-only",
                                   \&_get1,
                                   \$tmp);
        $RELAY_ONLY = $tmp;
    }

    $eng->subscribe("dnsbl", "net.rcvd", \&chain);
    $eng->subscribe("dnsbl", "net.finish", \&finish);
    $eng->initialised("dnsbl");
}

sub chain {
    my ($eng, $xtra, $rm, $chain) = @_;
    my $x = $xtra->{P}{dnsbl} = {};
    my $now = time;
    for my $rcvd (@$chain) {
        my ($bx, $hx, $ix, $prh, $i) = @$rcvd;
	next unless $prh;
	my $ip = $prh->{ip};
        next if $i->{internal} || $i->{trusted};
        my $yaler = join('.', reverse split /\./, $ip);
        $rules->enumerate(sub {
            my ($data, $rule, $def) = @_;
            my ($eng, $x, $rm, $yaler, $i, $now) = @$data;
            my ($kind, $class, $zone, $part) = split ' ', $def;
            return 1 if $kind ne 'b' and $kind ne 'w';

            if ($part eq 'relay') {
                return 1 unless $HAS_RELAY && $i->{relay};
            }
            elsif ($part eq 'external') {
                return 1 if $HAS_RELAY && $i->{relay};
                return 1 if $RELAY_ONLY && !$i->{relay};
            }
            if ($kind eq 'w') {
                return 1 unless $i->{white};
            }
            my $time_since_lasttry = ($now - $timeout{$zone}{lasttry}) || 1;
            my $sickness = $timeout{$zone}{count} / $time_since_lasttry;
            if ($sickness >= 1) {
                $eng->log("DEBUG", "dnsbl: skipping $zone because of timeouts");
                return 1;
            }
            my $lookup = "$yaler.$zone";
            $eng->log("DEBUG", "dnsbl: lookup $lookup");
            push @{$x->{requests}}, {
                id      => $rm->lookup($lookup),
                rule    => $rule,
                class   => $class,
                zone    => $zone,
            };
            return 1;
        }, [$eng, $x, $rm, $yaler, $i, $now]);
    }
}

sub finish {
    my ($eng, $xtra, $rm) = @_;
    my $x = $xtra->{P}{dnsbl} or return;
    my $now = time;
    my $name_results = {};
    my $class_results = {};
    my $timeout_results = {};
    my $error_results = {};
    for my $entry (@{$x->{requests}}) {
        my $request = $rm->request($entry->{id}) or next;
        my ($rule, $class, $zone) = @$entry{qw(rule class zone)};

        my $status = $request->status;
        my $domain = $request->requested_domain;

        # Handle timeouts
        $timeout{$zone}{lasttry} = $now;
        if ($status == REQUEST_TIMEOUT) {
            $eng->log("DEBUG", sprintf("dnsbl: [] from $domain "
                                     . "(timeout) times r=%.3fs",
                                     $request->elapsed_time));
            my $new = int((1 + $timeout{$zone}{count})*1.5);
            $new = 60*60*24 if $new > 60*60*24;
            $timeout{$zone}{count} = $new;
            ++$timeout_results->{$zone};
            next;
        }
        elsif ($status == REQUEST_ERROR) {
            my $error = $request->error;
            $error =~ s/\s+at .* line \d+\.\s*$//;
            $eng->log("DEBUG", sprintf("dnsbl: error (%s) from $domain "
                                     . "times r=%.3fs ",
                                     $error,
                                     $request->elapsed_time,
                                     ));
            my $new = int((1 + $timeout{$zone}{count})*1.5);
            $new = 60*60*24 if $new > 60*60*24;
            $timeout{$zone}{count} = 0;
            ++$error_results->{$zone};
            next;
        }
        $timeout{$zone}{count} = 0;

        my $ips = $request->returned_ips;
        next unless @$ips;
        $eng->log("DEBUG", do {
            local $" = ", ";
            sprintf("dnsbl: [@$ips] from $domain times: r=%.3fs",
                    $request->elapsed_time);
        });
        for my $ip (@$ips) {
            $name_results->{$rule} = $domain;
            $class_results->{$class}{$ip} = 1;
        }
    }

    for my $dnsbl_name (keys %$timeout_results) {
        # what rulename do we want to fire?
        # should these rules be declared in the .rules file?
        $eng->addfeature($xtra, '__' . $dnsbl_name . '_TIMEOUT',
            $timeout_results->{$dnsbl_name});
    }
    for my $dnsbl_name (keys %$error_results) {
        $eng->addfeature($xtra, '__' . $dnsbl_name . '_ERROR',
            $error_results->{$dnsbl_name});
    }

    # We can't cache this, since weights can change per-message.
    my $sorted = [];
    $rules->enumerate(sub { push @{$_[0]}, [$_[1], $_[2]]; 1 }, $sorted);
    sort_by_weight($xtra, $sorted);

    # Fire the most important rule within each class, for each "kind".
    my %seen;
    for my $entry (@$sorted) {
        my ($rule, $val) = @$entry;
        my ($kind, $class, $zone, $part) = split ' ', $val;
        next if $seen{$kind}{$class};
        if ($kind eq 'm') {
            if ($class_results->{$class}{$zone}) {
                $eng->addfeature($xtra, $rule);
                $seen{$kind}{$class} = 1;
            }
        }
        elsif ($kind eq 'b' or $kind eq 'w') {
            if ($name_results->{$rule}) {
                $eng->addfeature($xtra, $rule, $name_results->{$rule});
                $seen{$kind}{$class} = 1;
            }
        }
    }
}

sub sort_by_weight {
    my ($xtra, $ary) = @_;
    my @by_weight = map { $_->[1] }
                    sort { $b->[0] <=> $a->[0] }
                    map {
                        my $w = $xtra->{w}->get($_->[0]);
                        [$w ? abs($w) : 0, $_]
                    }
                    @$ary;
    @$ary = @by_weight;
}

1;
