package ActiveState::SpamEngine::Plugin::net;

use strict;

use Time::HiRes qw(gettimeofday tv_interval);

use ActiveState::DNS::RequestManager qw();
use ActiveState::DNS::Request qw(:status);
use ActiveState::DNS::Util qw(:types);
use ActiveState::SpamEngine qw(PMX_TYPE_CALLBACK PMX_TYPE_STRING
                               PMX_TYPE_INTEGER PMX_TYPE_HASH
                               _get1 _getall);

use Email::Received qw(parse_received);

use base 'ActiveState::SpamEngine::Plugin';

sub TMOUT() { 2 }
sub RETRY() { 2 }

#our $RCVD_ANTE_IP = qr/(?:[\[( \t]|\(ident_lookup\@)/;
#our $RCVD_POST_IP = qr/[\]) \t;]/;
our $IP_ADDR      = qr/(?:\d{1,3}\.){3}\d{1,3}/;
our $DNS_MANAGER;

our @SP_REQUIRE = qw(part);
our @SP_ATTRIB  = (
    # KEY                   TYPE                DEFAULT
    "internal-hosts",       PMX_TYPE_HASH,      undef,
    "trusted-relays",       PMX_TYPE_HASH,      undef,
    "skip-dns-checks",      PMX_TYPE_INTEGER,   0,
    "dns-timeout",          PMX_TYPE_INTEGER,   TMOUT,
    "dns-retry",            PMX_TYPE_INTEGER,   RETRY,
    "dns-servers",          PMX_TYPE_STRING,    undef,

    #**************************************************************************
    # These attributes are deprecated.
    #**************************************************************************
    "dns-retrans",                  PMX_TYPE_INTEGER,   undef,
);

sub plugin {
    my ($eng, $name) = @_;

    # Load the "part" plugin so we can subscribe to its events
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);

    # Subscribe to the appropriate events
    $eng->subscribe($name, 'core.compile', \&compile);
    $eng->subscribe($name, 'core.unload', \&unload);

    1;
}

#=============================================================================
# Checks
#=============================================================================

sub unload {
    $DNS_MANAGER = undef;
}

{
    my ($db_trusted, $db_internal);
    sub is_internal_host {
        # Omit the IANA non-routables
        return 1 if $_[0] =~ /^(?:10|127|192\.168|172\.(?:1[6789]|2[0-9]|3[01]))\./;
        return 0 unless $db_internal;
        return $db_internal->get($_[0]);
    }
    sub is_trusted_relay {
        return 0 unless $db_trusted;
        return $db_trusted->get($_[0])
    }

    sub compile {
        my $eng = shift;

        # If we've been told not to check dns, abort early.
        for my $attr (qw(
            core.local-tests-only
            plugin.net.skip-rbl-checks
            plugin.net.skip-dns-checks
            plugin.re.skip-rbl-checks
            plugin.re.skip-dns-checks
        )) {
            my $skip;
            $eng->enumerate_attributes($attr, \&_get1, \$skip);
            return if $skip;
        }

        $eng->enumerate_attributes("plugin.net.internal-hosts",
                                   \&_get1, \$db_internal);
        $eng->enumerate_attributes("plugin.net.trusted-relays",
                                   \&_get1, \$db_trusted);

        init_request_manager($eng) or return;

        # Register our callbacks
        $eng->subscribe("net", "part.eoh",     \&init);
        $eng->subscribe("net", "part.animate", \&anim);
        $eng->subscribe("net", "part.eom",     \&fini);
        $eng->initialised("net");
    }
}

sub init_request_manager {
    my $eng = shift;
    my ($timeout, $retry, $retrans, @ns);

    my $rm = ActiveState::DNS::RequestManager->new or return;

    $timeout = TMOUT;
    for my $attr (qw(plugin.net.dns-timeout plugin.re.dns-timeout)) {
        last if $eng->enumerate_attributes($attr, \&_get1, \$timeout);
    }

    $retry = RETRY;
    for my $attr (qw(plugin.net.dns-retry plugin.re.dns-retry)) {
        last if $eng->enumerate_attributes($attr, \&_get1, \$retry);
    }

    # retrans delay is emulated by extending the timeout.
    $retrans = 0;
    for my $attr (qw(plugin.net.dns-retrans plugin.re.dns-retrans)) {
        last if $eng->enumerate_attributes($attr, \&_get1, \$retrans);
    }
    $timeout += $retrans;

    $eng->enumerate_attributes('plugin.net.dns-servers', \&_getall, \@ns);
    @ns = ActiveState::DNS::Util::system_nameservers() unless @ns;

    # If we have nameservers, add them, then check that they "stick" (i.e.
    # they were valid IP addresses).
    if (@ns) {
        $rm->set_nameservers(@ns);
        @ns = @{$rm->nameservers};
    }

    # Now throw an error if no nameservers are defined
    if (@ns) {
        $eng->log('DEBUG', "net: nameservers=" . join(", ", @ns));
    }
    else {
        $eng->log("ERROR", "net: no nameservers, disabling net plugin.");
        return;
    }
    $rm->set_timeout($timeout);
    $rm->set_retry($retry);
    return $DNS_MANAGER = $rm;
}

sub init {
    my ($eng, $xtra, $msg) = @_;
    my $rm = $DNS_MANAGER or return;
    $eng->log('DEBUG', 'net: starting net checks');

    $rm->reset; # just in case fini() was never called

    my $message_opts;
    $eng->enumerate_attributes('core.message-opts', \&_get1, \$message_opts);

    my $net_plugin_notes = $xtra->{P}{net} ||= {};

    # Each element of the received @chain is an array that contains the
    # following items:
    #
    #   [$block, $header, $count, $ip, \%info]
    #
    # After the received chain has been populated, we post-process it and fill
    # in the extra information.
    my @chain;
    my $block = 0;
    my $trusted_block = 0;

    my $relay = $msg->get_relay;
    my $helo = $msg->get_helo;

    if ($relay) {
        while ($relay =~ /($IP_ADDR)/g) {
            push @chain, [0, 0, 0, { ip => $1, helo => $helo }, {}, undef ];
            $block = 1;
            $trusted_block = 1;
        }
    }

    $msg->enumerate_headers(undef,
        sub {
            my ($d, $h, $v) = @_;
            my ($rcvd, $st, $bx, $hx) = @$d;
            if (lc $h eq 'received') {
                if ($st == 0) {     # starting a received header block
                    $d->[1] = 1;    # state = inside received header block
                }
                # Extract sendmail-style Received headers
                my $ix = 0;
		#while ($v =~ /$RCVD_ANTE_IP($IP_ADDR)$RCVD_POST_IP/g) {
		#    next if $seen{$1}++; # don't report the same IP twice
		#    push @$rcvd, [$bx, $hx, $ix++, $1];
		#}
		my $parsed_received_header = parse_received($v);
		return 1 unless $parsed_received_header->{ip};
		if ($parsed_received_header) {
		    $ix += 1;
		    push @$rcvd, [$bx, $hx, $ix, $parsed_received_header, {}, $v];
		}
                ++$d->[3];          # increment the header number
            }
            elsif ($st == 1) {      # exiting a received header block
                $d->[1] = 0;        # state = outside received header block
                ++$d->[2];          # track block number
                $d->[3] = 0;        # reset the header index
            }
            return 1;
        },
        [\@chain, 0, $block, 0]
    );

    # Add metadata: whether the IP address is internal, trusted.
    my %s;
    for my $link (@chain) {
        my ($bx, $hx, $ix, $prh, $info, $header) = @$link;

        my $k = $info->{_k} = "$bx:$hx";
        $info->{trusted_block} = $bx <= $trusted_block;
        my $internal = $info->{internal} = is_internal_host($prh->{ip});
        my $trusted  = $info->{trusted}  = is_trusted_relay($prh->{ip});
        if ($internal || $trusted) {
            $info->{ext_hops} = 0;
            $s{count}{$k}{i}++;
        }
        else {
            $s{count}{$k}{x}++;
            $s{hop}++ unless $s{hhop}{$k}++;
            $info->{ext_hops} = $s{hop};
        }
    }

    # Now decide which IPs to label as relays.
    for my $link (@chain) {
        my ($bx, $hx, $ix, $prh, $info, $header) = @$link;
        my $k = $info->{_k};
        if ($s{count}{$k}{x}) {
            $s{relay}{$k}++;
            # keep going if there are any internals in this header
            last unless $s{count}{$k}{i};
        }
    }

    for my $link (@chain) {
        my ($bx, $hx, $ix, $prh, $info, $header) = @$link;
        my $k = $info->{_k};
        $info->{relay} = 1 if $s{relay}{$k} && $info->{ext_hops};
        $info->{white} = 1 if $info->{relay} && !$s{count}{$k}{i}
                                && $s{count}{$k}{x} == 1;
    }

    # If the message is not authenticated, check the relays.
    #
    if (!$message_opts || !$message_opts->get('authed')) {
        my $last_trusted = 1;
        my @rip; #second and subsequent untrusted relays

	for my $link (@chain) {
	    my ($bx, $hx, $ix, $prh, $info, $header) = @$link;
	    next if ($info->{trusted} || $info->{internal});
	    if ($last_trusted) {
		$info->{first_untrusted} = 1;
		$last_trusted = 0;
		if ($prh) {
		    for my $header_piece (qw(ip helo)) { #(sort keys %$prh) {
			$eng->broadcast(
			    $xtra,
			    "part.fur_$header_piece",
			    $prh->{$header_piece}
			);
		    }
		    if (exists $prh->{ip}) {
			$net_plugin_notes->{rdns_request_id} = $rm->reverse_lookup($prh->{ip});
			$xtra->{fur} = $prh->{ip};
		    }
		}
		if ($header) {
		    $eng->broadcast($xtra, "part.fur_header", $header);
		}
		$eng->broadcast($xtra, "net.fur", $prh, $header, $rm);
	    } elsif($prh && exists $prh->{ip}) {
		push @rip, $prh->{ip};
	    }
	}
	if(scalar @rip) {
	    $eng->broadcast($xtra, "net.rip", \@rip);
	}
    }

    # Log information about the received chain (presumably we'll only do this
    # at a very high log level, or when an attribute is enabled).
    for my $link (@chain) {
        my ($bx, $hx, $ix, $prh, $info, $header) = @$link;
        $eng->log(
            "DEBUG",
            "relay ip chain [$bx:$hx] $prh->{ip} " . join(', ',
                $info->{internal} ? "internal"                      : (),
                $info->{trusted}  ? "trusted"                       : (),
                $info->{ext_hops} ? "external $info->{ext_hops}"    : (),
                $info->{relay}    ? "relay"                         : (),
                $info->{white}    ? "white ok"                      : (),
            )
        );
    }
    
    # If not authenticated, expose the parsed received chain, complete with metadata
    #
    if (!$message_opts || !$message_opts->get('authed')) {
        $eng->broadcast($xtra, "net.rcvd", $rm, \@chain);
    }
}

sub anim {
    my ($eng, $xtra) = @_;
    my $rm = $DNS_MANAGER or return;
    eval { $rm->animate };
    if ($@) {
        $eng->log("ERROR", "net: DNS failed with error '$@'");
        $rm->reset;
    }
}

sub fini {
    my ($eng, $xtra) = @_;
    $eng->log('DEBUG', 'net: finish net checks');
    my $rm = $DNS_MANAGER or return;
    my $net_plugin_notes = $xtra->{P}{net} ||= {};

    $eng->broadcast($xtra, "net.dispatch");

    # Measure how much time elapses while waiting for DNS queries to be returned or to time out
    my $start_time = [gettimeofday];

    $rm->join;

    $eng->statistics()->add('timing', 'dns', tv_interval($start_time) * 1000);

    if (exists $net_plugin_notes->{rdns_request_id}) {
	my $request = $rm->request($net_plugin_notes->{rdns_request_id});
	if ($request) {
	    my $status = $request->status;
	    if (defined $status) {
		if (REQUEST_OK == $status) {
		    if (@{$request->returned_domains}) {
			for my $domain (@{$request->returned_domains}) {
			    $eng->broadcast($xtra, "part.fur_rdns", $domain);
			}
		    }
		}
		elsif (REQUEST_ERROR == $status) {
                    my $error = $request->error();
                    $eng->log('DEBUG', sprintf('net: error (%s) from %s', $error, $request->requested_domain));
                    if ($error =~ m/^RCODE_(\S+)/) {
                        my $rcode_error = $1;
                        $eng->broadcast($xtra, "part.fur_rdns", $rcode_error);
                    }
		}
		elsif (REQUEST_TIMEOUT == $status) {
		    $eng->addfeature($xtra, "__NET_FUR_RDNS_TIMEOUT");
		}
	    }
	}
    }
    $eng->log("DEBUG", "net: net checks completed");
    $eng->broadcast($xtra, "net.finish", $rm);
    $rm->reset;
}

1;
