package ActiveState::SpamEngine::Plugin::part;

use strict;
use ActiveState::MIME::Header qw(header_utf8);
use HTML::Parser;
use HTML::Tagset;
use HTML::Entities;
use ActiveState::SpamEngine qw(PMX_TYPE_INTEGER PMX_TYPE_DOUBLE _get1 _getall DEBUG);
use ActiveState::Transformer::Plugin::HTMLSimplify;
use ActiveState::Transformer::Plugin::GIFMetadata;
use ActiveState::Transformer::Plugin::PNGMetadata;
use ActiveState::Transformer::Plugin::JPEGMetadata;
use Sophos::Debug qw(DEBUG _debug);

use base 'ActiveState::SpamEngine::Plugin';


our @SP_ATTRIB = (
    # KEY               TYPE                DEFAULT
    'log-uris',         PMX_TYPE_INTEGER,   undef,

    # $VISIBILITY_THRESHOLD is the minimum colour difference the human eye can
    # see, approximated in several ways. I wrote a sample HTML page with some
    # text in various different background colours, finding the point at which
    # I could no longer distinguish them. Thinking of the two colours as RGB
    # vectors, we can calculate the distance between them as the length of the
    # difference.
    #
    # The magic vector (for my eyes) appears to be [12,12,12] => 20.8. If two
    # colours are closer than about 21 units, I see them as the same colour.
    #
    # That's actually not quite true: I can distinguish [9,9,9] near white.
    # Near black I can't distinguish as well: [16,16,16] is as close as two
    # colours can be before they blend.
    #
    # NOTE: since I'm now older than when I wrote that, I can't differentiate
    # colours as well anymore. So up goes the threshold -- I assume most of
    # our customers have worse eyesight than I do.
    'visibility-threshold', PMX_TYPE_DOUBLE, sqrt(3 * 20**2),
);

my @types = (
        ['part.type.pdf', 'pdf', 'application_acrobat application_pdf application_vnd.pdf application_x-pdf text_pdf text_x-pdf'],
        ['part.type.word', 'doc dot', 'application_msword application_doc appl_text application_vnd.msword application_vnd.ms-word application_winword application_word application_x-msw6 application_x-msword'],
        ['part.type.flash', 'swf swfl flv', 'application_x-shockwave-flash'],
        ['part.type.excel', 'xls xla xlc xlm xlt xlw', 'application_vnd.ms-excel application_msexcel application_x-msexcel application_x-ms-excel application_x-excel application_x-dos_ms_excel application_xls application_x-xls zz-application_zz-winassoc-xls'],
        ['part.type.powerpoint', 'pot pps ppt', 'application_vnd.ms-powerpoint application_mspowerpoint application_ms-powerpoint application_mspowerpnt application_vnd-mspowerpoint application_powerpoint application_x-powerpoint application_x-mspowerpoint'],
        ['part.type.outlook', 'msg', 'application_vns.ms-outlook application_msoutlook'],
        );

my (%extensions, %mimetypes);
for my $entry (@types) {
    my ($type, $extensions, $mimetypes) = @$entry;

    $extensions{$_} = $type foreach (split / /, $extensions);
    $mimetypes{$_} = $type foreach (split / /, $mimetypes);
}
@types = ();

sub plugin {
    my ($engine, $name) = @_;
    $engine->declare_attributes(\@SP_ATTRIB, $name);
    $engine->subscribe($name, 'core.compile', \&reset);
    1;
}

my $log_uris;
my $VISIBILITY_THRESHOLD;
# my $TLDDB;

sub reset {
    my $engine = shift;
    # $TLDDB = eval { $engine->nested_db("tld") };
    $engine->enumerate_attributes('plugin.part.log-uris', \&_get1, \$log_uris);
    $engine->enumerate_attributes('plugin.part.visibility-threshold',
                               \&_get1, \$VISIBILITY_THRESHOLD);
    $engine->subscribe('part', 'core.scan_message', \&part);
    $engine->initialised('part');
}

# normalizes the array by converting string elements to lowercase, eliminating
# duplicate elements and removing hyphens from each element
# XXX this should be moved off to AS-Transformer
sub normalize_charset_arr {
    my @new_ar = ();
    my %h = ();
    for my $temp (@{$_[0]}) {
        $temp =~ s/-//g;
        $temp = lc $temp;
        $h{$temp} = 0;
    }
    for my $temp (@{$_[0]}) {
        next if $h{$temp};
        push @new_ar, $temp;
        $h{$temp} = 1;
    }
    @{$_[0]} = @new_ar;
}


sub part {
    my ($engine, $extra, $msg) = @_;
    my $animate = $engine->listeners('part.animate');

    {
        my %event = (
            get_envelope_sender => 'e_from',
            get_helo            => 'helo',
        );

	my $message_opts;
	$engine->enumerate_attributes('core.message-opts', \&_get1, \$message_opts);
        
        if (DEBUG) {
            if ($message_opts) {
                _debug("core.message-opts{authed}=".($message_opts->get('authed') || "undef"));
            }
            else {
                _debug("core.message-opts is undefined");
            }
        }

	if (!$message_opts || !$message_opts->get('authed')) {
	    $event{get_relay} = 'relay';
	    $event{get_relay_hostname} = 'relay_hostname';
    	}

        for my $method (keys %event) {
            my $string = $msg->$method();
            $engine->broadcast($extra, "part.$event{$method}", $string)
                if defined $string;
        }
    }

    _debug("Enumerate enveloper recipeients") if DEBUG;
    my @recipients;
    $msg->enumerate_envelope_recipients(sub {
        my ($data, $recipient) = @_;
        push @$data, $recipient;
        1;
    }, \@recipients);
    $engine->broadcast($extra, 'part.e_to', @recipients) if @recipients;

    my @charset = ();

    my (%headers, %rheaders, @order);
    # Check the headers and meta tests
    {
        _debug("Enumerate headers") if DEBUG;
        $msg->enumerate_headers(undef, sub {
            my ($data, $header, $value) = @_;
            my $hlc = lc $header;
            my $newval = header_utf8($value);
            push @{$data->[0]{$hlc}}, $newval;
            push @{$data->[1]{$hlc}}, $value,
            push @{$data->[2]}, $hlc;
            1;
        }, [\%headers, \%rheaders, \@order]);

        my %done;
        $headers{__ORDER_DUPS__} = \@order;
        for my $h (@order) {
            next if $done{$h}++;
            push @{$headers{__ORDER__}}, $h;
            (my $ev = $h) =~ s/-/_/g;
            # unwrap header data before propagation (work on a copy)
            my @h = @{$headers{$h}};
            foreach (@h) {
                s/\A\r?\n[\t\s]*//;       # strip the initial line wrap
                s/[ \t]*\r?\n[\t\s]*/ /g; # wrap header
            }
            $engine->broadcast($extra, "part.h_$ev", @h);

            if ($h eq 'content-type') {
                for my $ct (@h) {
                    while ($ct =~ /
                        charset\s*=\s*
                        (                         # start capture
                            \"[^\"]+\"            # quoted string (no embedded quotes)
                        |                         # or
                            (?!\")                # next char can't be a quote (zero-width)
                            [^\s\;]+              # string terminated by whitespace or a ';'
                        )                         # end capture
                        /xg
                    ) {
                        my $temp = $1;
                        $temp =~ s/^\"(.*)\"$/$1/;
                        push @charset, $temp;
                    }
                }
            }

            # extract email addresses and generate :addr events for certain
            # headers
            if ($h =~ /^(from|to|cc|bcc|reply-to|return-path)$/i) {
                my $addr = [];
                foreach ( @h ) {
                    chomp;
                    s/\r?\n//g;
                    s/\s*\(.*?\)//g;
                    s/^[^<]*?<(.*?)>.*$/$1/;
                    s/ ;$//gs;
                    push @$addr, $_;
                }
                $engine->broadcastm($extra, "part.h_$ev:addr", $addr);
            }
        }
        $engine->broadcast($extra, 'part.hdr_hash', \%headers);
        $engine->broadcast($extra, 'part.hdr_rhash', \%rheaders);
        $engine->broadcast($extra, 'part.hdr_raw', $msg->get_header);

        # XXX See comments for part.eob and part.eom
        $engine->broadcast($extra, 'part.eoh', $msg);

        # include every charset mentioned in a raw header
        for my $va (values %rheaders) {
            for my $v (@$va) {
                while($v =~ /\=\?([^\?]+)\?([QqBb])\?([^\?]+)\?\=/g) {
                    push @charset, $1;
                }
            }
        }
    }

    my $info = parse_message($engine, $msg, $extra, \@charset);
    normalize_charset_arr(\@charset);

    # fire off the charset based header rules, note we only do this for "top"
    # level headers.
    for my $h (keys %rheaders) {
        my $va = $rheaders{$h};
        (my $ev = $h) =~ s/-/_/g;
        my @h_charset;
        for my $v (@$va) {
            while($v =~ /\=\?([^\?]+)\?([QqBb])\?([^\?]+)\?\=/g) {
                push @h_charset, $1;
            }
        }
        normalize_charset_arr(\@h_charset);
        for my $cs (@h_charset) {
            $engine->broadcast($extra, "part.h_$ev;charset=$cs", $headers{$h});
        }

        # if the header didn't contain any =????= blocks
        unless(@h_charset) {
            # and we have charsets declared elsewhere in the message
            if (@charset) {
                # try them all
                for my $cs (@charset) {
                    $engine->broadcast($extra, "part.h_$ev;charset=$cs", $headers{$h});
                }
            }
        }
    }

    $engine->broadcast($extra, "part.info", $msg);

    foreach my $info_key (keys %$info) {
        next unless $info_key =~ /^uri:(.*)$/;
        next unless $engine->listeners("part.$info_key");
        $info->{$info_key} = [] unless $info->{$info_key};
        $engine->log('DEBUG', "extracted $info_key: [@{$info->{$info_key}}]") if $log_uris;
        $engine->broadcastm($extra, "part.$info_key", $info->{$info_key});
    }

    $engine->log('DEBUG', "extracted uris: [@{$info->{uri}}]") if $log_uris;
    $engine->broadcastm($extra, 'part.uri', $info->{uri});

    # XXX Consider wrapping some of these with if ($eng->listeners($ev)) { }

    # [part name, info key]
    my @broadcasted_parts = (
        # event name                           ref to event content   don't broadcast if part is empty?
        #                                                             (default: 0 == always broadcast)

        [ 'part.b_paras'                    => \$info->{b_paras} ],
        ( map { ["part.b_paras;charset=$_"  => \$info->{b_paras} ] } @charset),
        [ 'part.b_paras'                    => \$info->{b_paras_invisible} ],
        [ 'part.b_paras_visible'            => \$info->{b_paras} ],
        [ 'part.b_paras_invisible'          => \$info->{b_paras_invisible} ],
        [ 'part.b_lines'                    => \$info->{b_lines} ],
        [ 'part.b_lines'                    => \$info->{b_lines_invisible} ],
        [ 'part.b_lines_visible'            => \$info->{b_lines} ],
        [ 'part.b_lines_invisible'          => \$info->{b_lines_invisible} ],
        [ 'part.b_joinlines'                => \$info->{b_joinlines} ],
        [ 'part.b_joinlines_visible'        => \$info->{b_joinlines} ],
        [ 'part.b_joinlines_invisible'      => \$info->{b_joinlines_invisible} ],
        [ 'part.b_raw'                      => \$info->{b_raw} ],
        [ 'part.b_joinraw'                  => \$info->{b_joinraw} ],
        [ 'part.b_html_simplified'          => \$info->{b_html_simplified}, 1 ],
        [ 'part.b_mime_tree'                => \$info->{b_mime_tree} ],
        [ 'part.b_mime_tree_ex'             => \$info->{b_mime_tree_ex} ],
        [ 'part.b_gif_metadata'             => \$info->{b_gif_metadata}, 1 ],
        [ 'part.b_png_metadata'             => \$info->{b_png_metadata}, 1 ],
        [ 'part.b_png_normalized'           => \$info->{b_png_normalized}, 1 ],
        [ 'part.b_jpeg_metadata'            => \$info->{b_jpeg_metadata}, 1 ],
        [ 'part.b_jpeg_normalized'           => \$info->{b_jpeg_normalized}, 1 ],
        [ 'part.b_html_simplified_and_gif_metadata'  => \$info->{b_html_simplified_and_gif_metadata}, 1 ],
        [ 'part.b_html_simplified_and_png_metadata'  => \$info->{b_html_simplified_and_png_metadata}, 1 ],
        [ 'part.b_html_simplified_and_jpeg_metadata' => \$info->{b_html_simplified_and_jpeg_metadata}, 1 ],
        [ 'part.b_pdf_raw'                  => \$info->{b_pdf_raw}, 1],
    # XXX we shouldn't expose 'eob' and 'eom' as part at all - they are just
    # internal engine events that other engine components (e.g. genotype) can
    # trigger off of - so there needn't be any arguments
        [ 'part.eob'                        => \$msg ],
        [ 'part.m_full'                     => \$info->{m_full} ],
        [ 'part.eom'                        => \$msg ],
    );

    foreach my $part_descriptor (@broadcasted_parts) {
        $engine->broadcast($extra, "part.animate") if $animate;
        my ($part_name, $part_ref, $dont_if_empty) = @$part_descriptor;

        if (ref $$part_ref eq 'ARRAY') {
            next if $dont_if_empty and not int(@$$part_ref);
            $engine->broadcastm($extra, $part_name, $$part_ref);
        }
        else {
            next if $dont_if_empty and not $$part_ref;
            $engine->broadcast($extra, $part_name, $$part_ref);
        }
    }
}

sub _process_image_metadata {
    my ($state) = @_;
    my $info = $state->{'info'};


    for my $image_info (
        ['ActiveState::Transformer::Plugin::GIFMetadata',  'gif'],
        ['ActiveState::Transformer::Plugin::PNGMetadata',  'png'],
        ['ActiveState::Transformer::Plugin::JPEGMetadata', 'jpeg']) {

        my ($plugin_name, $type) = @$image_info;

        my $content = $state->{"${type}_content"};
        next unless @{$content};

        my ($partname, $sigpart) = ("b_${type}_metadata", "b_${type}_normalized");
        my $transformer = ${plugin_name}->new();
        
        foreach my $image_details (@$content) {

            my $metadata = $transformer->transform(${$image_details->{data}}, $image_details->{length});
            next unless $metadata;
            my $normalized = $transformer->normalize($metadata) if $transformer->can('normalize'); 
            push @{$info->{$partname}}, $metadata;

            # HTML and image metadata combined parts
            if ($normalized) {
                push @{$info->{$sigpart}}, $normalized;
                if (int(@{$info->{b_html_simplified}})) {
                    $info->{"b_html_simplified_and_${type}_metadata"} =
                        join("\n", $info->{b_mime_tree}, @{$info->{b_html_simplified}}, @{$info->{$sigpart}});
                }
            # If no normalized part, just use the normal part (GIFs, for one)
            } elsif (int(@{$info->{b_html_simplified}}) and int(@{$info->{$partname}})) { 
                    $info->{"b_html_simplified_and_${type}_metadata"} =
                        join("\n", $info->{b_mime_tree}, @{$info->{b_html_simplified}}, @{$info->{$partname}});
            }
        }
    }
}

sub parse_message {
    my ($engine, $msg, $extra, $charset_array) = @_;

    my $max_bytes;
    $engine->enumerate_attributes('core.max-bytes-scanned', \&_get1, \$max_bytes);

    my $h_subj = ''; # default subject is empty
    $msg->enumerate_headers('Subject', \&_get1, \$h_subj);
    my $subj = 'Subject: ' . header_utf8($h_subj) . "\n\n";
    my $sraw = "Subject: $h_subj\n\n";

    my %info = (
        b_lines             => [],
        b_lines_invisible   => [],
        b_paras             => [],
        b_paras_invisible   => [],
        b_html_simplified   => [],
        b_raw               => [$h_subj ? $sraw : ("")],
        header_uri          => {},
        uris                => [],
        b_gif_metadata      => [],
        b_png_metadata      => [],
        b_jpeg_metadata     => [],
        b_png_normalized    => [],
        b_jpeg_normalized   => [],
        b_pdf_raw           => [],
    );

    my %state = (
        # These two are used to generate the current part id
        parent  => undef,
        part    => 1,

        seen    => 0,       # number of bytes scanned so far
        max     => $max_bytes,
        info    => \%info,
        html    => {
            bgcolor     => ['#ffffff'],
            bgtag       => [],
        },
        html_size => 0, # size of HTML parts after stripping
        html_seen => 0, # raw size of HTML parts
        uri_size  => 0,
        __htmlinvisible__ => [],
        uri     => [],
        map     => {},

        # used for generating more body parts (b_gif*, b_html*)
        gif_content => [],
        png_content => [],
        jpeg_content => [],
        html_content => [],
    );

    
    
    {
        my $h_from = [];
        $msg->enumerate_headers('From', \&_getall, $h_from);

        @$h_from = map header_utf8($_), @$h_from;
        my $e_from = $msg->get_envelope_sender;
        if ($e_from) {
            push @$h_from, $e_from;
        }
        for my $from (@$h_from) {
            while ($from =~ /([\w\.\-]+\@([\w\.\-]+))/g) {
                my ($mailto, $domain) = ($1, $2);
                push @{$state{info}{header_uri}{'uri:from'}}, lc("mailto:$mailto");
                push @{$state{info}{header_uri}{'uri:from'}}, lc("http://$domain");
            }
        }
    }

    # XXX this should be done much sooner, so we can broadcast these and then
    # get out of dodge if we hit the blacklist
    $msg->enumerate_headers(undef,
        sub {
            my ($state, $header, $value) = @_;
            return 1 if $header =~ /^x-/i;
            my $uri_header = 'uri:' . lc($header);
            $state->{info}{header_uri}{$uri_header} ||= [];
            
            _debug("scan header '%s' for uris in '%s'", $uri_header, $value);
            get_uris($engine, $state->{info}{header_uri}{$uri_header}, $value);
            return 1;
        },
        \%state
    );

    # XXX We should now be done with all header processing.  We could go ahead
    # and broadcast all header parts now, allowing us to potentially bail
    # early before parsing the message body.

    # Generate a map of the message. We can use this to scan parts in priority
    # order, rather than in MIME order -- although we don't really jump around
    _map_message([$state{map}, undef, 1, $charset_array], $msg);

    $state{parser} = HTML::Parser->new(
        api_version     => 3,
        handlers        => [(
            start       => [sub { h_tag($engine, $extra, \%state, @_) }, "tagname,attr,'+1',tokens"],
            end         => [sub { h_tag($engine, $extra, \%state, @_) }, "tagname,attr,'-1'"],
            text        => [sub { h_text($engine, $extra, \%state, @_) }, "dtext"],
            comment     => [sub { h_cmnt($engine, $extra, \%state, @_) }, "text"],
        )],
        marked_sections => 1,
    );

    # leave entities alone
    $state{parser}->attr_encoded(1);

    eval {
        # Parse the subject
        push_plain(\%state, $subj) if $h_subj;

        _parse_callback([\%state, $engine, $extra], $msg);
        for (@{$state{multipart_queue}}) {
            $state{multipart_partid} = $_;
            _parse_callback([\%state, $engine, $extra], $msg);
            last;
        }
    };
    $engine->log('WARNING', "error parsing: $@") if $@;

    # XXX note that this only partially works now
    meta_features($engine, $extra, \%state);

    # XXX now we cobble together all the information we gathered in myriad
    # ways as the various engine parts... most of the work has already been
    # done by now

    # Calculate the b_join* and m_full parts
    $info{b_joinlines_visible} = join('', @{$info{b_lines}});
    $info{b_joinlines_invisible} = join('', @{$info{b_lines_invisible}});
    $info{b_joinlines} = join("\n",
        @info{qw(b_joinlines_visible b_joinlines_invisible)});

    _debug("b_joinlines: %s", $info{b_joinlines});
    get_uris($engine, $state{info}{uris}, $info{b_joinlines});

    my $t = $info{b_raw};
    $info{b_joinraw} = join('', @$t);

    $info{m_full} = join('', $msg->get_header, @$t[1 .. $#$t]);

    $info{uri} = [];
    my %check = ();
    for my $uri (@{delete $info{uris}}) {
        push @{$info{uri}}, $uri unless $check{$uri};
        $check{$uri} = 1;
    }
    foreach my $uri_header (keys %{$info{header_uri}}) {
        %check = ();
        $info{$uri_header} = [];
        for my $uri (@{delete $info{header_uri}{$uri_header}})
        {
            push @{$info{$uri_header}}, $uri unless $check{$uri};
            $check{$uri} = 1;
        }
    }

    foreach my $partid (sort keys %{$state{map}}) {
        $info{b_mime_tree} .= join(' ', $partid, $state{map}->{$partid}{ct}) . "\n";
        $info{b_mime_tree_ex} .= join(' ', $partid, "$state{map}->{$partid}{ct},", $state{map}->{$partid}{bsize}) . "\n";
    }

    my $html_transformer = ActiveState::Transformer::Plugin::HTMLSimplify->new();
    for my $html_chunk (@{$state{html_content}}) {
        my $html_simplified = $html_transformer->transform($html_chunk);
        next unless $html_simplified;
        push @{$state{info}{b_html_simplified}}, $html_simplified;
    }

    _process_image_metadata(\%state);

    $info{map} = $state{map};
    %state = ();
    $info{__meta__} = $extra->{P}{part};

    return $extra->{P}{part} = \%info;
}

# XXX these are only for the spamassassin plugin - and uri_size is now broken
# since sparkle (equant rewrites of uri parsing)
sub meta_features {
    my ($engine, $extra, $state) = @_;
    $extra->{P}{part}{html_ratio} = do {
        my $hsz = $state->{html_size};
        my $rsz = $state->{html_seen};
        my $r   = 0;
        $r      = ($rsz - $hsz) / $rsz if $rsz;
        $r      = 0 if $r < 0;
        $r;
    };
    $extra->{P}{part}{uri_ratio} = $state->{seen} ? $state->{uri_size} / $state->{seen} : 0;
}

# XXX do we collapse duplicates in the charset array?
sub _map_message {
    my ($data, $part) = @_;
    my ($map, $parent, $thispart, $charset_array) = @$data;
    my $partid = defined $parent ? "$parent.$thispart" : $thispart;
    $part->enumerate_headers(
        undef,
        sub {
            my ($csa, $h, $v) = @_;
            # include every charset mentioned in a raw header
            while($v =~ /\=\?([^\?]+)\?[QqBb]\?[^\?]+\?\=/g) {
                push @$csa, $1;
            }
            return 1;
        },
        $charset_array
    );

    # Extract the content type first
    my $raw_content_type = 'text/plain; charset="US-ASCII"';
    $part->enumerate_headers('Content-Type', \&_get1, \$raw_content_type);
    my ($content_type, $params);
    if ($raw_content_type =~ qr{^\s*([^;\s]*/(?:plain|html))([^;\s]+.*)$}s) {
        # DEF15932 - MS email parser can concatenate content type and parameters
        ($content_type, $params) = ($1, $2);
    } elsif ($raw_content_type =~ qr{^\s*([^;\s]*)[;\s]*(.*)$}s) {
        ($content_type, $params) = ($1, $2);
    }

    # Parse the Content-Type parameters (pass-through RFC822 comments)
    $params = '' unless defined $params;
    my %params;
    while (length $params) {
        if ($params =~ s/^([A-Za-z-]+)="([^";]+)"//) {
            $params{lc $1} ||= $2;
        }
        elsif ($params =~ s/^([A-Za-z-]+)='([^';]+)'//) {
            $params{lc $1} ||= $2;
        }
        elsif ($params =~ s/^([A-Za-z-]+)=([^\s;]+)//) {
            $params{lc $1} ||= $2;
        }
        else {
            $params{$params} = '';
            last;
        }
        $params =~ s/^\s*//;
    }
    if (defined $params{charset}) {
        push @$charset_array, $params{charset};
    }
    $params{charset} = "US-ASCII" unless defined $params{charset};

    my $content_transfer_encoding = "8bit";
    $part->enumerate_headers("Content-Transfer-Encoding", \&_get1,
        \$content_transfer_encoding);

    my $content_disposition = "inline";
    $part->enumerate_headers("Content-Disposition", \&_get1, \$content_disposition);

    # Save pertinent information about this part
    $map->{$partid} = {
        ct => lc($content_type),
        ct_raw => $content_type,
        ct_params => \%params,
        te => $content_transfer_encoding,
        cd => $content_disposition,
        bsize => $part->get_body_size, # *raw* size, not decoded
    };

    # Recurse into subparts (record how many children there are)
    $map->{$partid}{kids} = $part->enumerate_parts(
        \&_map_message,
        # 'thispart' in the arrayref will be modified before each call by the
        # call before it
        [$map, $partid, 1, $charset_array]
    );

    # ++$thispart - for next sibling
    ++$data->[2];
    return 1;
}

sub _add_image_part {
    my ($part, $info, $state, $type) = @_;
    my $body = $part->get_decoded_body(0, 0);
    my $data = $state->{max} == 0 ? $body : substr($body, 0, $state->{max});

    push @{$state->{$type}}, {
        data => \$data,
        ct => $info->{ct_raw},
        'length' => length $body,
    };
}

sub _expose_image_from_base64_header  {
    my ($part, $info, $state) = @_;

    my $image_header = $part->get_body(0, 10);

    # GIF87a -> R0lGODdh  GIF89a -> R0lGODlh
    if ($image_header =~ /^R0lGOD[dl]h/) {
        _add_image_part($part, $info, $state, 'gif_content');
    }
    # PNG
    elsif ($image_header =~ /^iVBORw0K/)
    {
        _add_image_part($part, $info, $state, 'png_content');
    }
    # JFIF JPEG
    elsif ($image_header =~ m#^/9j/[247]#)
    {
        _add_image_part($part, $info, $state, 'jpeg_content');
    }
}

# XXX refactor: at startup, the various part implementations should register the content
# types that they are interested in

# The parser will descend into the following types.
# XXX this is no longer even accurate... refactor to be clear.
my %allowed_content_type = map { $_ => 1 } qw(
message/delivery-status
plain/text
text
text/plain
text/html
txt/plain
);

sub _parse_callback {
    my ($data, $part) = @_;
    my ($state, $engine, $extra) = @$data;

    # Info about this part
    my $partid = $state->{multipart_partid};
    unless($partid) {
        $partid = defined $state->{parent}
            ? "$state->{parent}.$state->{part}"
            : $state->{part};
    }
    my $info = $state->{map}{$partid};
    my $ctype = $info->{ct};

    if(!$state->{multipart_partid}) {
        if($ctype =~ /^multipart\//) {
            push @{$state->{multipart_queue}}, $partid;
            goto DONE;
        }
    }

    # Broadcast attachment events
    if ($partid ne '1') {
        local $extra->{P}{part}{partid} = $partid;
        $engine->broadcast($extra, 'part.a_hdr', $partid.' '.$part->get_header);
        $engine->broadcast($extra, 'part.a_msg', $part);
        # XXX fix: should check if animate events were requested
        $engine->broadcast($extra, 'part.animate');
    }

  PART:
    {
        my $isPDF = grep ($_ eq 'part.type.pdf', _get_generic_attachment_events($info, $ctype));

        $isPDF = 0 if exists($ENV{'SOPHOS_SCANPDF'}) && $ENV{'SOPHOS_SCANPDF'} eq '0';

        # XXX This CT check could be refactored into a hashref of subs to
        # invoke if a desired content type is found
        if (0
            # Scan recognised text parts
            || $allowed_content_type{$ctype}
            # Scan the preamble and epilog for multipart messages
            || $state->{multipart_partid}
            # We're going to treat PDF content as text.
            || $isPDF
        )
        {
            # Check for base64 encoding
            if (!$isPDF && $info->{te} eq 'base64') {
                my $cs = lc $info->{ct_params}{charset};
                if ($cs =~ /us-?ascii/ || $cs =~ /iso-8859/) {
                    $extra->{P}{part}{base64_enc_text} = $info->{te};
                }
            }

            next PART if $state->{seen} >= $state->{max};
            my $body = $part->get_decoded_body(0, $state->{max} - $state->{seen});
            next PART unless defined $body and length $body;

            if ($isPDF) {
                # This is a PDF attachment, either via the mime type, or the filename's extension.
                my $data = $part->get_decoded_body(0, 0);

                my $filtered = filter_pdf_expand(\$data);
                next PART unless $filtered;
                # lots of the really good stuff in PDF is at the end of the
                # file, so we don't really want to limit b_pdf_raw to the first
                # 10K.  Bear in mind that this is a new and very specific part,
                # so SPOPS will keep the potential size of this part in mind
                # when writing rules which use it.
                push @{$state->{info}{b_pdf_raw}}, $filtered;

                # Truncate the filtered text before parsing it for text,
                # because the current text parser is rather slow
                my $limit = ($state->{max} || 10000) * 5;
                substr($filtered, $limit) = '' if length($filtered) > $limit;
                my $text = filter_pdf_totext(\$filtered);
                next PART unless $text;
                $body = substr($text, 0, $state->{max} - $state->{seen});
            }

            $state->{seen} += length($body);

            push @{$state->{info}{b_raw}}, split /^/, $body;

            # Check for uuencoded attachments. Currently we'll just strip
            # these and continue. We do this by spooling through the decoded
            # body removing uuencoding, until the end of the body or until
            # we've seen $state->{max} non-uuencoded bytes.
            my $UU_BEGIN = qr/^begin[ \t]+[0-7]{3,4}[ \t]+\S/;
            if ($body =~ /$UU_BEGIN/m) {
                my @notuu;
                my $chunksz = 65536;
                my $in_uu = 0;
                my $offset = 0;
                my $left = $state->{max};
                while ($left > 0) {
                    $body = $part->get_decoded_body($offset, $chunksz)
                        or last;
                    $offset += length $body;
                    for (split /^/, $body) {
                        if ($in_uu && /^end[\r\n]*$/)   { $in_uu = 0; next }
                        elsif (/$UU_BEGIN/)             { $in_uu = 1; next }
                        unless ($in_uu) {
                            push @notuu, $_;
                            $left -= length;
                        }
                    }
                }
                $body = join("", @notuu);
            }

            # XXX boy I bet we could sure optimize the performance of this
            # code, and extract it on a just-in-time basis too, rather than
            # all up front...

            # test for html tags here
            my $test = substr($body, 0, 512);
            if ($extra->{P}{part}{base64_enc_text} && $test =~ /[\x80-\xff]/) {
                delete $extra->{P}{part}{base64_enc_text};
            }

            my $multipart_html_disguised_as_text = 0;
            if ($ctype !~ /html/ &&
                ($test =~ /<\/.*?>/i || $test =~ /<\s*?(?:img|a)\b/i))
            {
                $extra->{P}{part}{html_disguised_as_text} = 1;
                $multipart_html_disguised_as_text = 1
                    if $partid eq '1';
            }

            # XXX I'm doing this before the plain and invisible processing,
            # because those subs cruelly stomp on $body
            if ($ctype =~ /^text\/html/ or
                $extra->{P}{part}{html_disguised_as_text}) {
                push @{$state->{html_content}}, $body;
            }

            # Parse HTML ($state->{plain_text}), extract URIs
            if ($state->{multipart_partid} and 
                $partid eq '1' and
                (!$multipart_html_disguised_as_text)) {
                push_plain($state, $body);
            }
            elsif ($ctype =~ /html/ or $extra->{P}{part}{html_disguised_as_text}) {
                $state->{html_text} = [];
                $state->{html_text_invisible} = [];
                $state->{parser}->parse($body);
                $state->{parser}->eof;
                my $text = join('', @{$state->{html_text}});
                $state->{html_seen} += length $body;
                $state->{html_size} += length $text;
                push_plain($state, $text);
                push_invisible($state, join("\n", @{$state->{html_text_invisible}}))
                    if $state->{html_text_invisible};
                push_invisible($state, join("\n", @{$state->{html_cmnt}}))
                    if $state->{html_cmnt};
            }
            # If we see HTML entities in text/plain, keep the plain text but
            # expose the decoded text as "invisible".
            elsif ($test =~ /(?:&#\d+;){2}/) {
                push_plain($state, $body);
                push_invisible($state, decode_entities($body));
            }
            else {
                push_plain($state, $body);
            }
        }

        if ($ctype !~ /^text/ && !$isPDF) {
            next PART unless $info->{te} eq 'base64';

            _expose_image_from_base64_header($part, $info, $state);
        }

    }

    if (my @events = _get_generic_attachment_events($info, $ctype)) {
        my $decoded_body = $part->get_decoded_body(0, $state->{max});

        $engine->broadcast($extra, $_, $decoded_body) foreach @events;
    }

DONE:
    if(!$state->{multipart_partid}) {
        {
            local $state->{parent} = $partid;
            local $state->{part}   = 1;
            $part->enumerate_parts(\&_parse_callback, $data);
        }
        $state->{part}++;
    }
    1;
}

sub _get_attachment_filename_extensions {
    my ($info) = @_;
    my %extensions = ();

    if (defined $info->{cd}) {
        if ($info->{cd} =~ m/filename\s*=\s*"[^;]+\.([^";]+)"/) {
            $extensions{$1} = 1;
        }
        elsif ($info->{cd} =~ m/filename\s*=\s*'[^;]+\.([^';]+)'/) {
            $extensions{$1} = 1;
        }
        elsif ($info->{cd} =~ m/filename\s*=\s*[^\s;]+\.([^\s;]+)/) {
            $extensions{$1} = 1;
        }
    }

    if (defined $info->{ct_params} && defined $info->{ct_params}{name}) {
        my $name = $info->{ct_params}{name};
        if ($name =~ m/"[^;]+\.([^";]+)"/) {
            $extensions{$1} = 1;
        }
        elsif ($name =~ m/'[^;]+\.([^';]+)'/) {
            $extensions{$1} = 1;
        }
        elsif ($name =~ m/[^\s;]+\.([^\s;]+)/) {
            $extensions{$1} = 1;
        }
    }
    return keys %extensions;
}

sub _get_generic_attachment_events {
    my ($info, $ctype) = @_;

    my %events;

    $ctype =~ s{/}{_}g if $ctype;

    $events{_get_event_name_from_extension($_)}++ foreach _get_attachment_filename_extensions($info);
    $events{_get_event_name_from_mimetype($ctype)}++ if $ctype;

    return keys %events;
}

sub _get_event_name_from_extension {
    my ($ext) = @_;

    return $extensions{$ext} || "part.extension.$ext";
}

sub _get_event_name_from_mimetype {
    my ($mimetype) = @_;

    return $mimetypes{$mimetype} || "part.mimetype.$mimetype";
}

#=============================================================================
# This chunk takes decoded text chunks (either from the HTML decoder or
# directly from a message part) and extracts URIs, lines, and paragraphs.
#=============================================================================

# XXX evil bug! this sub shouldn't be stomping on its arguments, as other
# things want to make use of them farther down the chain...  this is, among
# other things, resulting in text/plain parts being stripped of trailing
# newlines
sub push_plain {
    my $state = shift;
    for (@_) {
        s/\r//g;                  # delete CR (makes \n\n hard to detect)
        push @{$state->{info}{b_lines}}, split /^/, $_;
        s/[ \t\xa0]+/ /g;         # horizontal white -> SP
        s/\n+ *\n+/\f/g;          # double NL -> FF
        s/[ \x0b\r\n]+/ /g;       # white -> SP
        tr/\f/\n/;                # FF -> NL
        push @{$state->{info}{b_paras}}, split /^/, $_;
    }
}

# XXX ditto about stomping on arguments
sub push_invisible {
    my $state = shift;
    for (@_) {
        s/\r//g;                  # delete CR (makes \n\n hard to detect)
        push @{$state->{info}{b_lines_invisible}}, split /^/, $_;
        s/[ \t\xa0]+/ /g;         # horizontal white -> SP
        s/\n+ *\n+/\f/g;          # double NL -> FF
        s/[ \x0b\r\n]+/ /g;       # white -> SP
        tr/\f/\n/;                # FF -> NL
        push @{$state->{info}{b_paras_invisible}}, split /^/, $_;
    }
}

#=============================================================================
# Extract URIs from plain text: construct get_uris sub.
#
# NOTE: This code is UGLY. it is written to be FAST, not readable.
#
#=============================================================================

my %URI_QUOTE_TABLE = (
    '>' => '<',
    '<' => '>',
    '[' => ']',
    '(' => ')',
    '{' => '}',
    '"' => '"',
    "'" => "'",
);

my $URI_START_QUOTE_RE = qr/[\>\<\[\(\{\"\']/;


# Note that this just *gets* the uris from the provided text - it is up to the
# caller to decide what to do with them
# XXX and holy shit, this should just be implemented as a 1:many transformer
# -- just need to pass in a log method
sub get_uris {
    my $engine = shift;
    my $result = shift;

    my @chunks = ();
    {
        for my $in (@_) {
            my $chunk = $in;
            $chunk =~ s/\s+/ /g;
            push @chunks, split(/ /, $chunk);

            # This is a special case because this pattern relies on seeing
            # spaces in the original text
            while ($chunk =~ / ((?=[fw])(?:w w w|f t p) (?:[\/:\.\w-] )+)/gi) {
                my $temp = $1;
                $temp =~ s/\s+//g;
                push @chunks, $temp;
            }
        }
    }
    _debug("Chunks %d", scalar @chunks) if DEBUG;


    # stupid perl - blocks need labels if we want to use 'next'
    DECODE_PASS:
    for (my $index = 0; $index <= $#chunks; $index += 1) {
        my $chunk = $chunks[$index];
        _debug("HTML/entity/QP decode chunks[%d] '%s'", $index, $chunk) if DEBUG;

        {
            my $chunk_copy = $chunk;
            if ($chunk_copy =~ s/%([0-9a-f]{2})/chr(hex($1))/ige) {
                push @chunks, $chunk_copy;
                next DECODE_PASS;
            }
        }

        {
            my $chunk_copy = $chunk;
            if ($chunk_copy =~ s/=([0-9a-f]{2})/chr(hex($1))/ige) {
                push @chunks, $chunk_copy;
                next DECODE_PASS;
            }
        }

        {
            my $chunk_copy = $chunk;
            $chunk_copy = decode_entities($chunk);
            if ($chunk_copy ne $chunk) {
                push @chunks, $chunk_copy;
                next DECODE_PASS;
            }
        }
    }

    # Now we have an initial list of chunks to parse through.  We do two things:
    # - try to transform each chunk into something that might be meaningful
    # - try to find something meaningful in each chunk
    #
    # As soon as we find *anything* meaningful in the chunk, add it to the uri
    # part, and add the surrounding text as two new chunks to work on
    #
    # We *will* eventually exhaust our chunk list; heat-death cases cannot
    # happen.

    # XXX todo: consider:
    # my $max_size = 2*int(@chunks); if (int(@chunks) > $max_size) { # crap }
    while (@chunks) {
        my $chunk = shift @chunks;

        _debug("scanning chunk='%s' for things that look like URIs", $chunk) if DEBUG;
 
        # This is an arbitrary value from the labs and there should be no big impact of changing it
        #  - Basically, this is saying that we don't care about uri paths longer than $max_post_size
        my $max_post_size = 255;
        if ($chunk =~ /^(.*?)($URI_START_QUOTE_RE)?(\w{3,9}:[\\\/]{2})(.{0,$max_post_size})(.*)$/o) {
            my ($pre, $squote, $schema, $post1, $post2) = ($1, $2, $3, $4, $5);
            $post2 ||= '';
            $schema =~ s/\\/\//g;

            
            _debug("Looks like a URI") if DEBUG;

            if ($squote) {
                my $equote = $URI_QUOTE_TABLE{$squote};
                my $pos = rindex($post1, $equote);
                if ($pos>0) {
                    my $temp = $schema . substr($post1, 0, $pos);
                    _debug("Saving uri $temp (". __LINE__. ")") if DEBUG;
                    push @$result, $temp;
                }
                else {
                    _debug("Adding quote='%s' back onto pre='%s' ", $squote, $pre) if DEBUG;
                    $pre .= $squote;
                }
            } else {
                _debug("Saving uri $schema$post1 (". __LINE__. ")") if DEBUG;
                push @$result, "$schema$post1";
            }
            if ($pre) {
                _debug("Adding new chunk pre='%s' ", $pre) if DEBUG;
                push @chunks, $pre;
            }
            if ($post1) {
                _debug("Adding new chunk post='%s' ", $post1 . $post2) if DEBUG;
                push @chunks, $post1 . $post2;
            }
            next;
        }

        if ($chunk =~ /^(.*?)($URI_START_QUOTE_RE)?((?=[mj])(?:mailto|javascript):)(.{0,$max_post_size})(.*)$/io) {
            my ($pre, $squote, $schema, $post1, $post2) = ($1, $2, $3, $4, $5);
            $post2 ||= '';

            _debug("Looks like a mailto|javascript") if DEBUG;
            if ($squote) {
                my $equote = $URI_QUOTE_TABLE{$squote};
                my $pos = rindex($post1, $equote);
                if ($pos>0) {
                    my $temp = $schema . substr($post1, 0, $pos);
                    push @$result, $temp;
                }
                else {
                    $pre .= $squote;
                }
            } else {
                _debug("Saving uri $schema$post1 (". __LINE__. ")") if DEBUG;
                push @$result, "$schema$post1";
            }
            push @chunks, $pre if $pre;
            push @chunks, $post1 . $post2 if $post1;
            next;
        }

        #    something-like.this@some.domain.with.less-than.nine.elements
        if ($chunk =~ /^(.*?)\b(\w[\w\-.]{0,64}\@((?:[\w-]{2,63}\.){1,}\w{2,10}))\b(.{0,$max_post_size})(.*)$/o) {
            my ($pre, $email, $host, $post1, $post2) = ($1, $2, $3, $4, $5);
            $post2 ||= '';

            _debug("Looks like a something-like.this\@some.domain.with.less-than.nine.elements") if DEBUG;
            if (length($host) < 255) {
                push @$result, "mailto:$email";
                push @$result, "mailto:$email$post1";
                push @$result, $host;
                _debug("Saving uri mailto:$email (". __LINE__. ")") if DEBUG;
                _debug("Saving uri mailto:$email$post1 (". __LINE__. ")") if DEBUG;
                push @chunks, $pre if $pre;
                push @chunks, $post1 . $post2 if $post1;
            }
            next;
        }
        #   some.domain.with.less-than.nine.elements
        if ($chunk =~ /^(.*?)\b((?:[\w-]{2,63}\.){2,}\w{2,10})\b(.{0,$max_post_size})(.*)$/o) {
            my ($pre, $domain, $post1, $post2) = ($1, $2, $3, $4);
            $post2 ||= '';

            _debug("Looks like a some.domain.with.less-than.nine.elements") if DEBUG;
            _debug("domain length=%d", length($domain)) if DEBUG;
            if (length($domain) < 255) {
                push @$result, $domain;
                push @$result, "$domain$post1";
                _debug("Saving uri $domain (". __LINE__. ")") if DEBUG;
                _debug("Saving uri $domain$post1 (". __LINE__. ")") if DEBUG;
                push @chunks, $pre if $pre;
                push @chunks, $post1 . $post2 if $post1;
            }
            next;
        }

        #   whatever.cn (any.whatever.cn is caught by the previous capture)
        if ($chunk =~ /^(.*?)\b([\w-]{2,63}\.(?=[cnoibt])(?:com|cn|net|org|info|biz|tw))\b(.{0,$max_post_size})(.*)$/io) {
            my ($pre, $domain, $post1, $post2) = ($1, $2, $3, $4);
            $post2 ||= '';

            _debug("Looks like a whatever.cn (any.whatever.cn is caught by the previous capture)") if DEBUG;
            push @$result, $domain;
            push @$result, "$domain$post1";
            _debug("Saving uri $domain (". __LINE__. ")") if DEBUG;
            _debug("Saving uri $domain$post1 (". __LINE__. ")") if DEBUG;
            push @chunks, $pre if $pre;
            push @chunks, $post1 . $post2 if $post1;
            next;
        }
    }
}


#=============================================================================
# The HTML parsing logic is responsible for decoding as much as possible, and
# detecting HTML obfuscation.
#=============================================================================

my %tags_to_check_for_style_colour = map {$_=>1} qw(div span p small);

# Hash of html tags that are especially likely to contain uris, and the
# attribute where they can be found -- anything in these attributes will
# automatically be added to the uri part (even if they don't match what we
# would normally call a uri) -- and moreover we don't do any further searching
# inside html tags other than in the attributes we list here.
my %uri_tags = (
    ( map {$_ => 'href' }       qw(a area link) ),
    ( map {$_ => 'src' }        qw(img frame iframe embed script) ),
    ( map {$_ => 'background' } qw(body table tr td) ),
    ( map {$_ => 'action' }     qw(form) ),
    ( map {$_ => 'codebase' }   qw(object) ),
);

### HTML::Parser handlers (only)
# args:
# - engine
# - $x (from outside package)
# - state obj ref
# - 'start' or 'end' handler's arg, as specified in the object declaration:
#   - tagname
#   - hashref of all decoded attributes in this tag (undef if 'end' event)
#   - +1 or -1 signaling start or end of this tag
#   - arrayref of tokens: tag name, and if 'start' event, followed by attr name/value pairs, not decoded
sub h_tag {
    my ($engine, $extra, $state, $tag, $attr_encoded, $num, $tokens) = @_;
    my $attr = {};
    for my $key (keys %$attr_encoded) {
        $attr->{$key} = decode_entities($attr_encoded->{$key});
    }

    $engine->broadcast($extra, 'part.html_tag', $tag, $attr, $num);

    for my $key (keys %$attr) {
        $engine->broadcast($extra, "part.html_attr.$tag", "$key=$attr->{$key}");
    }

    unless ($HTML::Tagset::isKnown{$tag}) {
        # If it was an unknown HTML tag, it may be caused by our heuristic
        # guessing wrong about whether this part is HTML.
        $extra->{P}{part}{html_tag_unknown} ||= $tag;
        return;
    }

    if ($uri_tags{$tag} && exists $attr->{$uri_tags{$tag}}) {
        my $uri = $attr_encoded->{$uri_tags{$tag}};
            push @{$state->{info}{uris}}, $uri;
            $uri =~ s/\s//g;
            push @{$state->{info}{uris}}, $uri;

        $uri = $attr->{$uri_tags{$tag}};
            push @{$state->{info}{uris}}, $uri;

            $uri =~ s/\s//g;
            push @{$state->{info}{uris}}, $uri;
            _debug("Extract URIs from tag %s", $tag) if DEBUG;
            get_uris($engine, $state->{info}{uris}, $uri);
    }

    if (exists $attr->{style}) {
        my @uris = $attr->{style} =~ m/:\s*url\s*(?:\(['"]?|['"])(.*?)['")]/g;
        push @{$state->{info}{uris}}, @uris;
    }

    $state->{html_inside}{$tag} += $num;
    check_for_invisible_colour($state, $extra, $attr, $num) if $tag eq 'font';

    check_for_invisible_colour_in_style($state, $extra, $attr, $num) if
        $tags_to_check_for_style_colour{$tag};

    check_for_span_display_none($state, $extra, $attr, $num) if $tag eq 'span';

    if ($num == 1) {
        # Add whitespace for certain tags
        if ($tag eq 'p' or $tag eq 'hr') {
            push @{$state->{html_text}}, "\n\n";
        }
        elsif ($tag eq 'br') {
            push @{$state->{html_text}}, "\n";
        }
        elsif ($attr->{width}) {
            push @{$state->{html_text}}, " ";
        }

        if ($attr->{alt}) {
            push @{$state->{html_text_invisible}}, "\n\n", $attr->{alt}, "\n\n";
        }

        # update the background colour for invisibility checks
        if (exists $attr->{bgcolor}) {
            push @{$state->{html}{bgcolor}}, name_to_rgb(lc $attr->{bgcolor});
            push @{$state->{html}{bgtag}}, $tag;
        }

        $state->{html_last_tag} = $tag;

        # Look for the base URI; a base URI will be ignored by browsers unless
        # it is an absolute URI of a standard protocol.

        if ($tag eq 'base' and $attr->{href}) {
            # Cache this between invocations
            # XXX handling 'base' tags isn't required (yet)
        }
    }
    else {
        if ($tag eq 'td') {
            push @{$state->{html_text}}, " ";
        }
        elsif ($tag eq 'tr') {
            push @{$state->{html_text}}, "\n\n";
        }
        if (@{$state->{html}{bgtag}} and $tag eq $state->{html}{bgtag}[-1]) {
            pop @{$state->{html}{bgcolor}};
            pop @{$state->{html}{bgtag}};
        }
    }
}

sub h_text {
    my ($engine, $extra, $state, undef) = @_;
    my $store_in = 'html_text';
    # Skip <SCRIPT> text, CSS <style>, and invisible text
    for my $in (qw(script style invisible)) {
        if ($state->{html_inside}{$in}) {
            $store_in = 'html_text_invisible';
            last;
        }
    }

    push @{$state->{$store_in}}, $_[3];
    # Use a temporary reference in an attempt to keep Devel::Cover happy
    my $content = \$state->{$store_in}[-1];
    $$content =~ s/[\r\n]+/ /g;
}


sub h_cmnt {
    my ($engine, $extra, $state, undef) = @_;
    $engine->broadcast($extra, 'part.html_cmnt', $_[3]);
}

#=============================================================================
# Colour parsing and visibility prediction
#=============================================================================

sub push_block_visibility {
    my ($state, $v) = @_;
    push @{$state->{__htmlinvisible__}}, $state->{html_inside}{invisible};
    $state->{html_inside}{invisible} = $v;
}

sub pop_block_visibility {
    my ($state) = @_;
    if (int @{$state->{__htmlinvisible__}}) {
        $state->{html_inside}{invisible} = pop @{$state->{__htmlinvisible__}};
    }
    else {
        $state->{html_inside}{invisible} = 0;
    }
}

sub extract_colour_from_style {
    my $style = shift;
    my $C;
    if ($style) {
        if ($style =~ /colou?r\s*:\s*rgb\s*\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)/i) {
            $C = sprintf("%02X%02X%02X",$1,$2,$3);
        }
        elsif ($style =~ /colou?r\s*:\s*#?([0-9A-F]{6})/i) {
            $C = $1;
        }
    }
    return $C;
}

sub check_for_invisible_colour_in_style {
    my ($state, $extra, $attr, $num) = @_;
    if ($num == 1) {
        my $C = extract_colour_from_style($attr->{style});
        my $inv = 0;

        if (defined $C) {
            my $bg = lc $state->{html}{bgcolor}[-1];
            $C = name_to_rgb(lc $C);
            if (colour_distance($bg, $C) < $VISIBILITY_THRESHOLD) {
                $extra->{P}{part}{html_font_invisible} ||= $C;
                $inv = 1;
            }
        }
        push_block_visibility($state, $inv);
    }
    else {
        pop_block_visibility($state);
    }
}
sub check_for_invisible_colour {
    my ($state, $extra, $attr, $num) = @_;
    if ($num == 1) {
        my $C = $attr->{color};
        if ($attr->{style}) {
            $C = extract_colour_from_style($attr->{style});
        }
        my $inv = 0;
        if (defined $C) {
            my $bg = lc $state->{html}{bgcolor}[-1];
            $C = name_to_rgb(lc $C);
            if (colour_distance($bg, $C) < $VISIBILITY_THRESHOLD) {
                $extra->{P}{part}{html_font_invisible} ||= $C;
                $inv = 1;
            }
        }
        push_block_visibility($state, $inv);
    }
    else {
        pop_block_visibility($state);
    }
}

sub check_for_span_display_none {
    my ($state, $extra, $attr, $num) = @_;

    if ($num == 1) { # opening tag
        my $style = $attr->{style};
        if($style && $style =~ /display\s*:\s*none/i) {
            $extra->{P}{part}{html_font_invisible} ||= 'display:none';
            push_block_visibility($state, 1);
        } else {
            push_block_visibility($state, 0);
        }
    }
    else { # closing tag
        pop_block_visibility($state);
    }
}

sub colour_distance {
    my ($bg, $fg) = @_;
    for ($bg, $fg) {
        return $VISIBILITY_THRESHOLD unless defined and /^(#?)[0-9A-Fa-f]{6}$/;
        substr($_, 0, 1, '') if $1;
    }
    return 0 if $bg eq $fg;

    # Turn the string 'ffffff' into [255, 255, 255]
    my @bg = unpack('C*', pack('H*', $bg));
    my @fg = unpack('C*', pack('H*', $fg));
    my $d = 0;
    for my $i (0 .. $#bg) {
        my $di = $bg[$i] - $fg[$i];
        $d += $di ** 2;
    }
    my $dist = sqrt($d);
#warn "colour distance: bg=$bg fg=$fg dist=$dist min=$VISIBILITY_THRESHOLD\n";
    return $dist;
}

my %HTML_COLOUR = (
    aliceblue                => '#f0f8ff',
    antiquewhite             => '#faebd7',
    aqua                     => '#00ffff',
    aquamarine               => '#7fffd4',
    azure                    => '#f0ffff',
    beige                    => '#f5f5dc',
    bisque                   => '#ffe4c4',
    black                    => '#000000',
    blanchedalmond           => '#ffebcd',
    blue                     => '#0000ff',
    blueviolet               => '#8a2be2',
    brown                    => '#a52a2a',
    burlywood                => '#deb887',
    cadetblue                => '#5f9ea0',
    chartreuse               => '#7fff00',
    chocolate                => '#d2691e',
    coral                    => '#ff7f50',
    cornflowerblue           => '#6495ed',
    cornsilk                 => '#fff8dc',
    crimson                  => '#dc143c',
    cyan                     => '#00ffff',
    darkblue                 => '#00008b',
    darkcyan                 => '#008b8b',
    darkgoldenrod            => '#b8860b',
    darkgray                 => '#a9a9a9',
    darkgreen                => '#006400',
    darkkhaki                => '#bdb76b',
    darkmagenta              => '#8b008b',
    darkolivegreen           => '#556b2f',
    darkorange               => '#ff8c00',
    darkorchid               => '#9932cc',
    darkred                  => '#8b0000',
    darksalmon               => '#e9967a',
    darkseagreen             => '#8fbc8f',
    darkslateblue            => '#483d8b',
    darkslategray            => '#2f4f4f',
    darkturquoise            => '#00ced1',
    darkviolet               => '#9400d3',
    deeppink                 => '#ff1493',
    deepskyblue              => '#00bfff',
    dimgray                  => '#696969',
    dodgerblue               => '#1e90ff',
    feldspar                 => '#d19275',
    firebrick                => '#b22222',
    floralwhite              => '#fffaf0',
    forestgreen              => '#228b22',
    fuchsia                  => '#ff00ff',
    gainsboro                => '#dcdcdc',
    ghostwhite               => '#f8f8ff',
    gold                     => '#ffd700',
    goldenrod                => '#daa520',
    gray                     => '#808080',
    green                    => '#008000',
    greenyellow              => '#adff2f',
    honeydew                 => '#f0fff0',
    hotpink                  => '#ff69b4',
    indianred                => '#cd5c5c',
    indigo                   => '#4b0082',
    ivory                    => '#fffff0',
    khaki                    => '#f0e68c',
    lavender                 => '#e6e6fa',
    lavenderblush            => '#fff0f5',
    lawngreen                => '#7cfc00',
    lemonchiffon             => '#fffacd',
    lightblue                => '#add8e6',
    lightcoral               => '#f08080',
    lightcyan                => '#e0ffff',
    lightgoldenrodyellow     => '#fafad2',
    lightgrey                => '#d3d3d3',
    lightgreen               => '#90ee90',
    lightpink                => '#ffb6c1',
    lightsalmon              => '#ffa07a',
    lightseagreen            => '#20b2aa',
    lightskyblue             => '#87cefa',
    lightslateblue           => '#8470ff',
    lightslategray           => '#778899',
    lightsteelblue           => '#b0c4de',
    lightyellow              => '#ffffe0',
    lime                     => '#00ff00',
    limegreen                => '#32cd32',
    linen                    => '#faf0e6',
    magenta                  => '#ff00ff',
    maroon                   => '#800000',
    mediumaquamarine         => '#66cdaa',
    mediumblue               => '#0000cd',
    mediumorchid             => '#ba55d3',
    mediumpurple             => '#9370d8',
    mediumseagreen           => '#3cb371',
    mediumslateblue          => '#7b68ee',
    mediumspringgreen        => '#00fa9a',
    mediumturquoise          => '#48d1cc',
    mediumvioletred          => '#c71585',
    midnightblue             => '#191970',
    mintcream                => '#f5fffa',
    mistyrose                => '#ffe4e1',
    moccasin                 => '#ffe4b5',
    navajowhite              => '#ffdead',
    navy                     => '#000080',
    oldlace                  => '#fdf5e6',
    olive                    => '#808000',
    olivedrab                => '#6b8e23',
    orange                   => '#ffa500',
    orangered                => '#ff4500',
    orchid                   => '#da70d6',
    palegoldenrod            => '#eee8aa',
    palegreen                => '#98fb98',
    paleturquoise            => '#afeeee',
    palevioletred            => '#d87093',
    papayawhip               => '#ffefd5',
    peachpuff                => '#ffdab9',
    peru                     => '#cd853f',
    pink                     => '#ffc0cb',
    plum                     => '#dda0dd',
    powderblue               => '#b0e0e6',
    purple                   => '#800080',
    red                      => '#ff0000',
    rosybrown                => '#bc8f8f',
    royalblue                => '#4169e1',
    saddlebrown              => '#8b4513',
    salmon                   => '#fa8072',
    sandybrown               => '#f4a460',
    seagreen                 => '#2e8b57',
    seashell                 => '#fff5ee',
    sienna                   => '#a0522d',
    silver                   => '#c0c0c0',
    skyblue                  => '#87ceeb',
    slateblue                => '#6a5acd',
    slategray                => '#708090',
    snow                     => '#fffafa',
    springgreen              => '#00ff7f',
    steelblue                => '#4682b4',
    tan                      => '#d2b48c',
    teal                     => '#008080',
    thistle                  => '#d8bfd8',
    tomato                   => '#ff6347',
    turquoise                => '#40e0d0',
    violet                   => '#ee82ee',
    violetred                => '#d02090',
    wheat                    => '#f5deb3',
    white                    => '#ffffff',
    whitesmoke               => '#f5f5f5',
    yellow                   => '#ffff00',
    yellowgreen              => '#9acd32',
);

# XXX this could also be an AS-Transformer
sub name_to_rgb { $HTML_COLOUR{lc $_[0]} || $_[0] }

sub flate_decode {
    my ($prefix, $compressed, $postfix) = @_;
    #print STDERR "DEBUG: flate_decode(prefix='$prefix', compressed data is " . length($compressed) . " bytes, postfix='$postfix')\n";

    use Compress::Zlib;
    my $inflated = '';
    my $error = '';
    my $inflate = Compress::Zlib::inflateInit();
    if ($inflate) {
        my ($buff, $stat, $tmp);
        # hardcoded deadline of ten real-time seconds to complete decryption
        # Always decrypt first 1KB, even if it takes forever...
        my $deadline = time + 10;
        while ($compressed) {

            my $section = substr($compressed, 0, 1024);
            ($buff, $stat) = $inflate->inflate($section);
            substr($compressed, 0, 1024) = "";

            if ($stat == Compress::Zlib::Z_STREAM_END() || $stat == Compress::Zlib::Z_OK()) {
                # Only want ASCII character data
                $buff =~ tr[\200-\377][]d;    # strip top-bit set characters
                $buff =~ tr[\000-\011][]d;    # strip control characters
                $buff =~ tr[\013-\037][]d;    #  except for CR
                $inflated .= $buff;

                last if $stat == Compress::Zlib::Z_STREAM_END();
            } else {
                $error = 'CORRUPT';
                last;
            }

            if ($buff eq "") {
                $error = 'EMPTY';
                last;
            }
            if (time > $deadline) {
                $error = 'TIMEOUT';
                last;
            }
            if (length($inflated) > 100000) {
                # limit size of compressed section to 100KB
                $error = 'TOOBIG';
                last;
            }
        }
    } else {
        $error = 'NOZLIB';
    }
    $error .= ":" if $inflated && $error;
    return "$prefix<<<$error$inflated>>>$postfix";
}

sub filter_pdf_expand {
    my ($pdf) = @_;

    my $tmp = $$pdf;
    $tmp =~ s#(<<[^<]*/Filter\s*/FlateDecode[^>]*>>\s*stream(?:\n\r|\r\n|\n|))(.*?)(endstream)#flate_decode($1, $2, $3)#seg;

    return $tmp;

    return $$pdf;
}

sub filter_hex {
    my ($text) = @_;

    my $result = pack("H*", $text);

    $result =~ s/([()\\])/\\$1/g;
    return $result;
}

sub filter_pdf_totext {
    my ($pdf) = @_;

    # TODO: This is *really* simple text extraction code.  It has the following problems:
    #  - doesn't handle <01>Tj style numeric characters
    #  - takes no account of character mapping
    #  - non-ASCII glyphs in fonts would render garbage
    #  - order of characters in document need not match rendered locations 

    # We can fix the first one relatively easily.  The second is much harder,
    # but not impossible.  Number three is a nightmare scenario.  Number four
    # is also hard to decode, but would leave a 'fingerprint' in the PDF which
    # would be easily to tag as suspicious.

    my @result = ();
    my $index;
    my $next = 0;

    my $text = $$pdf;
    $text =~ s{([^<])<([A-Fa-f0-9]+)>(\s*Tj)}{"$1(" . filter_hex($2) . ")$3"}esg; # Turn hex string into conventional string
    $text =~ s{\-[\d\.]+[ \t]+[\-\d\.]+[ \t]+Td}{(\\n)Tj}gis; # turn minus positions into \\n
    $text =~ s{[\d\.]+[ \t]+[\d\.]+[ \t]+rg\b}{(\\n)Tj}gis; # turn absolute locations into a CR

    SCAN :
    while (($index = index($text, '(', $next)) != -1) {
        $next = $index + 1;
        next if $index > 0 && substr($text, $index - 1, 1) eq '\\';
        for (substr($text, $index + 1)) {
            my $pos = 0;
            my $recurse = 0;
            BRACKETMATCH :
            while (1) {
                my $close = $pos;
                while (($close = index($_, ')', $close)) != -1) {
                    last if $close == 0 || substr($_, $close - 1, 1) ne '\\';
                    $close = $close + 1;
                }
                next SCAN if $close == -1; # short circuit; this isn't text.

                next SCAN if $close > 2 && index($_, "\n") < $close;
                next SCAN if $close > 80; # Don't add more than 80 characters at a time.

                my $open = $pos;
                while (($open = index($_, '(', $open)) != -1) {
                    last if $open == 0 || substr($_, $open - 1, 1) ne '\\';
                    $open = $open + 1;
                }

                if ($open != -1 && $open < $close) {
                    $recurse++;
                    $pos = $open + 1;
                    next BRACKETMATCH;
                } else {
                    $pos = $close + 1;
                    last BRACKETMATCH if $recurse == 0;
                    $recurse--;
                }
            }
            if (substr($_, $pos) =~ /^\s*Tj/s) {
                push @result, substr($_, 0, $pos - 1);
                $next = $index + $pos;
            }
        }
    }

    $_ = join('', @result);

    s{\\n}{\n}gs;      # newlines can be escaped
    s{(?<!\\)\\(\d{1,3})}{oct($1) > 255 ? "\\$1" : chr(oct($1))}ges; # characters defined using octal
    s{\\([()\\])}{$1}gs; # remove backslashes from certain escaped characters
    tr[\200-\377][]d;    # strip top-bit set characters
    tr[\000-\011][]d;    # strip control characters
    tr[\013-\037][]d;    #  except for CR

    return $_;
}

1;
