package ActiveState::SpamEngine::Plugin::redb;

use strict;

use ActiveState::SpamEngine qw(_get1 PMX_TYPE_CALLBACK PMX_TYPE_INTEGER);
use ActiveState::Transformer;
use ActiveState::SpamEngine::DataDefinedTransformer;

use base 'ActiveState::SpamEngine::Plugin';

our @SP_REQUIRE = qw(part);
our @SP_ATTRIB  = (
    "pcre-limit",           PMX_TYPE_INTEGER,   undef,
    "pcre-recurse",         PMX_TYPE_INTEGER,   undef,
    "log-queries",          PMX_TYPE_INTEGER,       0,
);

sub plugin {
    my ($engine, $name) = @_;
    for my $dep (@SP_REQUIRE) { $engine->load_plugin($dep); }
    $engine->declare_attributes(\@SP_ATTRIB, $name);
    $engine->subscribe($name, "core.compile", \&compile);
    1;
}

my (
    $limit,
    $recurse,
    $log_queries,
    $db_events,
    $db_events_nore,
    $db_dbs,
    $db_map,
    $db_tp,
    $db_qmax,
    $db_ddt,
    $transformer,
);

sub compile {
    my $engine = shift;
    $engine->enumerate_attributes("plugin.re.pcre-limit", \&_get1, \$limit);
    $engine->enumerate_attributes("plugin.re.pcre-recurse", \&_get1, \$recurse);
    $engine->enumerate_attributes("plugin.redb.pcre-limit", \&_get1, \$limit);
    $engine->enumerate_attributes("plugin.redb.pcre-recurse", \&_get1, \$recurse);
    $engine->enumerate_attributes("plugin.redb.log-queries", \&_get1, \$log_queries);

    $db_events = eval { $engine->nested_manager("redb.events") };
    $db_events_nore = eval { $engine->nested_manager("redb.events_nore") };
    $db_dbs = eval { $engine->nested_manager("redb.dbs") };
    $db_tp = eval { $engine->nested_db("redb.tp") };
    $db_qmax = eval { $engine->nested_db("redb.qmax") };
    $db_map = eval { $engine->nested_db("redb.map") };
    $db_ddt = eval { $engine->nested_db("redb.ddt") };

    if ($db_events && $db_events_nore && $db_dbs && $db_map && $db_tp && $db_qmax && $db_ddt) {
        eval {
            $db_ddt->enumerate(
                sub {
                    my ($db_events, $name) = @_;
                    my $code = $db_events->get($name);
                    ActiveState::SpamEngine::DataDefinedTransformer::define(
                        $name,
                        $code
                    );
                }, $db_ddt
            );

            my $transformer_plugins = $db_tp->get("#TRANSFORMER_PLUGINS#");
            my @tplist = split /\x00/, $transformer_plugins;
            $transformer = ActiveState::Transformer->new;
            for my $tp (@tplist) {
                $transformer->load_plugin($tp);
            }
        };
        if (!$@) {
            $db_events->master->enumerate(sub {
                my ($engine, $event) = @_;
                $engine->subscribe("redb", $event, \&handler);
                1;
            }, $engine);
	    $db_events_nore->master->enumerate(sub {
		my ($engine, $event) = @_;
                $engine->subscribe("redb", $event, \&handler_nore);
                1;
            }, $engine);
        }
        else {
            $engine->log("ERROR", "Error in setting up global redb transformer: $@\n");
        }
    }

    $engine->initialised("redb");
    1;
}

sub handler {
    my ($engine, $extra) = splice(@_, 0, 2);
    $extra->{P}{redb}{query_info} ||= {};
    my $event = $extra->{EVENT};
    my $definition = $db_events->nest($event);
    $engine->pcre->match_asdb_v2(
        {
            code => \&check_db,
            data => [$engine, $extra],
            logger => $engine,
            asdb => $definition,
            when => 1,
            save => 2,
            limit => $limit,
            recurse => $recurse,
        },
        @_,
    );
}

sub handler_nore {
    my ($engine, $extra) = splice(@_, 0, 2);
    $extra->{P}{redb}{query_info} ||= {};
    my $event = $extra->{EVENT};
    my $definition = $db_events_nore->nest($event);
    $definition->enumerate(
	sub{
	    my $data = shift;
	    my $rule = shift;
	    check_db(
		[ $data->[0], $data->[1] ],
		$rule,
		$data->[2],
		$data->[2]
	    );
	},
	[$engine, $extra, "@_" ]
    );

}

sub check_db {
    my ($data, $rule, $m0, $m1) = @_;
    my ($engine, $extra) = @$data;

    my $qinfo = $extra->{P}{redb}{query_info};
    unless (exists $qinfo->{$rule}) {
        $qinfo->{$rule}{max} =
            $db_qmax->get($rule) || 1;
        $qinfo->{$rule}{count} = 0;
    }
    $qinfo->{$rule}{count} += 1;
    if ($log_queries) {
        $engine->log("DEBUG", "REDB QUERY DEPTH: $rule $qinfo->{$rule}{count}");
    }
    return 0 if $qinfo->{$rule}{count} >
                $qinfo->{$rule}{max};

    my $check_db_name = $db_map->get($rule);
    my $check_db = $db_dbs->nest($check_db_name);
    my $pipeline_list = $db_tp->get($rule);
    my $result = $m1;

    if ($pipeline_list) {
        my @tp_list = split /\x00/, $pipeline_list;
        my $pipeline = $transformer->pipeline(@tp_list);
        $result = $pipeline->transform($m1);
    }
    if ($log_queries) {
        $engine->log("DEBUG", "REDB QUERY LOG: $rule $m0 [$m1] [$result] ($check_db_name)");
    }

    if ($check_db && $check_db->get($result)) {
        $engine->addfeature($extra, $rule, "$m0 [$m1] [$result]");
    }
    return 1;
}

1;
