package ActiveState::SpamEngine::Plugin::sig;

use strict;

use ActiveState::SpamEngine qw(PMX_TYPE_CALLBACK PMX_TYPE_INTEGER
                               PMX_TYPE_STRING DEBUG
                               _get1 _getall);
use ActiveState::Sig;
use base 'ActiveState::SpamEngine::Plugin';

# XXX To do:
# - document this plugin.
#   - structure of the part that is exposed to the engine
#   - class methods exported
#   - attributes and where/how they are set


our @SP_REQUIRE = qw(part sxl);
our @SP_ATTRIB  = (
    # KEY           TYPE                DEFAULT
    'log-sums',     PMX_TYPE_INTEGER,   undef,
    'save-match',   PMX_TYPE_INTEGER,   undef,
    'max-bytes',    PMX_TYPE_INTEGER,   undef,      # XXX this needs to be
                                                    # definable per sig plugin
                                                    # (or sig algo)
);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);
    $eng->subscribe($name, 'core.compile', \&compile);
    return 1;
}

my ($asdb_cache, $log_sums, $save_match, $max_bytes, $sig_plugins, $sig, %sxl);

sub compile {
    my $eng = shift;
    $asdb_cache = eval { $eng->nested_db("sig.cache") } or return;
    $sig_plugins = eval { $eng->db()->get("sig.plugins") } or return;

    $sxl{'nw_body'} = $eng->sxl('sig'); # FIXME replace with sigb when zone becomes available
    $sxl{'nw_para'} = $eng->sxl('sig'); # FIXME replace with sigp when zone becomes available
    $sxl{'DEFAULT'} = $eng->sxl('sig');

    $eng->enumerate_attributes('plugin.sig.log-sums', \&_get1, \$log_sums);
    $eng->enumerate_attributes('plugin.sig.save-match', \&_get1, \$save_match);

    # Maximum bytes to scan?
    $eng->enumerate_attributes("plugin.sig.max-bytes", \&_get1, \$max_bytes);
    $eng->enumerate_attributes('core.max-bytes-scanned', \&_get1, \$max_bytes)
        if not defined $max_bytes;

    $eng->subscribe('sig', 'part.info', \&scan_message);
    $eng->initialised("sig");

    $sig = ActiveState::Sig->new(
        asdb_cache => $asdb_cache,
        save_match => $save_match,
        max_bytes  => $max_bytes,
    ) or do {
        $eng->log("ERROR", "sig: unable to get sig object");
        return 1;
    };

    for my $sig_plugin (split /\x00/, $sig_plugins) {
        $sig->load_plugin($eng, $sig_plugin);
    }

    1;
}

sub sxl_hit {
    my ($x, $query, $result) = @_;
    my ($engine, $xtra) = @$x;
    $engine->addfeature($xtra, $result, $query);
}

sub scan_message {
    my ($eng, $xtra, $msg) = @_;

    # Clear all the cached crud we save in the sig object (about each message)
    $sig->reset;

    # Calculates sigs against every plugin and algorithm; saves result in sig object.
    $sig->calculate($eng, $xtra, $msg);

    # Get number of sigs and the object containing them all
    my $sigs = $sig->sigs or return 1;
    my %info = $sig->info;

    $xtra->{P}{part}{sig} = \%info;
    my @info_keys = sort keys %info;

    # log the sums
    if ($log_sums) {
        for my $type (@info_keys) {
            my $info = $info{$type};
            $info = [$info] unless ref($info) eq "ARRAY";
            for my $item (@$info) {
                next unless $item->{sig};
                $eng->log(
                    'DEBUG',
                    join('',
                        "sig: checksum $type: $item->{sig}",
                        $item->{match} ? " [$item->{match}]" : (),
                    )
                ) if DEBUG;
            }
        }
    }

    # check the asdb_cache
    for my $algo (@info_keys) {
        my $info = $info{$algo};
        $info = [$info] unless ref($info) eq "ARRAY";
        for my $item (@$info) {
            my $rule = $item->{hit} or next;
            $eng->addfeature(
                $xtra,
                $rule,
                # XXX are these things in the right order? {sig} is a hex
                # value these days...
                $item->{match} || $item->{sig} || $algo
            );
            return 1;
        }
    }

    if (%sxl)
    {
        # Note that any response to the SXL query will now have to wait
        # until we receive the response.  The SXL plugin will call back to
        # the plugin if it receives a response

        my %sigs = $sig->sigs;
        for my $algorithm (keys %sigs) {
            if (exists $sxl{$algorithm}) {
                $sxl{$algorithm}->get($_, \&sxl_hit, [$eng, $xtra]) for (@{$sigs{$algorithm}});
            } else {
                $sxl{'DEFAULT'}->get($_, \&sxl_hit, [$eng, $xtra]) for (@{$sigs{$algorithm}});
            }
        }
    }
}

1;
