package ActiveState::SpamEngine::Plugin::spamassassin;

use strict;
use ActiveState::SpamEngine qw(_get1 _getall PMX_TYPE_INTEGER
                               PMX_TYPE_CALLBACK PMX_TYPE_STRING);

use base 'ActiveState::SpamEngine::Plugin';

our @SP_REQUIRE = qw(part);
our @SP_ATTRIB  = (
    "ok-locales",           PMX_TYPE_STRING,    'all',
    "save-match",           PMX_TYPE_INTEGER,   undef,
);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);
    $eng->subscribe($name, "core.compile", \&compile);
    1;
}

#=============================================================================
# Support code for accessing the headers hash
#=============================================================================
sub h_get {
    my ($hdr, $hdrs) = @_;
    $hdr = lc $hdr;
    my $addr = ($hdr =~ s/:addr$//);
    my @v = exists $hdrs->{$hdr} ? @{$hdrs->{$hdr}} : ();
    if ($addr) {
        # XXX there must be a better way
        @v = (split /,/, join(' ', @v));
        foreach (@v) {
            chomp;
            s/\r?\n//g;
            s/\s*\(.*?\)//g;            # strip out (comments)
            s/^[^<]*?<(.*?)>.*$/$1/;    # "Foo Blah" <jm@foo> or <jm@foo>
#           s/, .*$//gs;                # multiple addrs: return 1st
            s/ ;$//gs;                  # 'undisclosed-recipients: ;'
        }
    }
    return @v if wantarray;
    return join("\n", map { s/\r?\n[\s\t]+/ /g; $_;} @v) . "\n"
       if @v > 1 and ! $addr;
    return join(' ', @v);
}

# Used to compare positions of headers. In scalar context, returns the
# index of the first occurrence of '$hdr'. In list context, returns all
# positions of '$hdr'.
sub h_find {
    my ($hdr, $h) = @_;
    if (wantarray) {
        my @ix;
        for (0 .. $#{$h->{__ORDER_DUPS__}}) {
            push @ix, $_ if $h->{__ORDER_DUPS__}[$_] eq $hdr;
        }
        return @ix;
    }
    else {
        for (0 .. $#{$h->{__ORDER__}}) {
            return $_ if $h->{__ORDER__}[$_] eq $hdr;
        }
        return -1; # don't really know what to return here.
    }
}

#=============================================================================
# HTML tests
#=============================================================================
sub html_tag {
    my ($eng, $xtra, $tag, $attr, $num) = @_;
    my $x = $xtra->{P}{spamassassin}{html} ||= {};

    if ($num==1) {
        $x->{__existing_tags__}{lc $tag} = 1;
    }

    if ($tag eq 'table' and exists $attr->{border}
            and $attr->{border} =~ /^(\d+)/) {
        $x->{thick_border} ||= $1 if $1 > 1;
    }
    if ($tag eq 'body' or $tag eq 'frame') {
        for (keys %$attr) {
            if (/^on(?:Load|Unload|BeforeUnload)$/i) {
                $x->{javascript_very_unsafe} ||= $_;
            }
        }
    }
    if ($tag eq 'body' and exists $attr->{bgcolor}) {
        $x->{bgcolor_nonwhite} ||= $attr->{bgcolor}
            unless is_white($attr->{bgcolor});
    }
    if ($tag eq 'font' and exists $attr->{size}) {
        $x->{big_font} ||= $attr->{size} if (
            ($attr->{size} =~ /^\s*(\d+)/ and $1 >= 3) or
            ($attr->{size} =~ /\+(\d+)/   and $1 >= 2)
        );
    }
    if ($tag eq 'font' and exists $attr->{color}) {
        local $_ = lc $attr->{color};
        $x->{font_color_nohash} ||= $_ if /^[0-9a-f]{6}$/;
        $x->{font_color_unsafe} ||= $_ if (
            /^#?[0-9a-f]{6}$/ && ! /^#?(?:00|33|66|80|99|cc|ff){3}$/
        );
        $x->{font_color_name} ||= $_ if (
            !/^#?[0-9a-f]{6}$/ && !/^(?:navy|gray|red|white)$/
        );
        if (my ($r, $g, $b) = rgb_unpack($_)) {
            my ($h, $s, $v) = rgb_to_hsv($r, $g, $b);
            if (not defined $h) {
                $x->{font_gray} ||= $_ if !($v == 0 or $v == 255);
            }
            elsif ($h < 30 or $h > 330) {
                $x->{font_red} ||= $_;
            }
            elsif ($h < 90) {
                $x->{font_yellow} ||= $_;
            }
            elsif ($h < 150) {
                $x->{font_green} ||= $_;
            }
            elsif ($h < 210) {
                $x->{font_cyan} ||= $_;
            }
            elsif ($h < 270) {
                $x->{font_blue} ||= $_;
            }
            elsif ($h < 330) {
                $x->{font_magenta} ||= $_;
            }
        }
        else {
            $x->{font_color_unknown} ||= $_;
        }
    }
    if ($tag eq 'font' and exists $attr->{face}) {
        local $_ = $attr->{face};
        $x->{font_face_caps} ||= $_ if /[A-Z]{3}/;
        $x->{font_face_bad}  ||= $_ if !/^[a-z][a-z -]*[a-z](?:,\s*[a-z][a-z -]*[a-z])*$/i;

        my %face = (
            'arial' => 1,
            'arial (hebrew)' => 1,
            'arial black' => 1,
            'arial cyr' => 1,
            'arial narrow' => 1,
            'arial rounded mt bold' => 1,
            'arial unicode ms' => 1,
            'ariel' => 1,
            'book antiqua' => 1,
            'bookman old style' => 1,
            'brush script mt' => 1,
            'centaur' => 1,
            'century' => 1,
            'century gothic' => 1,
            'clarendon light' => 1,
            'comic sans ms' => 1,
            'cooper black' => 1,
            'copperplate gothic bold' => 1,
            'courier' => 1,
            'courier new' => 1,
            'cursive' => 1,
            'default' => 1,
            'french script mt' => 1,
            'futura' => 1,
            'futura bk' => 1,
            'futura md' => 1,
            'garamond' => 1,
            'garamond macron' => 1,
            'geneva' => 1,
            'georgia' => 1,
            'helv' => 1,
            'helvetica' => 1,
            'impact' => 1,
            'lucida' => 1,
            'lucida console' => 1,
            'lucida fax' => 1,
            'lucida grande' => 1,
            'lucida handwriting' => 1,
            'lucida sans' => 1,
            'lucida sans unicode' => 1,
            'microsoft sans serif' => 1,
            'mingliu' => 1,
            'monaco' => 1,
            'mono' => 1,
            'monospace' => 1,
            'monotype corsiva' => 1,
            'ms sans serif' => 1,
            'ms serif' => 1,
            'ms shell dlg' => 1,
            'optima' => 1,
            'palatino' => 1,
            'palatino linotype' => 1,
            'san serif' => 1,
            'sans ' => 1,
            'sans serif' => 1,
            'script' => 1,
            'serif' => 1,
            'sunsans regular' => 1,
            'swiss' => 1,
            'symbol' => 1,
            'tahoma' => 1,
            'thorndale' => 1,
            'times' => 1,
            'times new roman' => 1,
            'times new roman cyr' => 1,
            'trebuchet' => 1,
            'trebuchet ms' => 1,
            'utopia' => 1,
            'verdana' => 1,
            'webdings' => 1,
            'wingdings' => 1,
            'zapf chancery' => 1,
        );

        for (split ',', lc) {
            s/\s*$//;
            s/^\s*//;
            s/-/ /g;
            s/(["'])(.*)\1/$2/;
            $x->{font_face_odd} ||= $_ if not $face{$_};
        }
    }

    if ( ($tag eq 'img' and exists $attr->{src}) or
         ($tag =~ /^(?:body|table|tr|td)$/ and exists $attr->{background})
       )
    {
        my $src = ($tag eq 'img') ? $attr->{src} : $attr->{background};
        if ($src =~ /\.(?:pl|cgi|php|asp|jsp|cfm|shtml)/) {
            $x->{web_bugs} = $src;
        }
    }

    if ($tag eq 'img') {
        my $img_in_a  = grep { $_ eq 'a'  } @{$x->{__htmltree__}};
        if ($img_in_a || exists $attr->{'usemap'}) {
            my $img_in_td = grep { $_ eq 'td' } @{$x->{__htmltree__}};

            $x->{'imgspam'}{'a_img_outside_td'}++ if !$img_in_td;
            $x->{'imgspam'}{'a_img_inside_td'}++  if $img_in_td;
        }
    }

    if ($tag eq "meta" &&
        exists $attr->{'http-equiv'} &&
        exists $attr->{content} &&
        $attr->{'http-equiv'} =~ /Content-Type/i &&
        $attr->{content} =~ /\bcharset\s*=\s*["']?([^"']+)/i)
    {
        push @{$x->{'charsets'}}, $1;
    }


    $x->{javascript} ||= $tag if $tag eq 'script';
    $x->{relaying_frame} ||= $tag if $tag =~ /^i?frame$/;
    $x->{embeds} ||= $tag if $tag =~ /^(?:object|embed)$/;

    # Maintain a table stack.
    if (grep { $tag eq $_ } qw(form table td a)) {
        if ($num == 1) {
            push @{$x->{__htmltree__}}, $tag;
        }
        elsif ($x->{__htmltree__} and $tag) {
            # Pop everything after the last $tag off the stack.
            while (@{$x->{__htmltree__}}) {
                my $t = pop @{$x->{__htmltree__}};
                last if $t eq $tag;
            }
        }
    }
}

sub html_cmnt {
    my ($eng, $xtra, $cmnt) = @_;
    my $x = $xtra->{P}{spamassassin}{html} ||= {};
    $x->{comment_8bit} ||= $cmnt if $cmnt =~ /[\x80-\xff]{3}/;
    $x->{comment_saved_url} ||= $cmnt if $cmnt =~ /<!-- saved from url=\(\d{4}\)/;
    $x->{comment_unique_id} ||= $cmnt if (
        $cmnt =~ /<!(?:--)?\s*(?:[\d.]+|[a-fA-F\d]{5,}|\S{10,})\s*(?:--)?>/
    );
}

# The test itself is called from the spamassassins.
sub html_test {
    my ($eng, $xtra, undef, $test) = @_;
    push @{$xtra->{M}}, $xtra->{P}{spamassassin}{html}{$test};
    $xtra->{P}{spamassassin}{html}{$test};
}

# Test "base" flags set from the part plugin
sub base_test {
    my ($eng, $xtra, undef, $test, $ge, $lt) = @_;
    my $x = $xtra->{P}{part}{__meta__}{$test};
    return unless defined $x; # will always fail
    push @{$xtra->{M}}, $x;
    if (defined $ge and defined $lt) {
        return $x >= $ge && $x < $lt;
    }
    elsif (defined $ge) {
        return $x >= $ge;
    }
    else {
        return defined $x;
    }
}

sub rgb_unpack {
    my $rgb = name_to_rgb(shift);
    return unless $rgb =~ s/^#?([\da-f]{6})$/$1/i;
    my ($r, $g, $b) = unpack('C*', pack('H*', $rgb));
}

sub rgb_to_hsv {
    my ($r, $g, $b) = @_;
    my ($h, $s, $v);

    my ($max, $min) = ($r > $g) ? ($r, $g) : ($g, $r);
    $max = $b if $b > $max;
    $min = $b if $b < $min;
    $v = $max;
    $s = $max ? ($max - $min) / $max : 0;
    if ($s == 0) {
        $h = undef;
    }
    else {
        my $d = $max - $min;
        my $cr = ($max - $r) / $d;
        my $cg = ($max - $g) / $d;
        my $cb = ($max - $b) / $d;
        if ($r == $max) {
            $h = $cb - $cg;
        }
        elsif ($g == $max) {
            $h = 2 + $cr - $cb;
        }
        elsif ($b == $max) {
            $h = 4 + $cg - $cr;
        }
        $h *= 60;
        $h += 360 if $h < 0;
    }
    return ($h, $s, $v);
}

sub is_white {
    name_to_rgb($_[0]) eq '#ffffff';
}

*name_to_rgb = \&ActiveState::SpamEngine::Plugin::part::name_to_rgb;

#=============================================================================
# Rules compiler; nothing to see here.
#=============================================================================
my $db;
our $PHISHING_TARGETS_DB;
our $VALID_CHARSETS_DB;
sub compile {
    my $eng = shift;
    eval {
        $db = $eng->nested_manager("spamassassin");
        $PHISHING_TARGETS_DB = $eng->nested_db("sa.phished");
        $VALID_CHARSETS_DB   = $eng->nested_db("sa.charset");
    };
    return if $@;

    require ActiveState::SpamEngine::Plugin::spamassassin::Tests;
    require ActiveState::SpamEngine::Plugin::spamassassin::MailingList;
    clear_caches(); # this is defined in ::Tests;

    $eng->subscribe("spamassassin", "part.html_tag", \&html_tag);
    $eng->subscribe("spamassassin", 'part.html_cmnt', \&html_cmnt);

    my $save;
    $eng->enumerate_attributes("plugin.spamassassin.save-match", \&_get1, \$save);
    my $handler = $save ? \&_fwd_save : \&_fwd;
    $db->master->enumerate(sub {
        my ($data, $event) = @_;
        my ($eng, $handler) = @$data;
        $eng->subscribe("spamassassin", $event, $handler);
        1;
    }, [$eng, $handler]);

    $eng->initialised("spamassassin");
    1;
}

sub _fwd {
    my ($eng, $xtra) = splice(@_, 0, 2);
    my $event = $xtra->{EVENT};
    my $def = $db->nest($event);
    for my $loop (@_) {
        $def->enumerate(sub {
            my ($data, $rule, $def) = @_;
            my ($eng, $xtra, $loop) = @$data;
            my ($function, @args) = split /\x01/, $def;
            no strict 'refs';
            if (!defined &$function) {
                $eng->log(
                    "ERROR",
                    "spamassassin: rule $rule: function '$function' not found"
                );
            }
            elsif (&$function($eng, $xtra, $$loop, @args)) {
                $eng->addfeature($xtra, $rule);
            }
            1;
        }, [$eng, $xtra, \$loop]);
    }
}

sub _fwd_save {
    my ($eng, $xtra) = splice(@_, 0, 2);
    my @m; local $xtra->{M} = \@m;
    my $event = $xtra->{EVENT};
    my $def = $db->nest($event);
    for my $loop (@_) {
        $def->enumerate(sub {
            my ($data, $rule, $def) = @_;
            my ($eng, $xtra, $loop, $m) = @$data;
            my ($function, @args) = split /\x01/, $def;
            no strict 'refs';
            @$m = ();
            if (!defined &$function) {
                $eng->log(
                    "ERROR",
                    "spamassassin: rule $rule: function '$function' not found"
                );
            }
            elsif (&$function($eng, $xtra, $$loop, @args)) {
                $eng->addfeature($xtra, $rule, @$m ? join($;, @$m) : ());
            }
            1;
        }, [$eng, $xtra, \$loop, \@m]);
    }
}

1;
