#
# Copyright (c) 2002 ActiveState Corp.  All rights reserved.
#
# Portions of this file are derived from the Mail::SpamAssassin module
# from CPAN.  Such derived portions are used here under Perl's Artistic
# License.
#
# SpamAssassin is Copyright (c) Justin Mason <jm@jmason.org>.
#
# See http://spamassassin.org/ for further information on SpamAssassin.
#

package ActiveState::SpamEngine::Plugin::spamassassin::Tests;

package ActiveState::SpamEngine::Plugin::spamassassin;

use strict;
use Time::Local;

our $PHISHING_TARGETS_DB;
our $VALID_CHARSETS_DB;
use ActiveState::SpamEngine::IsIPV4 qw(is_ipv4);

my $IP_ADDRESS = qr/\b
                  (?:1\d\d|2[0-4]\d|25[0-5]|\d\d|\d)\.
                  (?:1\d\d|2[0-4]\d|25[0-5]|\d\d|\d)\.
                  (?:1\d\d|2[0-4]\d|25[0-5]|\d\d|\d)\.
                  (?:1\d\d|2[0-4]\d|25[0-5]|\d\d|\d)
                  \b/x;

# A mapping of known country codes to frequent charsets used therein.
# note that the ISO and CP charsets will already have been permitted,
# so only "unusual" charsets should be listed here.
#
# Country codes should be lowercase, charsets uppercase.
#
sub _h {
    my %h;
    $h{$_}++ for @_;
    return \%h;
}
my %charsets_for_locale = (
  # Japanese
  'ja' => _h(qw'EUCJP JISX020119760 JISX020819830 JISX020819900 JISX020819970
                JISX021219900 JISX021320001 JISX021320002 SHIFT_JIS SHIFTJIS
                ISO2022JP SJIS JIS7'),
  # Korea
  'ko' => _h(qw'EUCKR EUCKR'),
  # Cyrillic
  'ru' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  'ka' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  'tg' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  'be' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  'uk' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  'bg' => _h(qw'KOI8R KOI8U KOI8T ISOIR111 CP1251 GEORGIANPS CP1251 PT154'),
  # Thai
  'th' => _h(qw'TIS620'),
  # Chinese (simplified and traditional)
  'zh' => _h(qw'GB2312 GB231219800 GB18030 GBK BIG5HKSCS BIG5 EUCTW'),
);

sub is_charset_ok_for_locales {
    my ($cs, @locales) = @_;

    $cs = uc $cs; $cs =~ s/[^A-Z0-9]//g; study $cs;

    # always OK (the net speaks mostly roman charsets)
    return 1 if ($cs =~ /^USASCII/);
    return 1 if ($cs =~ /^ISO8859/);
    return 1 if ($cs =~ /^ISO10646/);
    return 1 if ($cs =~ /^UTF/);
    return 1 if ($cs =~ /^UCS/);
    return 1 if ($cs =~ /^CP125/);
    return 1 if ($cs =~ /^WINDOWS125/);
    return 1 if ($cs eq 'IBM852');
    return 1 if ($cs =~ /^UNICODE11UTF[78]/); # wtf? never heard of it
    return 1 if ($cs eq 'XUNKNOWN'); # added by sendmail when converting to 8bit

    foreach my $locale (@locales) {
        if (!defined($locale) || $locale eq 'C') { $locale = 'en'; }
        $locale = substr($locale, 0, 2);
        return 1 if $charsets_for_locale{$locale}{$cs};
    }

    return 0;
}

###########################################################################
# HEAD TESTS:
###########################################################################
# From and To have same address, but are not exactly the same and
# neither contains intermediate spaces.
sub check_for_from_to_same {
  my ($eng, $xtra, $h, $type) = @_;

  my $hdr_from = h_get('From', $h);
  my $hdr_to = h_get('To', $h);
  return 0 if (!length($hdr_from) || !length($hdr_to) ||
               $hdr_from eq $hdr_to);

  my $addr_from = h_get('From:addr', $h);
  my $addr_to = h_get('To:addr', $h);
  # BUG: From:addr and To:addr sometimes contain whitespace
  $addr_from =~ s/\s+//g;
  $addr_to =~ s/\s+//g;
  return 0 if (!length($addr_from) || !length($addr_to) ||
               $addr_from ne $addr_to);

  if ($hdr_from =~ /^\s*\S+\s*$/ && $hdr_to =~ /^\s*\S+\s*$/) {
    return 1;
  }
}

sub sorted_recipients {
  my ($eng, $xtra, $h) = @_;

  if (!exists $xtra->{P}{spamassassin}{tocc_sorted}) {
    _check_recipients($xtra, $h);
  }
  return $xtra->{P}{spamassassin}{tocc_sorted};
}

sub similar_recipients {
  my ($eng, $xtra, $h, $min, $max) = @_;

  if (!exists $xtra->{P}{spamassassin}{tocc_similar}) {
    _check_recipients($xtra, $h);
  }
  return (($min eq 'undef' || $xtra->{P}{spamassassin}{tocc_similar} >= $min) &&
          ($max eq 'undef' || $xtra->{P}{spamassassin}{tocc_similar} < $max));
}

# best experimentally derived values
use constant TOCC_SORTED_COUNT => 7;
use constant TOCC_SIMILAR_COUNT => 5;
use constant TOCC_SIMILAR_LENGTH => 2;

sub _check_recipients {
  my ($xtra, $h) = @_;

  my @address;

  # ToCc: pseudo-header works best, but sometimes Bcc: is better
  for ('To:addr', 'Cc:addr', 'Bcc:addr') {
    my $to = h_get($_, $h);     # get recipients
    $to =~ s/\(.*?\)//g;        # strip out the (comments)
    @address = ($to =~ m/([\w.=-]+\@\w+(?:[\w.-]+\.)+\w+)/g);
    last if scalar(@address) >= TOCC_SIMILAR_COUNT;
  }

  # ideas that had both poor S/O ratios and poor hit rates:
  # - testing for reverse sorted recipient lists
  # - testing To: and Cc: headers separately
  $xtra->{P}{spamassassin}{tocc_sorted} = (scalar(@address) >= TOCC_SORTED_COUNT &&
                          join(',', @address) eq (join(',', sort @address)));

  # a good S/O ratio and hit rate is achieved by comparing 2-byte
  # substrings and requiring 5 or more addresses
  $xtra->{P}{spamassassin}{tocc_similar} = 0;
  if (scalar (@address) >= TOCC_SIMILAR_COUNT) {
    my @user = map { substr($_,0,TOCC_SIMILAR_LENGTH) } @address;
    my @fqhn = map { m/\@(.*)/ } @address;
    my @host = map { substr($_,0,TOCC_SIMILAR_LENGTH) } @fqhn;
    my $hits = 0;
    my $combinations = 0;
    for (my $i = 0; $i <= $#address; $i++) {
      for (my $j = $i+1; $j <= $#address; $j++) {
        $hits++ if $user[$i] eq $user[$j];
        $hits++ if $host[$i] eq $host[$j] && $fqhn[$i] ne $fqhn[$j];
        $combinations++;
      }
    }
    $xtra->{P}{spamassassin}{tocc_similar} = $hits / $combinations;
  }
}

sub check_for_multiple_user_agents {
    my ($eng, $xtra) = @_;
    my $n = grep /^__USER_AGENT_/, keys %{$eng->{feats}};
    $n > 1;
}

###########################################################################
# yet another test for faked Received: headers (FORGED_RCVD_TRAIL)
sub check_for_forged_received_trail {
    my ($eng, $xtra, $hdrs) = @_;

    my @received = grep(/\S/, split(/\n/, h_get('Received', $hdrs)));
    my @by;
    my @from;
    my @fromip;
    my $mismatch = 0;

    for (my $i = 0; $i <= $#received; $i++) {
        if ($received[$i] =~ s/\bby[\t ]+(\w+(?:[\w.-]+\.)+\w+)//i) {
            $by[$i] = lc($1);
            $by[$i] =~ s/.*\.(\S+\.\S+)$/$1/;
        }
        if ($received[$i] =~ s/\bfrom[\t ]+(\w+(?:[\w.-]+\.)+\w+)//i) {
            $from[$i] = lc($1);
            $from[$i] =~ s/.*\.(\S+\.\S+)$/$1/;
        }
        if ($received[$i] =~ s/^ \((?:\S+ |)\[(${IP_ADDRESS})\]\)//i) {
            $fromip[$i] = $1;
        }

        if (defined ($from[$i]) && defined($fromip[$i])) {
            if ($from[$i] =~ /^localhost(?:\.localdomain|)$/) {
                if ($fromip[$i] eq '127.0.0.1') {
                    # valid: bouncing around inside 1 machine, via the localhost interface.
                    # freshmeat newsletter does this.
                    $from[$i] = undef;
                }
            }
        }

        if ($i > 0 && defined($by[$i]) && defined($from[$i - 1]) &&
            ($by[$i] ne $from[$i - 1]))
        {
            $mismatch++;
        }
    }
    return ($mismatch > 1);
}



###########################################################################

sub check_for_forged_hotmail_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my $from = h_get('From:addr', $hdrs);
    if ($from !~ /hotmail.com/) { return 0; }

    my $rcvd = h_get('Received', $hdrs);
    $rcvd =~ s/\s+/ /gs;

    my $ip = h_get('X-Originating-Ip', $hdrs);

    # Hotmail formats its received headers like this:
    # Received: from hotmail.com (f135.law8.hotmail.com [216.33.241.135])
    # spammers do not ;)
    if (gated_through_received_hdr_remover($hdrs)) { return 0; }

    return 1 unless ($ip =~ /\d+\.\d+\.\d+\.\d+/);

    return 0 if $rcvd =~ /from\s+\S*hotmail.com\s+\(\s*(?:\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\]\s*)?(?i:helo[=\s]|\S+\.)hotmail(?:\.msn)?\.com[ \)]/;
    return 0 if $rcvd =~ /from \S+ by \S+\.hotmail(?:\.msn)?\.com with HTTP\;/;

    return 1;
}

# MSN_GROUPS
sub check_for_msn_groups_headers {
    my ($eng, $xtra, $h) = @_;

    return 0 unless (h_get('To', $h) =~ /<(\S+)\@groups\.msn\.com>/i);
    my $listname = $1;

    # from Theo Van Dinter, see
    # http://www.hughes-family.org/bugzilla/show_bug.cgi?id=591
    return 0 unless h_get('Message-Id', $h) =~ /^<$listname-\S+\@groups\.msn\.com>/;
    return 0 unless h_get('X-Loop', $h) =~ /^notifications\@groups\.msn\.com/;
    return 0 unless h_get('Return-Path', $h) =~ /<$listname-bounce\@groups\.msn\.com>/;

    $_ = h_get('Received', $h);
    return 0 if !/from mail pickup service by groups\.msn\.com\b/;
    return 1;

    # MSN Groups
    # Return-path: <ListName-bounce@groups.msn.com>
    # Received: from groups.msn.com (tk2dcpuba02.msn.com [65.54.195.210]) by
    #    dogma.slashnull.org (8.11.6/8.11.6) with ESMTP id g72K35v10457 for
    #    <zzzzzzzzzzzz@jmason.org>; Fri, 2 Aug 2002 21:03:05 +0100
    # Received: from mail pickup service by groups.msn.com with Microsoft
    #    SMTPSVC; Fri, 2 Aug 2002 13:01:30 -0700
    # Message-id: <ListName-1392@groups.msn.com>
    # X-loop: notifications@groups.msn.com
    # Reply-to: "List Full Name" <ListName@groups.msn.com>
    # To: "List Full Name" <ListName@groups.msn.com>
}

###########################################################################
sub check_for_forged_eudoramail_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my $from = h_get('From:addr', $hdrs);
    if ($from !~ /eudoramail.com/) { return 0; }

    my $rcvd = h_get('Received', $hdrs);
    $rcvd =~ s/\s+/ /gs;          # just spaces, simplify the regexp

    my $ip = h_get('X-Sender-Ip', $hdrs);
    if ($ip =~ /\d+\.\d+\.\d+\.\d+/) { $ip = 1; } else { $ip = 0; }

    # Eudoramail formats its received headers like this:
    # Received: from Unknown/Local ([?.?.?.?]) by shared1-mail.whowhere.com;
    #      Thu Nov 29 13:44:25 2001
    # Message-Id: <JGDHDEHPPJECDAAA@shared1-mail.whowhere.com>
    # Organization: QUALCOMM Eudora Web-Mail  (http://www.eudoramail.com:80)
    # X-Sender-Ip: 192.175.21.146
    # X-Mailer: MailCity Service

    if (gated_through_received_hdr_remover($hdrs)) { return 0; }

    if ($rcvd =~ /by \S*whowhere.com\;/ && $ip) { return 0; }

    return 1;
}

###########################################################################
sub check_for_forged_excite_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my $from = h_get('From:addr', $hdrs);
    if ($from !~ /excite.com/) { return 0; }

    my $rcvd = h_get('Received', $hdrs);
    $rcvd =~ s/\s+/ /gs;

    # Excite formats its received headers like this:
    # Received: from bucky.excite.com ([198.3.99.218]) by vaxc.cc.monash.edu.au
    #    (PMDF V6.0-24 #38147) with ESMTP id
    #    <01K53WHA3OGCA5W9MM@vaxc.cc.monash.edu.au> for luv@luv.asn.au;
    #    Sat, 23 Jun 2001 13:36:20 +1000
    # Received: from hippie.excite.com ([199.172.148.180]) by bucky.excite.com
    #    (InterMail vM.4.01.02.39 201-229-119-122) with ESMTP id
    #    <20010623033612.NRCY6361.bucky.excite.com@hippie.excite.com> for
    #    <luv@luv.asn.au>; Fri, 22 Jun 2001 20:36:12 -0700
    # spammers do not ;)

    if (gated_through_received_hdr_remover($hdrs)) { return 0; }

    if ($rcvd =~ /from \S*excite.com (\S+) by \S*excite.com/) { return 0; }

    return 1;
}

###########################################################################
sub check_for_forged_yahoo_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my $from = h_get('From:addr', $hdrs);
    if ($from !~ /yahoo\.com$/) { return 0; }

    my $rcvd = h_get('Received', $hdrs);
    $rcvd =~ s/\s+/ /gs;

    if (gated_through_received_hdr_remover($hdrs)) { return 0; }

    if ($rcvd =~ /by web\S+\.mail\.yahoo\.com via HTTP/) { return 0; }
    if ($rcvd =~ /by smtp\.\S+\.yahoo\.com with SMTP/) { return 0; }
    if ($rcvd =~
        /from \[$IP_ADDRESS\] by \S+\.(?:groups|grp\.scd)\.yahoo\.com with NNFMP/) {
      return 0;
    }

    # used in "forward this news item to a friend" links.  There's no better
    # received hdrs to match on, unfortunately.  I'm not sure if the next test is
    # still useful, as a result.
    #
    # search for msgid <20020929140301.451A92940A9@xent.com>, subject "Yahoo!
    # News Story - Top Stories", date Sep 29 2002 on
    # <http://xent.com/pipermail/fork/> for an example.
    #
    if ($rcvd =~ /\bmailer\d+\.bulk\.scd\.yahoo\.com\b/
        && $from =~ /\@reply\.yahoo\.com$/) { return 0; }

    if ($rcvd =~ /by \w+\.\w+\.yahoo\.com \(\d+\.\d+\.\d+\/\d+\.\d+\.\d+\) id \w+/) {
        # possibly sent from "mail this story to a friend"
        return 0;
    }

    return 1;
}

sub check_for_forged_juno_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my $from = h_get('From:addr', $hdrs);
    if($from !~ /\bjuno\.com/) { return 0; }

    if(gated_through_received_hdr_remover($hdrs)) { return 0; }

    my $xmailer = h_get('X-Mailer', $hdrs);
    my $xorig = h_get('X-Originating-IP', $hdrs);
    my $rcvd = h_get('Received', $hdrs);
    if (!$xorig) {  # New style Juno has no X-Originating-IP header, and other changes
        if($rcvd !~ /from.*\b(?:juno|untd)\.com.*[\[\(]$IP_ADDRESS[\]\)].*by/
            && $rcvd !~ / cookie\.(?:juno|untd)\.com /) { return 1; }
        if($xmailer !~ /Juno /) { return 1; }
    } else {
        if($rcvd !~ /from.*\bmail\.com.*\[$IP_ADDRESS\].*by/) { return 1; }
        if($xorig !~ /$IP_ADDRESS/) { return 1; }
        if($xmailer !~ /\bmail\.com/) { return 1; }
    }

    return 0;
}

sub check_for_from_domain_in_received_headers {
    my ($eng, $xtra, $hdrs, $domain, $desired) = @_;
    my $st = $xtra->{P}{spamassassin}{from_domain_in_received};

    if ($st) {
        if (exists $st->{$domain}) {
            if ($desired eq 'true') {
                # See use of '0e0' below for why we force int() here:
                return int($st->{$domain});
            }
            else {
                # And why we deliberately do NOT use integers here:
                return !$st->{$domain};
            }
        }
    }
    else {
        $st = $xtra->{P}{spamassassin}{from_domain_in_received} = {};
    }

    my $from = h_get('From:addr', $hdrs);
    if ($from !~ /\b\Q$domain\E/i) {
        # '0e0' is Perl idiom for "true but zero":
        $st->{$domain} = '0e0';
        return 0;
    }

    my $rcvd = h_get('Received', $hdrs);
    if ($rcvd =~ /from.*\b\Q$domain\E.*[\[\(][0-9]+\.[0-9]+\.[0-9]+\.[0-9]+[\]\)].*by.*\b\Q$domain\E/) {
        $st->{$domain} = 1;
        return ($desired eq 'true');
    }

    $st->{$domain} = 0;
    return ($desired ne 'true');
}

sub gated_through_received_hdr_remover {
    my ($hdrs) = @_;

    my $txt = h_get("Mailing-List", $hdrs);
    if (defined $txt && $txt =~ /^contact \S+\@\S+\; run by ezmlm$/) {
        my $dlto = h_get("Delivered-To", $hdrs);
        my $rcvd = h_get("Received", $hdrs);

        # ensure we have other indicative headers too
        if ($dlto =~ /^mailing list \S+\@\S+/ &&
            $rcvd =~ /qmail \d+ invoked by .{3,20}\); \d+ ... \d+/)
        {
            return 1;
        }
        # jm: this line *was* included:
        #   $rcvd =~ /qmail \d+ invoked from network\); \d+ ... \d+/ &&
        # but I've found FPs where it did not appear in the mail; it's
        # not required.
    }

    if (h_get("Received", $hdrs) !~ /\S/) {
        # we have no Received headers!  These tests cannot run in that case
        return 1;
    }

    # MSN groups removes Received lines. thanks MSN
    if (h_get("Received", $hdrs) =~ /from groups\.msn\.com \(\S+\.msn\.com /) {
        return 1;
    }

    return 0;
}

###########################################################################

sub check_subject_for_lotsa_8bit_chars {
    my ($eng, $xtra, $hdrs) = @_;

    local $_ = h_get('Subject', $hdrs);
    # cut [ and ] because 8-bit posts to mailing lists may not get
    # hit otherwise. e.g.: Subject: [ILUG] XǯuX .  Also cut
    # *, since mail that goes through spamassassin multiple times will
    # not be tagged on the second pass otherwise.
    s/[\[\]\* ]//g;

    push @{$xtra->{M}}, $_;
    return are_more_high_bits_set($_);
}

sub are_more_high_bits_set {
    local($_) = shift;
    return unless defined;
    my $numhis = tr/\200-\377//;
    my $numlos = length($_) - $numhis;

    ($numlos <= $numhis && $numhis > 3);
}

sub check_for_missing_to_header {
    my ($eng, $xtra, $hdrs) = @_;
    return not (exists $hdrs->{to} or exists $hdrs->{'apparently-to'});
}

sub check_for_missing_header {
    my ($eng, $xtra, $hdrs, $tag) = @_;
    return not exists $hdrs->{lc $tag};
}

sub check_for_missing_body {
    my ($eng, $xtra, $msg) = @_;
    # This test is slightly different in 4.0 -- we seem to miss the Jewellery
    # spam if we simply check for length == 0. Instead, if the message is very
    # small (under 10 bytes) we'll just check if it's entirely whitespace.
    my $sz = $msg->get_body_size;
    if ($sz < 10) {
        my $buf = $msg->get_body(0, $sz);
        return 1 unless $buf =~ /\S/;
    }
    0;
}

###########################################################################

sub check_lots_of_cc_lines {
    my ($eng, $xtra, $hdrs) = @_;
    local $_ = h_get('Cc', $hdrs);
    my @count = /\n/gs;
    if ($#count > 20) { return 1; }
    return 0;
}

###########################################################################

sub check_for_forged_gw05_received_headers {
    my ($eng, $xtra, $hdrs) = @_;

    local $_ = h_get('Received', $hdrs);

    # e.g.
    # Received: from mail3.icytundra.com by gw05 with ESMTP; Thu, 21 Jun 2001 02:28:32 -0400
    my ($h1, $h2) = (
          m/\nfrom\s(\S+)\sby\s(\S+)\swith\sESMTP\;\s+\S\S\S,\s+\d+\s+\S\S\S\s+
                          \d{4}\s+\d\d:\d\d:\d\d\s+[-+]*\d{4}\n$/xs);

    if (defined ($h1) && defined ($h2) && $h2 !~ /\./) {
        return 1;
    }

    0;
}

###########################################################################
sub check_for_faraway_charset {
    my ($eng, $xtra, $msg) = @_;
    my $type;
    $msg->enumerate_headers('Content-Type', \&_get1, \$type);
    my @locales = get_my_locales($eng);
    return 0 if grep { $_ eq "all" } @locales;

    $type = get_charset_from_ct_line($type);
    if (defined $type && !is_charset_ok_for_locales($type, @locales))
    {
        my $data = $msg->get_decoded_body(0, 2048);
        if (are_more_high_bits_set($data)) {
            push @{$xtra->{M}}, $type;
            return 1;
        }
    }

    0;
}

sub check_for_faraway_charset_in_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my @locales = get_my_locales($eng);
    return 0 if grep { $_ eq "all" } @locales;
    for my $h (qw(from subject)) {
        next unless $hdrs->{$h};
        my $hdr = join(' ', @{$hdrs->{$h}});
        while ($hdr =~ /=\?(.+?)\?.\?.*?\?=/g) {
            my $type = $1;
            unless (is_charset_ok_for_locales($type, @locales)) {
                push @{$xtra->{M}}, $type;
                return 1;
            }
        }
    }
    0;
}

sub html_charset_faraway {
    my ($eng, $xtra) = @_;
    my @locales = get_my_locales($eng);
    return 0 if grep { $_ eq 'all' } @locales;
    my $x = $xtra->{P}{spamassassin}{html} ||= {};
    return 0 if !exists $x->{'charsets'};

    my $okay=0;
    my $bad=0;

    for my $charset (@{$x->{'charsets'}}) {
        if (is_charset_ok_for_locales($charset, @locales)) {
            ++$okay;
        } else {
            push @{$xtra->{M}}, $charset;
            ++$bad;
        }
    }
    return ($bad && ($bad >= $okay));
}


sub html_imgspam {
    my ($eng, $xtra, $msg, $type, $min, $max) = @_;

    my $x = $xtra->{P}{spamassassin}{html} ||= {};

    return 0 unless defined $type;
    return 0 unless defined $min;
    return 0 unless exists $x->{'imgspam'};
    return 0 unless exists $x->{'imgspam'}{$type};

    my $count = $x->{'imgspam'}{$type};
    return 0 unless defined $count;

    if (defined $max) {
        return ($count >= $min and $count <= $max);
    }
    return ($count >= $min);
}

sub count_content_type_headers {
    my ($data, $part) = @_;
    my $ct;
    $part->enumerate_headers('content-type', \&_get1, \$ct);
    if (defined $ct) {
        _interp_content_type($ct, $data, $part);
    }
    return 1;
}

sub _interp_content_type {
    my ($ct, $data, $part) = @_;
    ++${$data->[2]};
    if ($ct =~ /html/i) {
        my $cd;
        $part->enumerate_headers('Content-disposition', \&_get1, \$cd);
        ++${$data->[0]} unless $cd && (index(lc $cd, 'inline')==-1);
    } elsif ($ct =~ /text/i) {
        ++${$data->[1]};
    } elsif ($ct =~ /multipart/i) {
        --${$data->[2]};
        $part->enumerate_parts(
            \&count_content_type_headers,
            $data
        );
    }
    if ($ct =~ /charset="?([^"]{1,128})"?/i) {
        push @{$data->[3]}, $1;
    }
}

sub _parse_part_content_headers {
    my ($eng, $xtra, $msg) = @_;

    my $x = $xtra->{P}{spamassassin}{mime} ||= {};
    return if exists $x->{'html'};

    my $html = 0;
    my $text = 0;
    my $total = 0;
    $x->{'charsets'} ||= [];

    my $ct = 'text/plain';
    $msg->enumerate_headers('content-type', \&_get1, \$ct);
    _interp_content_type($ct, [\$html, \$text, \$total, $x->{'charsets'}], $msg);

    $msg->enumerate_parts(
        \&count_content_type_headers,
        [\$html, \$text, \$total, $x->{'charsets'}]
    );

    $x->{'html'} = $html;
    $x->{'nonhtml'} = $total - $html;
    $x->{'text'} = $text;
    $x->{'nontext'} = $total - $text;
}

sub check_for_mime_html {
    my ($eng, $xtra, $msg) = @_;
    _parse_part_content_headers($eng, $xtra, $msg);

    my $x = $xtra->{P}{spamassassin}{mime};
    if ( $x->{'html'} > 0 ) { return 1; }
    return undef
}

sub check_for_invalid_charset {
    return undef unless $VALID_CHARSETS_DB;
    my ($eng, $xtra, $msg) = @_;

    _parse_part_content_headers($eng, $xtra, $msg);

    my $x = $xtra->{P}{spamassassin}{mime};
    for my $charset (@{$x->{'charsets'}}) {
        my $tmp = $charset;
        $tmp =~ s/;.*$//s;
        $tmp = lc $tmp;
        unless( $VALID_CHARSETS_DB->get($tmp) ) {
            push @{$xtra->{M}}, $charset, $tmp;
            return 1;
        }
    }
    return undef;
}

sub check_for_mime_html_only {
    my ($eng, $xtra, $msg) = @_;
    _parse_part_content_headers($eng, $xtra, $msg);

    my $x = $xtra->{P}{spamassassin}{mime};
    if ( ($x->{'html'} > 0) and ($x->{'nonhtml'} == 0) ) { return 1;}
    return undef;
}

sub check_for_mime_text_only {
    my ($eng, $xtra, $msg) = @_;
    _parse_part_content_headers($eng, $xtra, $msg);

    my $x = $xtra->{P}{spamassassin}{mime};
    return 1 if $x->{text} > 0 and $x->{nontext} == 0;
    undef;
}

sub check_for_mime_charset_faraway {
    my ($eng, $xtra, $msg) = @_;
    my @locales = get_my_locales($eng);
    return 0 if grep { $_ eq 'all' } @locales;
    _parse_part_content_headers($eng, $xtra, $msg);

    my $x = $xtra->{P}{spamassassin}{mime};

    for my $charset (@{$x->{'charsets'}}) {
        if (!is_charset_ok_for_locales($charset, @locales)) {
            push @{$xtra->{M}}, $charset;
            return 1;
        }
    }
    return undef;
}

=cut

NOTE: spamassassin.pm needs to fill in $x->{$test} ('total_comment_ratio')
      before we can enable this.

sub html_range {
    my ($eng, $xtra, $msg, $test, $min, $max) = @_;
    my $x = $xtra->{P}{spamassassin}{html} ||= {};
    return 0 unless exists $x->{$test};

    $test = $x->{$test};

    # not all perls understand what "inf" means, so we need to do
    # non-numeric tests!  urg!
    if ( !defined $max || $max eq "inf" ) {
        return ( $test eq "inf" ) ? 1 : ($test > $min);
    }
    elsif ( $test eq "inf" ) {
        # $max < inf, so $test == inf means $test > $max
        return 0;
    }
    else {
        # if we get here everything should be a number
        return ($test > $min && $test <= $max);
    }
}

=cut

sub html_tag_exists {
    my ($eng, $xtra, $msg, $tag) = @_;
    my $x = $xtra->{P}{spamassassin}{html} ||= {};
    return 0 unless exists $x->{__existing_tags__};
    return exists $x->{__existing_tags__}{lc $tag};
}

sub get_charset_from_ct_line {
    my $type = shift or return;
    if ($type =~ /charset=\"([^\"]+)\"/i) { return $1; }
    if ($type =~ /charset=\'([^\']+)\'/i) { return $1; }
    if ($type =~ /charset=(\S+)/i) { return $1; }
    return undef;
}

sub get_my_locales {
    my ($eng) = @_;
    my @locales;
    $eng->enumerate_attributes('plugin.re.ok-locales', \&_getall, \@locales);
    @locales = grep defined, @locales;
    return @locales if @locales;

    $eng->enumerate_attributes('plugin.spamassassin.ok-locales',
                               \&_getall, \@locales);
    @locales = grep defined, @locales;
    return @locales;
}

###########################################################################

sub check_for_shifted_date {
    my ($eng, $xtra, $hdrs, $min, $max) = @_;

    if (!exists $xtra->{P}{spamassassin}{date_diff}) {
        _check_date_diff($eng, $xtra, $hdrs);
    }
    return (($min eq 'undef' || $xtra->{P}{spamassassin}{date_diff} >= (3600 * $min)) &&
            ($max eq 'undef' || $xtra->{P}{spamassassin}{date_diff} <  (3600 * $max)));
}

sub _check_date_diff {
    my ($eng, $x, $h) = @_;
    $x->{P}{spamassassin}{date_diff} = 0;

    my $rcvd = h_get('Received', $h) or return;

    # a Resent-Date: header takes precedence over any Date: header
    my $date = h_get('Resent-Date', $h) || h_get('Date', $h) or return;
    chomp ($date);

    my $time = _parse_rfc822_date($eng, $date);
    # parse_rfc822_date failed
    return if !defined($time);

    # use second date. otherwise fetchmail Received: hdrs will screw it up
    my @rcvddatestrs = ($rcvd =~ /\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+/g);
    my @rcvddates = ();
    foreach $rcvd (@rcvddatestrs) {
        $eng->log("DEBUG", "trying Received header date for real time: $rcvd");
        $rcvd = _parse_rfc822_date($eng, $rcvd);
        push (@rcvddates, $rcvd) if defined $rcvd;
    }

    unless (@rcvddates) {
        $eng->log("DEBUG", "no dates found in Received headers, not raising flag");
        return;
    }

    my @diffs;

    foreach $rcvd (@rcvddates) {
        my $diff = $time - $rcvd;
        $eng->log("DEBUG", "time_t from date=$time, rcvd=$rcvd, diff=$diff");
        push(@diffs, $diff);
    }

    # if the last Received: header has no difference, then we choose to
    # exclude it
    if ($#diffs > 0 && $diffs[$#diffs] == 0) {
        pop(@diffs);
    }

    # use the date with the smallest absolute difference
    # (experimentally, this results in the fewest false positives)
    @diffs = sort { abs($a) <=> abs($b) } @diffs;
    $x->{P}{spamassassin}{date_diff} = $diffs[0];
}

sub _parse_rfc822_date {
    my ($eng, $date) = @_;
    my ($yyyy, $mmm, $dd, $hh, $mm, $ss, $mon, $tzoff);

    # make it a bit easier to match
    local $_ = " $date "; s/, */ /gs; s/\s+/ /gs;

    # now match it in parts.  Date part first:
    if (s/ (\d+) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{4}) / /i) {
        $dd = $1; $mon = $2; $yyyy = $3;
    } elsif (s/ (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) +(\d+) \d+:\d+:\d+ (\d{4}) / /i) {
        $dd = $2; $mon = $1; $yyyy = $3;
    } elsif (s/ (\d+) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{2,3}) / /i) {
        $dd = $1; $mon = $2; $yyyy = $3;
    } else {
        $eng->log("DEBUG", "time cannot be parsed: $date");
        return undef;
    }

    # handle two and three digit dates as specified by RFC 2822
    if (defined $yyyy) {
        if (length($yyyy) == 2 && $yyyy < 50) {
            $yyyy += 2000;
        }
        elsif (length($yyyy) != 4) {
            # three digit years and two digit years with values between 50 and 99
            $yyyy += 1900;
        }
    }

    # hh:mm:ss
    if (s/ ([\d\s]\d):(\d\d)(:(\d\d))? / /) {
        $hh = $1; $mm = $2; $ss = $4 || 0;
    }

    # numeric timezones
    if (s/ ([-+]\d{4}) / /) {
        $tzoff = $1;
    }
    # UT, GMT, and North American timezones
    elsif (s/ (UT|GMT|[ECMP][DS]T) / /) {
        if    ($1 eq "UT"  || $1 eq "GMT") { $tzoff = "+0000"; }
        elsif ($1 eq "EDT")                { $tzoff = "-0400"; }
        elsif ($1 eq "EST" || $1 eq "CDT") { $tzoff = "-0500"; }
        elsif ($1 eq "CST" || $1 eq "MDT") { $tzoff = "-0600"; }
        elsif ($1 eq "MST" || $1 eq "PDT") { $tzoff = "-0700"; }
        elsif ($1 eq "PST")                { $tzoff = "-0800"; }
    }
    # all other timezones are considered equivalent to "-0000"
    $tzoff ||= '-0000';

    if (!defined $mmm && defined $mon) {
        my @months = qw(jan feb mar apr may jun jul aug sep oct nov dec);
        $mon = lc($mon);
        my $i; for ($i = 0; $i < 12; $i++) {
            if ($mon eq $months[$i]) { $mmm = $i+1; last; }
        }
    }

    $hh ||= 0; $mm ||= 0; $ss ||= 0; $dd ||= 0; $mmm ||= 0; $yyyy ||= 0;

    my $time;
    eval {              # could croak
        $time = timegm($ss, $mm, $hh, $dd, $mmm-1, $yyyy);
    };

    if ($@) {
      $eng->log("DEBUG", "time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss");
      return undef;
    }

    if ($tzoff =~ /([-+])(\d\d)(\d\d)$/)        # convert to seconds difference
    {
        $tzoff = (($2 * 60) + $3) * 60;
        if ($1 eq '-') {
            $time += $tzoff;
        } else {
            $time -= $tzoff;
        }
    }
    return $time;
}

sub subject_is_all_caps {
    my $subject = $_[2];
    $subject =~ s/^\s+//;
    $subject =~ s/\s+$//;
    return 0 if $subject !~ /\s/;       # don't match one word subjects
    return 0 if (length $subject < 10); # don't match short subjects
    $subject =~ s/[^a-zA-Z]//g;         # only look at letters
    push @{$_[1]{M}}, $subject;
    return length($subject) && ($subject eq uc($subject));
}

###########################################################################

sub message_from_bugzilla {
    my ($eng, $xtra, $all) = @_;

    # Let's look for a Bugzilla Subject...
    if ($all =~ /^Subject: [^\n]{0,10}\[Bug \d+\] /m && (
            # ... in combination with either a Bugzilla message header...
            $all =~ /^X-Bugzilla-[A-Z][a-z]+: /m ||
            # ... or sender.
            $all =~ /^From: bugzilla/mi
        ))
    {
        push @{$xtra->{M}}, substr($all, $-[0], $+[0]-$-[0]);
        return 1;
    }

    return 0;
}

sub message_from_debian_bts {
    my $eng  = shift;
    my $xtra = shift;
    for my $all (@_) { # full headers
        # This is the main case; A X-<Project>-PR-Message header exists and
        # the Subject looks "buggy". Watch out: The DBTS is used not only by
        # Debian but by other <Project>s, eg. KDE, too.
        if ($all    =~ /^X-[A-Za-z0-9]+-PR-Message: [a-z-]+ \d+$/m &&
            $all    =~ /^Subject: Bug#\d+: /m) {
            push @{$xtra->{M}}, substr($all, $-[0], $+[0]-$-[0]);
            return 1;
        }
        # Sometimes the DBTS sends out messages which don't include the X-
        # header.  In this case we look if the message is From a DBTS account
        # and Subject and Message-Id look good.
        elsif ($all =~ /^From: owner\@/mi &&
               $all =~ /^Subject: Processed(?: \([^)]+\))?: /m &&
               $all =~ /^Message-ID: <handler\./m) {
            push @{$xtra->{M}}, substr($all, $-[0], $+[0]-$-[0]);
            return 1;
        }
    }
    return 0;
}

###########################################################################

# This test was originally based on RFC 2369 compliance.  However, the
# Mailing-List header was added so that Yahoo!Groups mails would not be
# matched.
sub suspect_list_headers {
    my ($eng, $xtra, $hdrs) = @_;

    my @headers = ('List-Help', 'List-Subscribe', 'List-Unsubscribe',
                   'List-Post', 'List-Owner', 'List-Archive',
                   'Mailing-List');
    my %count;

    foreach my $header (@headers) {
        my $h = $hdrs->{lc $header};
        $count{$header} += @$h if $h;
    }

    # without RFC 2369 fields
    return 0 unless %count;

    # header appears more than once (violates RFC 2369)
    foreach my $count (values %count) {
        return 1 if $count > 1;
    }

    # List-Help field is used (good RFC 2369 practice)
    return 0 if exists $count{'List-Help'};

    # List-Unsubscribe without much else (spammer mind trick)
    return (exists $count{'List-Unsubscribe'} && scalar keys %count < 2);
}


###########################################################################
# FULL-MESSAGE TESTS:
###########################################################################

sub check_for_uppercase {
    my ($eng, $xtra, $body, $min, $max) = @_;
    my $x = $xtra->{P}{spamassassin};

    if (exists $x->{uppercase}) {
        return (
            $x->{uppercase} > $min &&
            $x->{uppercase} <= $max
        );
    }

    # remove shift-JIS charset codes
    $body =~ s/\x1b\$B.*\x1b\(B//gs;

    # count numerals as lower case, otherwise 'date|mail' is spam
    my $lower = $body =~ tr/a-z0-9//d;
    my $upper = $body =~ tr/A-Z//;

    if (($upper + $lower) == 0) {
      $x->{uppercase} = 0;
    }
    else {
      $x->{uppercase} = ($upper / ($upper + $lower)) * 100;
    }

    return (
        $x->{uppercase} > $min &&
        $x->{uppercase} <= $max
    );
}

sub check_for_yelling {
    my ($eng, $xtra) = @_;
    my $x = $xtra->{P}{spamassassin};

    return $x->{num_yelling_lines} if defined $x->{num_yelling_lines};

    # Make local copy of lines in the body that have some non-letters
    my @lines = grep /[^A-Za-z]/, split /^/, $_[2];

    # Try to eliminate lines which might be newsletter section headers,
    # which are often in all caps; we do this by removing most lines
    # that start with whitespace.  However, some spam will match
    # this as well, so keep lines which have "!" or "$$" (spam often
    # has a yelling line indent with spaces, but surround by dollar
    # signs), or a "." which appears to end a sentence.
    @lines = grep(/^\S|!|\$\$|\.(?:\s|$)/, @lines);

    for (@lines) {
        # Get rid of everything but upper AND lower case letters
        tr/A-Za-z \t//cd;

        # Remove leading and trailing whitespace
        s/^\s+//;
        s/\s+$//;
    }

    # Now that we have a mixture of upper and lower case, see if it's
    # 1) All upper case
    # 2) 20 or more characters in length
    # 3) Has at least one whitespace in it; we don't want to catch things
    #    like lines of genetic data ("...AGTAGC...")

    @lines = grep { /^[A-Z\s]{20,}$/ && /\s/ } @lines;
    @{$xtra->{M}} = @lines;
    $x->{num_yelling_lines} = @lines;
    return @lines > 0;
}

sub check_for_num_yelling_lines {
    my ($eng, $xtra, undef, $threshold) = @_;
    return ($xtra->{P}{spamassassin}{num_yelling_lines} >= $threshold);
}

sub check_signature {
    my ($eng, $xtra, undef, $min, $max, $blank) = @_;

    if (!exists $xtra->{P}{spamassassin}{signature_lines}) {
        _check_signature($xtra->{P}{spamassassin}, $_[2]);
    }
    return (($xtra->{P}{spamassassin}{signature_lines} >= $min) &&
            ($xtra->{P}{spamassassin}{signature_lines} <= $max) &&
            ($xtra->{P}{spamassassin}{signature_blank} == $blank));
}


sub _check_signature {
  my $x = shift;

  $x->{signature_blank} = 0;
  $x->{signature_lines} = 0;

  # signature must follow one non-whitespace character
  if ($_[0] =~ /\S\s*\n--\s*\n((.*\n){1,15}?)\s*\Z/m) {
    my $signature = $1;
    if ($signature =~ /\n\s*\n\s*\S/m) {
      $x->{signature_blank} = 1;
    }
    if ($signature =~ /\S/m) {
      $x->{signature_lines} = ($signature =~ tr/\n/\n/);
    }
  }
}

###########################################################################

sub check_for_fake_aol_relay_in_rcvd {
    local $_ = $_[2]; # received header

    # this is the hostname format used by AOL for their relays. Spammers love
    # forging it.  Don't make it more specific to match aol.com only, though --
    # there's another set of spammers who generate fake hostnames to go with
    # it!
    if (/ rly-[a-z][a-z]\d\d\./i) {
        return 0 if /\/AOL-\d+\.\d+\.\d+\)/;    # via AOL mail relay
        return 0 if /ESMTP id (?:RELAY|MAILRELAY|MAILIN)/; # AOLish
        push @{$_[1]{M}}, $_[2];
        return 1;
    }

# spam: Received: from unknown (HELO mta05bw.bigpond.com) (80.71.176.130) by
#    rly-xw01.mx.aol.com with QMQP; Sat, 15 Jun 2002 23:37:16 -0000

# non: Received: from  rly-xj02.mx.aol.com (rly-xj02.mail.aol.com [172.20.116.39]) by
#    omr-r05.mx.aol.com (v83.35) with ESMTP id RELAYIN7-0501132011; Wed, 01
#    May 2002 13:20:11 -0400

    return 0;
}

###########################################################################

sub check_for_to_in_subject {
    my ($eng, $xtra, $hdrs, $check) = @_;
    $check ||= 1;

    my $to = h_get('To:addr', $hdrs);
    return 0 unless $to; # no To:?
    $to =~ s/\@.*$//; # just the username please
    push @{$xtra->{M}}, $to;

    my $subject = h_get('Subject', $hdrs);

    return 1 if $check == 1 && $subject =~ /^\s*\Q$to\E,/;
    return 1 if $check == 2 && $subject =~ /^\s*\Q$to\E,/i;
    return 1 if $check == 3 && $subject =~ /^\s*\Q$to\E,\S/;
    return 1 if $check == 4 && $subject =~ /^\s*\Q$to\E,\S/i;
    return 1 if $check == 5 && $subject =~ /\b\Q$to\E\b/;
    return 1 if $check == 6 && $subject =~ /\b\Q$to\E\b/i;

    return 0;
}

sub check_messageid_not_usable {
    my ($eng, $xtra, $hdrs) = @_;
    local ($_);

    # Lyris eats message-ids.  also some ezmlm, I think :(
    $_ = h_get ("List-Unsubscribe");
    if (/<mailto:(?:leave-\S+|\S+-unsubscribe)\@\S+>$/) {
        push @{$xtra->{M}}, substr($_, $-[0], $+[0]-$-[0]);
        return 1;
    }

    # ezmlm again
    if(gated_through_received_hdr_remover($hdrs)) { return 1; }

    # Allen notes this as 'Wacky sendmail version?'
    $_ = h_get ("Received");
    if (/\/CWT\/DCE\)/) {
        push @{$xtra->{M}}, substr($_, $-[0], $+[0]-$-[0]);
        return 1;
    }

    # Apr  2 2003 jm: iPlanet rewrites lots of stuff, including Message-IDs
    if (/iPlanet Messaging Server/) {
        push @{$xtra->{M}}, "iPlanet Messaging Server";
        return 1;
    }

    # too old; older versions of clients used different formats
    # DmitryS: this is for checking old corpus mail - we don't need to support it yet
    # return 1 if (received_within_months('6','undef'));

    return 0;
}

###########################################################################

# valid Outlookish Message-Ids contain the top word of the system time
# when the message was sent!
# We can verify this, by decoding the Date header, extracting
# the time token from the Message-Id, and comparing them.
#
sub check_outlook_timestamp_token {
    my ($eng, $xtra, $hdrs) = @_;
    local ($_);

    # handle lists which move the real message-id to another
    # header for resending.
    my $id = join ("\n", grep { defined($_) && length($_) > 0 }
                h_get('X-Message-Id', $hdrs),
                h_get('Resent-Message-Id', $hdrs),
                h_get('Message-Id', $hdrs),
    );

    return 0 unless ($id =~ /^<[0-9a-f]{4}([0-9a-f]{8})\$[0-9a-f]{8}\$[0-9a-f]{8}\@/);

    my $timetoken = hex($1);

    # convert UNIX time_t to Windows FILETIME.  From MSDN:
    #
    #     LONGLONG ll = Int32x32To64(t, 10000000) + 116444736000000000;
    #     pft->dwLowDateTime = (DWORD) ll;
    #     pft->dwHighDateTime = ll >>32;
    #
    # IOW, ((tt * a) + b) / c = id .
    # Now to avoid using any kind of LONGLONG data type, we do this:
    #     => tt * (a/c) + (b/c) = id
    #     let x = (a/c) = 0.0023283064365387
    #     let y = (b/c) = 27111902.8329849
    #
    my $x = 0.0023283064365387;
    my $y = 27111902.8329849;

    # quite generous, but we just want to be in the right ballpark, so we
    # can handle mostly-correct values OK, but catch random strings.
    my $fudge = 250;

    $_ = h_get('Date', $hdrs);
    $_ = _parse_rfc822_date($eng, $_); $_ ||= 0;
    my $expected = int (($_ * $x) + $y);
    my $diff = $timetoken - $expected;

    if (abs ($diff) < $fudge) { return 0; }

    # also try last date in Received header, Date could have been rewritten
    $_ = h_get('Received', $hdrs);
    /(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+).*?$/;

    $_ = _parse_rfc822_date($eng, $_); $_ ||= 0;
    $expected = int (($_ * $x) + $y);
    $diff = $timetoken - $expected;

    if (abs ($diff) < $fudge) { return 0; }

    return 1;
}

sub __check_phishing_url {
    my $eng = shift;
    my $xtra = shift;
    my $matched = 0;
    $xtra->{P}{spamassassin}{gone_phishing} = [];
    for my $chunk (@_) {
        while ( $chunk =~ /<\s{0,9}a
              [^>]{0,99}
              href\s{0,9}=\s{0,9}(\"?(?:https?|ftp):\/\/
              ([^\/\">]{1,999})
              [^>]{0,999})>
              \s{0,9}(?=[hwf])(?:https?:\/\/|www\.|ftp\.)
              ([^><\/]{1,999})
            /ixg
            ) {
            my ($href, $a, $b) = ($1, lc $2, lc $3);
            my $oddbits = $href =~ /[\x00-\x19\x80\xff]/;
            # strip user/pass
            my $userpass = $a =~ s/^.*@//;
            if ($b && $b =~ /(?:\w+\.)+\w/ && index($b,$a) == -1) {
                push @{$xtra->{P}{spamassassin}{gone_phishing}},
                    [$a, $b, $a =~ /(?:\d{1,3}\.){3}\d{1,3}/ ? 1 : 0, 0,
                    $userpass || $oddbits, length $href];
                $matched = 1;
            }
        }
    }
    if ($matched) {
        if($PHISHING_TARGETS_DB) {
            $PHISHING_TARGETS_DB->enumerate(
                sub {
                    my ($xtra, $phish) = @_;
                    for my $data (@{$xtra->{P}{spamassassin}{gone_phishing}}) {
                        if (index($data->[1], $phish)!=-1) {
                            $data->[3] = 1;
                        }
                    }
                    return 1; # continue
                },
                $xtra
            );
        }
    }

    return 0;
}

sub check_phishing_from_and_iphref {
    return unless $PHISHING_TARGETS_DB;
    my ($eng, $xtra, $msg) = @_;
    my $from;
    $msg->enumerate_headers('from', \&_get1, \$from);
    return unless $from;
    my $matched = 0;
    $from = lc $from;
    $PHISHING_TARGETS_DB->enumerate(
        sub {
            my ($data, $phish) = @_;
            if (index($data->[0], $phish) != -1) {
                ${$data->[1]} = 1;
                return 0;
            }
            return 1;
        },
        [$from, \$matched]
    );
    return unless $matched;
    # from has a phished string in it
    my $body = $msg->get_decoded_body();
    while ($body =~ /https?:\/\/([^\/\"?>]+)/ig) {
        my $dest = $1;
        # allow digits, '.', 'a'-'f', 'x', and ':'
        # the 'x' and 'a'-'f' allow hex, the ':' allows ports
        next unless $dest =~ /^[\d\.abcdefx:]+$/;
        $dest =~ s/:.*$//; # strip port
        next unless is_ipv4($dest);
        # body contains an http://ip_addr
        return 1;
    }
    return 0;
}

sub check_phishing_url_ip {
    my ($eng, $xtra) = @_;
    __check_phishing_url(@_) unless exists $xtra->{P}{spamassassin}{gone_phishing};
    my $matched = 0;
    for my $data (@{$xtra->{P}{spamassassin}{gone_phishing}}) {
        next unless !$data->[3] && $data->[2];
        # text contains a hostname that isn't a phishing target
        # href contains an IP address
        $matched = 1;
        push @{$xtra->{M}}, $data->[0] . " : " . $data->[1];
    }
    return $matched;
}

sub check_phishing_url_targetted_ip {
    my ($eng, $xtra) = @_;
    __check_phishing_url(@_) unless exists $xtra->{P}{spamassassin}{gone_phishing};
    my $matched = 0;
    for my $data (@{$xtra->{P}{spamassassin}{gone_phishing}}) {
        next unless $data->[3] && $data->[2];
        # text contains a phishing target
        # href contains an IP address
        $matched = 1;
        push @{$xtra->{M}}, $data->[0] . " : " . $data->[1];
    }
    return $matched;
}

sub check_phishing_url_targetted_host {
    my ($eng, $xtra) = @_;
    __check_phishing_url(@_) unless exists $xtra->{P}{spamassassin}{gone_phishing};
    my $matched = 0;
    for my $data (@{$xtra->{P}{spamassassin}{gone_phishing}}) {
        next unless $data->[3] && (!$data->[2]) && ($data->[4] || ($data->[5]>128));
        # href contains a hostname
        # text contains a phishing target
        # href contains oddbits, user/pass, or is longer than 128 bytes
        $matched = 1;
        push @{$xtra->{M}}, $data->[0] . " : " . $data->[1];
    }
    return $matched;
}

sub check_hashbuster_block_lots_of_five_letter_words {
    my ($eng, $xtra, undef) = @_;
    while ($_[2] =~ /\b((?:\w{5}\s{1,99}){14,99})\b/g) {
        my $match = $1;
        my @w = split /\s/, $match;
        my %w = map { ($_ => 1) } @w;
        if (int(keys(%w))>14) {
            push @{$xtra->{M}}, $match;
            return 1;
        }
    }
    return undef;
}

sub check_hashbuster_block {
    my ($eng, $xtra, undef) = @_;

    while ($_[2] =~ /
          \b                                 # word boundary
          (
            \w{0,32}                         # alphanumeric cruft
            [bcdfghjklmnpqrstvwx0-9]{6,9}    # at least six consonants or numbers
            \w{0,32}                         # alphanumbeic cruft
          )
          [ \t]{1,9}                         # spaces or tabs
          (
            \w{0,32}                         # same as above
            [bcdfghjklmnpqrstvwx0-9]{6,9}
            \w{0,32}
          )
          \b                                 # word boundary
        /gix)
    {
        my ($a, $b) = ($1, $2);
        if (defined $a &&
            defined $b &&
            !(
                $a =~ /^\d{2,99}\D{0,99}$/         or # digits + letters, or all digits
                $a =~ /^\D{1,99}\d{2,99}$/         or # letters + digits
                $a =~ /^(.)\1{1,99}$/              or # all the same character
                $b =~ /^\d{2,99}\D{0,99}$/         or
                $b =~ /^\D{1,99}\d{2,99}$/         or
                $b =~ /^(.)\1{1,99}$/              or
                ($a =~ /^(?:0[xX])?[A-F0-9]+$/ && $b =~ /^(?:0[xX])?[A-F0-9]+$/) or # both HEX
                ($a =~ /^(?:0[xX])?[a-f0-9]+$/ && $b =~ /^(?:0[xX])?[a-f0-9]+$/) or # both hex
                ($a !~ /[^A-Z]/ && $b !~ /[^A-Z]/)          # both all uppercase
            )
        ) {
            push @{$xtra->{M}}, $a, $b;
            return 1;
        }
    }
    0;
}

sub _extract_addresses {
    my ($data, $header, $value) = @_;
    $value =~ s/\([^)]*?\)//g; # strip comments
    for my $to (split /\s*,\s*/, $value) {
        push @$data, $to;
    }
    return 1;
}
sub check_for_spammy_tocc {
    my ($eng, $xtra, $msg) = @_;
    for my $hdr ('to', 'cc') {
        my $to = [];
        $msg->enumerate_headers( $hdr, \&_extract_addresses, $to );

        # at least 4 addresses
        next unless int(@$to)>3;

        my $domains = {};
        my $firstchars = {};
        for my $addr (@$to) {
            $firstchars->{substr($addr, 0, 1)}++;
            if ($addr =~ /\@(.*)$/) { $domains->{lc $1}++; }
        }

        # all to the same domain,
        next if (int(keys %$domains)!=1);

        # all begin with same first character
        next if (int(keys %$firstchars)!=1);

        return 1;
    }
    return 0;
}


sub clear_caches {
    _ciphered_to_email_in_uri_test_clear_cached_tr_subs();
}


{
my $cached_tr_subs = {};

sub _ciphered_to_email_in_uri_test_clear_cached_tr_subs {
    $cached_tr_subs = {};
}

sub ciphered_to_email_in_uri_test {
    my ($eng, $xtra, $msg, $tr_lhs, $tr_rhs) = @_;

    my $to;
    $msg->enumerate_headers('to', \&_get1, \$to);
    return 0 unless $to;

    my $to_email;

    if ( $to =~ /<([^>]{4,75})>/ ) {
        $to_email = $1;
    }
    else {
        $to_email = $to;
        $to_email =~ s/^\s+//;
        $to_email =~ s/\s+$//;
    }

    return 0 if $to_email eq "";

    my $cache_key = "$tr_lhs\x00$tr_rhs";
    unless (exists $cached_tr_subs->{$cache_key}) {
        eval <<"EOC";
\$cached_tr_subs->{\$cache_key} = sub { \$_[0] =~ tr/$tr_lhs/$tr_rhs/ };
EOC
        return 0 if $@;
    }

    $cached_tr_subs->{$cache_key}($to_email);

    my $uris = $xtra->{'P'}{'part'}{'uri'};

    for my $uri (@$uris) {
        if ( -1 != index($uri, $to_email) ) {
            return 1;
        }
    }

    return 0;
}

}

1;
