package ActiveState::SpamEngine::Plugin::sxl;

use strict;

use ActiveState::SpamEngine qw(PMX_TYPE_CALLBACK PMX_TYPE_STRING
                               PMX_TYPE_INTEGER
                               _get1 _getall);

use base qw(ActiveState::SpamEngine::Plugin);

# Detect if we're in SDK debug mode. We do this at compile time so that DEBUG
# gets optimized away most of the time.
BEGIN {
    if ($ENV{PMX_SDK_DEBUG_LOGFILE}) {
        eval "sub DEBUG() { 1 }";
    }
    else {
        eval "sub DEBUG() { 0 }";
    }
}

our @SP_REQUIRE = qw(net fur rip);
our @SP_ATTRIB = (
    # KEY                  TYPE                DEFAULT
    'hashed-uris',         PMX_TYPE_INTEGER,   undef,
    'enable',              PMX_TYPE_INTEGER,   undef,
    'max-response-time',   PMX_TYPE_INTEGER,   undef,
);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->declare_attributes(\@SP_ATTRIB, $name);
    $eng->subscribe($name, "core.compile", \&compile);
}

sub compile {
    my $eng = shift;

    # Check those engine attributes which would make using SXL pointless...
    for my $attr (qw(
        core.local-tests-only
        plugin.net.skip-dns-checks
    )) {
        my $skip;
        $eng->enumerate_attributes($attr, \&_get1, \$skip);
        return if $skip;
    }

    # Give the SXL object a chance to load its configuration
    my $enable_sxl = $eng->{_sxl}->compile($eng);
    return unless $enable_sxl;

    # The Net plugin subscribes to compile too, where it will create the
    # RequestManager we will need to use to be able to subscribe to the
    # net.finish event.  We therefore replace the default RequestManager in the
    # engine's SXL object with the freshly created RequestManager from the net
    # plugin.
    $eng->{_sxl}->set_dns_manager($ActiveState::SpamEngine::Plugin::net::DNS_MANAGER);

    $eng->subscribe('sxl', 'net.dispatch', \&dispatch);
    $eng->subscribe('sxl', 'net.finish', \&collate);

    $eng->initialised("sxl");
}


sub collate {
    my ($eng, $xtra, $rm) = @_;
#    my $x = $xtra->{P}{sxl} or return;
#    print STDERR "DEBUG:    found plugin specific data\n";

    # Want just any old SXL implementation...
    my $sxl = $eng->sxl();
    $sxl->collect_results($eng);
}

# This is an opportunity to dispatch any pending queries.  We'll choose not to
# send any if the spam probability is already more than 50%.
sub dispatch {
    my ($eng, $xtra) = @_;
    #print STDERR "DEBUG: sxl plugin: dispatch\n";

    if (my $sxl = $eng->sxl()) {

        my $corr = $xtra->{w}->get('??corr??') || 0.5;
        my $p = $xtra->{prob} = 1 / (1 + exp(-$xtra->{sum} * $corr));

        if ($p >= .5) {
            $eng->log("DEBUG", sprintf(qq[%s() probability==%s; spam, so clear any outstanding SXL queries\n], (caller(0))[3], $p)) if (DEBUG);
            $sxl->clear();
            $eng->statistics()->increment('sxl_early_out');
        } else {
            $eng->log("DEBUG", sprintf(qq[%s() probability==%s; send any outstanding SXL queries\n], (caller(0))[3], $p)) if (DEBUG);
            $sxl->dispatch($eng);
        }
    }
}

1;
