package ActiveState::SpamEngine::Plugin::uri;

use strict;
use ActiveState::SpamEngine qw(DEBUG _get1);
use Sophos::TLD;
use Sophos::Debug qw(_debug);

use base 'ActiveState::SpamEngine::Plugin';

our @SP_REQUIRE = qw(part sxl);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->subscribe($name, "core.compile", \&compile);
    1;
}

my ($db, $rules, $sxl, $sxlemail, $sxlns, $tld_db);

sub compile {
    my $eng = shift;

    eval {
        my $temp;
        $eng->enumerate_attributes("plugin.uri.advert-url", \&_get1, \$temp);
        if ($temp) {
            $db = $temp;
        } else {
            $db = $eng->nested_db("uri.db");
        }
        $rules = $eng->nested_db("uri.rules");
    };
    return if $@;

    $sxl      = $eng->sxl('uri');
    $sxlemail = $eng->sxl('uriemail');
    $sxlns    = $eng->sxl('body_uri_ns');

    if (!defined $tld_db) {
        $tld_db = Sophos::TLD->new();
    }

    $eng->subscribe("uri", 'part.uri:from', \&uri_from);
    $eng->subscribe("uri", 'part.uri', \&uri);
    $eng->initialised("uri");
}

sub sxl_hit {
    my ($x, $query, $result) = @_;
    my ($engine, $xtra) = @$x;
    $engine->log("DEBUG", sprintf(qq[%s(query='%s',result='%s')\n], (caller(0))[3], $query, $result)) if (DEBUG);
    $engine->addfeature($xtra, $result, $query);
}

# result == { request_error_rcode => 'str', timeout => bool, 
#             domains => ['ns1', 'ns2'], glue => ["1.2.3.4", "5.6.7.8"]}
#
sub uri_ns_result {
    my ($x, $query, $result) = @_;
    my ($engine, $xtra) = @$x;
    $engine->log("DEBUG", sprintf(qq[%s(uri_ns_result(query='%s',result={%s}],
                                  (caller(0))[3], $query, join(",", keys(%{$result})))
                 .($result->{domains} ? ",domains=[".join(",", @{$result->{domains}})."]" : "").")\n") if (DEBUG);
    
    if (defined $result->{request_error_rcode} && $result->{request_error_rcode}) {
        _debug("case: error rcode") if DEBUG;
        if($result->{request_error_rcode} =~ m~ERROR_MISSING_RCODE~) {
            # labs don't want feature __URI_NS_ERROR_MISSING_RCODE
        } else {
            $engine->addfeature($xtra, '__URI_NS_' . $result->{request_error_rcode});
        }

        return;
    }
    
    if ($result->{timeout}) {
        _debug("case: timeout") if DEBUG;
        $engine->addfeature($xtra, '__URI_NS_TIMEOUT');
        return;
    }

    if (!exists($result->{domains})) {
        _debug("case: request ok but also no domains") if DEBUG;
        $engine->addfeature($xtra, '__URI_NS_NOANSWER');
    }
    else {
        _debug("case: request ok processing %s domains", scalar(@{$result->{domains}})) if DEBUG;
        $engine->addfeature($xtra, '__URI_NS');
        
        if (!defined $db) {
            $engine->log("WARNING", "uri: no database, returning");
            return;
        }

        for my $nameserver (@{$result->{domains}}) {
            # If we get an empty nameserver, fire this special rule and DO NOT look up in uri.d
            #
            if ($nameserver eq "") {
                $engine->addfeature($xtra, '__URI_NS_ROOTLABEL');
                next;
            }
            
            my $nameserver_domain = $tld_db->single_label_plus_tld($nameserver);
            _debug("looking up domain '%s' in uri.d (extracted from ns '%s')", $nameserver_domain, $nameserver) if DEBUG;

            my $matched = $db->get($nameserver_domain);

            if ($matched) {
                $engine->log("DEBUG", "uri: ns: found $matched nameserver $nameserver") if DEBUG;
                $engine->addfeature($xtra, "URI_".uc($matched)."_NS_NAME", "[$query]=>[$nameserver]");
            }
        }
    }
 
    if (defined($result->{glue})) {
        _debug("case: request ok processing %s glue records", scalar(keys(%{$result->{glue}}))) if DEBUG;
        
        if (!defined($db)) {
            $engine->log("WARNING", "uri: no database, returning");
            return;
        }

        for my $glue_name (keys(%{$result->{glue}})) {
            my $glue_ip = $result->{glue}->{$glue_name};

            _debug("looking up ip '%s' in uri.d (extracted from ns '%s')", 
                   $glue_ip, $glue_name) if DEBUG;
            my $matched = $db->get($glue_ip);

            if ($matched) {
                $engine->log("DEBUG", "uri: ns: found $matched nameserver ip $glue_ip") if DEBUG;
                $engine->addfeature($xtra, "URI_".uc($matched)."_NS_IP", "[$query]=>[$glue_name]=>[$glue_ip]");
            }
        }
    }
}

# Called back for body URIs
#
sub uri
{
    uri_shared('body', @_);
}

# Called back for from line URIs
#
sub uri_from
{
    uri_shared('from', @_);
}

sub uri_shared {
    my $type = shift;
    my $eng  = shift;
    my $xtra = shift;

    unless (defined $db) {
        $eng->log("WARNING", "uri: no database, returning");
        return;
    }

    if (DEBUG) {
        $eng->profile(3, "uri: known URI check starting");
        eval {
            require Data::Dump;
            $eng->log("DEBUG", "uri: URIs to search for are:\n".Data::Dump::dump(\@_)."\n");
        };
    }

    # Now do the actual searching in the DB
  URI:
    for my $uri_ref (@_) {
        # IMPORTANT! We must make a copy here, or any modifications we make to
        # uri will modify the part list itself
        my $uri = lc $uri_ref;
        my $scheme = 'http';
        if ($uri =~ /^(\w+):\/\/(.*)$/) {
            $scheme = $1;
            $uri = $2;
        }
        elsif ($uri =~ /^(mailto|javascript):/) {
            $scheme = $1;
        }

        if ($scheme eq 'mailto') {
            my $matched = $db->get("$uri");

            if ($matched) {
                $eng->log("DEBUG", "uri: found $matched mailto $uri") if DEBUG;
                $eng->addfeature($xtra, $rules->get("${matched}:mailto"), $uri);
                last;
            }

            #
            # SXL query for mailto?
            #

            (my $wild = $uri) =~ s/^mailto:[^@]+\@/mailto:\@/;
            $matched = $db->get($wild);

            if ($matched) {
                $eng->log("DEBUG", "uri: found $matched mailto $uri") if DEBUG;
                $eng->addfeature($xtra, $rules->get("${matched}:mailto"), $uri);
                last;
            }

            #
            # SXL query for hostname part of mailto
            #

            (my $dekrufted = $uri) =~ s/[^\w\@\-\.\:]//g;
            $matched = $db->get($dekrufted);

            if ($matched) {
                $eng->log("DEBUG", "uri: found $matched mailto $uri") if DEBUG;
                $eng->addfeature($xtra, $rules->get("${matched}:mailto"), $uri);
                last;
            }

            #
            # SXL query for de-crufted part of mailto
            #

            (my $dekrufted_wild = $wild) =~ s/^mailto:[^@]+\@/mailto:@/;
            $matched = $db->get($dekrufted_wild);

            if ($matched) {
                $eng->log("DEBUG", "uri: found $matched mailto $uri") if DEBUG;
                $eng->addfeature($xtra, $rules->get("${matched}:mailto"), $uri);
                last;
            }

            # 
            # SXL query for decrufted part of mailto following '@'
            #
            
            if ($sxlemail)
            {
                (my $domain = $uri) =~ s/^.*\@//;
                $sxlemail->get_uri($domain, \&sxl_hit, [$eng, $xtra]);
            }

            # Don't fire KNOWN_ADVERT_DOMAIN from email addresses
            next;
        }
        else {
            my ($username, $passwd, $host, $port, $path);
            my $matched;
            if ($uri =~ /^([^\\\?\@\:]+):([^\@]+)\@([\w\.-]+)((?::\d+)?)([\?\/].*)?$/) {
                ($username, $passwd, $host, $port, $path) = ($1, $2, $3, $4, $5);
            }
            elsif ($uri =~ /^([^\\\?\@]+)\@([\w\.-]+)((?::\d+)?)([\?\/].*)?$/) {
                ($username, $host, $port, $path) = ($1, $2, $3, $4);
            }
            elsif ($uri =~ /^([\w\.-]+)((?::\d+)?)([\?\/].*)?$/) {
                ($host, $port, $path) = ($1, $2, $3);
            }
            else {
                $host = $uri;
            }

            $path =~ s/\/$// if $path;

            my @path_list = ();
            @path_list = split /([\/\?])/, $path if $path;
            # print "XXX\nhost: $host\nusername: $username\npassword: $passwd\nport: $port\npath: $path\n";

            # Try uris with successively fewer leading hostnames...
            my @host_list = split /\./, $host;
            while (@host_list > 1) {
                my $domain = join('.', @host_list);
                if ($matched = $db->get($domain)) {
                    $eng->log("DEBUG", "uri: found $matched domain $domain") if DEBUG;
                    $eng->addfeature($xtra, $rules->get("${matched}:domain"), $domain);
                    last URI;
                }

                # ...with and without a port
                if ($port) {
                    if ($matched = $db->get($domain . $port)) {
                        $eng->log("DEBUG", "uri: found $matched domain $domain$port") if DEBUG;
                        $eng->addfeature($xtra, $rules->get("${matched}:domain"), "$domain$port");
                        last URI;
                    }
                }

                # ...and with successively fewer trailing pathnames
                my @path_list_copy = @path_list; # copy @p locally
                while (@path_list_copy) {
                    my $url = join('', $domain, @path_list_copy);
                    if ($matched = $db->get($url)) {
                        $eng->log("DEBUG", "uri: found $matched url $url") if DEBUG;
                        $eng->addfeature($xtra, $rules->get("${matched}:url"), $url);
                        last URI;
                    }
                    # we pop twice because we are saving the divider (/, ? etc)
                    pop @path_list_copy;
                    pop @path_list_copy;
                }
                shift @host_list;
            }

            if ($matched = $db->get($uri)) {
                $eng->log("DEBUG", "uri: found $matched url $uri") if DEBUG;
                $eng->addfeature($xtra, $rules->get("${matched}:url"), $uri);
                last;
            }

            #
            # SXL lookup
            #
            $sxl->get_uri($host, \&sxl_hit, [$eng, $xtra]);
            
            #
            # NS lookup (using Sophos::SXL) for URLs in the email body only
            #
            if ($type eq 'body') {
                my $label_re = qr/(?:\w{1,2}|\w[\w-]{1,61}\w)/;
                if ($host =~ /^(?:$label_re\.){1,19}$label_re$/o) {
                    $sxlns->get_uri_ns($tld_db->single_label_plus_tld($host), \&uri_ns_result, [$eng, $xtra]);
                }
                elsif (DEBUG) {
                    _debug("domain '%s' is invalid, not doing an NS lookup on it", $host);
                }
            }
        }
    }
    $eng->profile(3, "uri: known URI check finished") if DEBUG;
}

1;
