package ActiveState::SpamEngine::Plugin::word_obfu;

use strict;
use ActiveState::SpamEngine qw(PMX_TYPE_CALLBACK PMX_TYPE_INTEGER
                               _get1);
use ActiveState::ODE qw(ode_match);

use base 'ActiveState::SpamEngine::Plugin';

our @SP_REQUIRE = qw(part);

sub plugin {
    my ($eng, $name) = @_;
    for my $dep (@SP_REQUIRE) { $eng->load_plugin($dep); }
    $eng->subscribe($name, "core.compile", \&compile);
    1;
}

my ($db_words, $db_rules, $db_ode);

sub compile {
    my $eng = shift;
    eval {
        $db_words = $eng->nested_db("word_obfu.words");
        $db_rules = $eng->nested_db("word_obfu.rules");
        $db_ode   = $eng->nested_db("ode");
    };
    return if $@;

    $eng->subscribe("word_obfu", "part.b_joinlines", \&detect_wobfu);
    $eng->initialised("word_obfu");
    1;
}

sub escape {
    my $temp = $_[0];
    $temp =~ s/([\x00-\x1f\x80-\xff])/sprintf("\\x%02x",ord($1))/gex;
    return $temp;
}

sub detect_wobfu {
    my $eng = shift;
    my $xtra = shift;
    my $results = ode_match($eng->pcre, $db_ode->_asdb, $_[0]);

    my $matches = {};

    for my $data (@$results) {
        my $skip = 1;
        my ($word, $start, $len, $obfu, $logic, $sb, $eb, $context) = @$data;
        my $word_len = length($word);
        my $ex_logic = $logic;
        $logic =~ s/-//g;

        for(;;) {
            last if (!$sb || !$eb);

            last if ($word eq $obfu);

            # skips words that start with a '$'
            last if ($obfu =~ /^\$/);

            # skips words prefixed with a '$'
            last if ($start>0) && (substr($_[0], $start-1, 1) eq '$');

            last if ($obfu =~ /$word\z/);
            last if ($obfu =~ /\A$word/);

            # skip anything where we don't have a basic match in the first 3
            # characters or the last 3 characters
            last if ($logic =~ /^[^B]{3}/);
            last if ($logic =~ /[^B]{3}$/);

            # skip anything with IK or KI
            last if ($logic =~ /IK/);
            last if ($logic =~ /KI/);

            # skip anything without at least 3 basic matches
            last if $logic =~ tr/B// < 3;

            # skip three or more Kruft in a row
            last if $logic =~ /KKK/;

            # skip M.D.
            last if $obfu =~ /\A[A-Z]\.\s?[A-Z]\.\z/;

            # inserts only after at least one basic match
            last if $logic =~ /\A[^B]+I/;

            # skip anything with only a single space inserted
            last if $logic =~ /\AB+SB+\z/;

            # skip anything with only a single insert if the insert is
            # whitespace. (Not all whitespace matches as 'space' in ODE due to
            # newlines causing FPs)
            if ($logic =~ /\A(B+)IB+\z/) {
                my $skip_len = length ($1);
                my $inserted_char = substr($obfu, $skip_len, 1);
                last if $inserted_char =~ /\s/;
            }

            # skip two words split with [^B]S or S[^B]
            last if $logic =~ /\AB*[^B]SB*\z/;
            last if $logic =~ /\AB*S[^B]B*\z/;

            # skip matches that are all digits, or symbols commonly
            # found around digits
            last if $obfu =~ /\A[\d\.\:\$\)\(\-_\+\*\%]+\z/;

            if ($logic =~ /\AB+KB+\z/) {
                # skip contractions of known source words (e.g. Dick's or
                # Dick~Rs thanks to MS-Word)
                last if $obfu =~ /\A([^'\x92]+)['\x92]([^'\x92]+)\z/;

                # skip hyphenations of known source words (e.g. hard-core)
                last if $obfu =~ /\A([^-]+)-([^-]+)\z/;

            }

            # skip ...pus. Sy... (end of sentence followed by another word)
            last if $obfu =~ /[\.\,\!\?]\s/ && $logic =~ /\AB+[KP]SB+\z/;

            # skip if word and obfu are equal except for the ending
            # kruft, and the $obfu ends in a sentence seperator
            last if $logic =~ /\AB+[OPK]\z/ && $obfu =~ /[.,?!]$/;

            # obfu exists as a word
            last if $db_words->get(lc $obfu);

            $skip = 0;
            last;
        }

        $obfu = escape($obfu);
        $context = escape($context);
        if ($skip) {
            # $eng->log("DEBUG", "ode: $word written as '$obfu' [location=$start length=$len logic=$ex_logic sb=$sb eb=$eb context='$context']");
        }
        else {
            my $class = $db_words->get($word);

            push @{$matches->{$class}}, "$word: $obfu";
            $eng->log("DEBUG", "word_obfu: '$class/$word' written as '$obfu' [location=$start length=$len logic=$ex_logic sb=$sb eb=$eb] context='$context'");
        }
    }

    # Now decide what to fire
    for my $class (keys %$matches) {
        my $fired = $matches->{$class};
        for my $def (split ' ', $db_rules->get($class)) {
            my ($thresh, $rule) = split ':', $def;
            if (@$fired >= $thresh) {
                my $msg = join(", ", @$fired);
                $eng->addfeature($xtra, $rule, $msg);
                last;
            }
        }
    }

    1;
}

1;
