# Represents the data which will be returned to Sophos.

use strict;

package Sophos::Feedback::Report;

use Time::Local qw/timegm/;
use File::stat;
use Storable;

# API functions:
sub parse_report;
sub add_milter_file;
sub header;
sub traffix;
sub content;
sub size;

my @required_fields;

sub new {
    my ($this) = @_;
    my $class = ref $this || $this;
    return bless {
        sections => {},
        headers => {},
    }, $class;
}

# The format for this object is a *little* complicated...
#
# 'headers' contains a hash of name => values.  This is straightforward.
# 'sections' contains a hash of section hashes.  This is not.
# Each 'section' entry can contain either an array or a hash.  An array
# contains data which can't be numerically manipulated.
#   As a hash, there may be one of *three* values:
#    - A numeric value (RULE_NAME => '1')                 RULE_NAME=1 
#    - A string (DNS_ADDRESS => 'a=1 b=2 c=3'")           DNS_ADDRESS a=1 b=2 c=3
#    - A hash (DNS_ADDRESS => { a => 1, b => 2, c => 3 }) DNS_ADDRESS b=2 a=1 c=3

# Static function to create the correct format for time expected by a Sophos
# feedback report
sub time2pmx_time {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = gmtime($_[0]);
    return sprintf(
        "%04d-%02d-%02dT%02d:%02d:%02dZ",
         $year+1900,
         $mon+1,
         $mday,
         $hour,
         $min,
         $sec
    );
}

sub pmx_time2time {
    my ($pmxtime) = @_;
    return unless $pmxtime;
    my ($year, $month, $mday, $hour, $min, $sec) = ($pmxtime =~ m/(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)Z/);
    return undef unless $year;
    return timegm($sec, $min, $hour, $mday, $month-1, $year);
}

sub _breakup_stats {
    my ($element) = @_;
    return unless defined $$element;

    # Do nothing if $$element is numeric
    return $$element if _isNumeric($$element);
    # Do nothing unless $$element is a scalar
    return $$element if ref $$element;

    my @contents = split /\s+/, $$element;

    my $hash = {};

    for my $pair (@contents) {
        my ($a, $b) = split /=/, $pair, 2;
        $hash->{$a} = $b;
    }

    $$element = $hash;
    return $$element;
}

# Parse a server stats line which looks like this:
#   "0x12345678 a=1 b=2 c=3"
sub _process_server_stats
{
    my $self = shift;
    my $section = shift;
    my $line = shift;

    #print STDERR "DEBUG: _process_server_stats(section='$section', '$_')\n";

    return unless $line;

    # Remove leading and trailing white space
    $line =~ s/^\s*//;
    $line =~ s/\s*$//;

    if (exists $self->{'sections'}{"$section"}) {
        my $entry = $self->{'sections'}{"$section"};
        if (ref $entry eq 'ARRAY') {
            push @$entry, $line;
        } elsif (ref $entry eq 'HASH') {
            if ($line =~ m/^(\S+)(?:\s+|\s*=\s*)(.*)$/) {
                $entry->{"$1"} = $2;
            }
        }
    } else {
        if ($line =~ m/^\S+.*\s[^=]+(?:\s|$)/) {
            # Non-engine generated data, it gets output verbatim.
            $self->{'sections'}{"$section"} = [$line];
        } else {
            # Engine may produce reports with section data which may be one of:
            # 01234567 a=1 b=2 c=3
            # a=1
            if ($line =~ m/^(\S+)(?:\s+|\s*=\s*)(.*)$/) {
                $self->{'sections'}{"$section"}{"$1"} = $2;
            }
        }
    }
}
    
# NOTE: if $data is an ARRAY reference, then it will be modified.
sub parse_report {
    my ($self, $data) = @_;

    # Reset sections and headers
    $self->{headers} = {};
    $self->{sections} = {};

    my $headers = $self->{headers};
    my $sections = $self->{sections};

    if (!ref $data) {
        $data = [split /^/, $data];
    }

    my $report_version;

    # Check the first line of the report for the 'Report-Version' field
    my ($key, $value) = split /: /, $data->[0];
    if ($key eq 'Report-Version') {
        $report_version = $value;

        die "Unsupported report format $report_version" if $report_version > 3;

        shift @$data;
    }

    die "Report format unrecognised" unless defined $report_version;

    my $linenum = 1;
    while (@$data) {
        $_ = shift @$data;
        $linenum++;
        chomp;
        last if /^\s*$/;
        my ($key, $value) = split /: /, $_, 2;
        next unless defined $value;
        $value =~ s/\s*$//;
        $value =~ s/^\s*//;

        $headers->{$key} = $value;
    }

    if ($report_version == 2) {
        # VERSION 2 format - traffix data follows header
        # Now read traffix data
        while (@$data) {
            $_ = shift @$data;
            $linenum++;
            chomp;
            last if /^\s*$/;

            $self->_process_server_stats('traffic', $_);
        }
    } elsif ($report_version == 3) {
        while ($_ = shift @$data) {
            $linenum++;
            chomp;
            if (m/^\[(.*)\]\s*$/) {
                my $section = $1;

                while ($_ = shift @$data) {
                    $linenum++;
                    chomp;
                    last if /^\s*$/;

                    $self->_process_server_stats($section, $_);
                }
            } else {
                die "Invalid report format; no section provided at line $linenum: '$_'";
            }
        }
    }

    if (!defined $headers->{'Report-Time'}) {
        # Report-Time: 2007-03-13T16:02:49Z
        $headers->{'Report-Time'} = time2pmx_time(time);
    }
}

sub header {
    my ($self, $key) = @_;
    return $self->{headers}{$key};
}

sub section {
    my ($self, $section) = @_;
    my @contents = ();
    if (exists $self->{sections}{"$section"}) {
        # May be ARRAY or HASH
        my $hash = $self->{sections}{"$section"};
        if (!defined $hash || !ref $hash) {
            return [];
        } elsif (ref $hash eq 'ARRAY') {
            push @contents, @$hash;
        } elsif (ref $hash eq 'HASH') {
            # We expect this hash to contain something like:
            # {  12345678 => { a=> 1, b => undef, c => 3} }
            #     or
            # {  12345678 => 'a=1 b c=3' }
            #     or, just about the same...
            # {  12345678 => '1' }
            for my $ip (keys %$hash) {
                my $value = $hash->{"$ip"};
                if (ref $value) {
                    push @contents, "$ip " . join(' ', map {
                            defined $value->{$_} ? "$_=$value->{$_}" : $_
                            } keys %{$value});
                } else {
                    if (_isNumeric($value)) {
                        push @contents, "$ip=$value";
                    } else {
                        push @contents, "$ip $value";
                    }
                }
            }
        }
    }

    return \@contents;
}

sub content {
    my ($self) = @_;

    my @contents = ();
    push @contents, 'Report-Version: 3';

    for my $key (keys %{$self->{headers}}) {
        # Note: Report-Version was excluded from the headers.  If it appears
        # again, it is because it was present twice in the source document.
        push @contents, "$key: $self->{headers}{$key}";
    }

    for my $section (keys %{$self->{sections}}) {
        push @contents, '';
        push @contents, "[$section]";
        for my $entry (@{$self->section("$section")}) {
            push @contents, $entry;
        }
    }

    return join "\n", @contents;
}

sub size {
    my ($self) = @_;
    return length($self->content());
}

sub _find_element {
    my $self = shift;
    my $current = $self->{sections};

    while (@_) {
        my $key = shift;
        return \$current->{"$key"} unless @_;
        if (exists $current->{"$key"}) {
            _breakup_stats(\$current->{"$key"}) unless ref $current->{"$key"};
            return undef unless ref $current->{"$key"};
        } else {
            $current->{$key} = {};
        }
        $current = $current->{$key};
    }
    return \$current;
}

sub _isNumeric {
    my ($value) = @_;
    return 0 unless defined $value;
    return 0 if $value =~ m/^\s*$/;
    return 1 if $value =~ m/^(?:\d+|\d+\.\d+)$/;
    return 0;
}


sub increment {
    my $self = shift;
    return $self->add(@_, 1);
}

sub add {
    my $self = shift;
    my $increment = pop;
    my $entry = $self->_find_element(@_);
    return unless defined $entry;
    return if ref $$entry;
    if (_isNumeric($increment)) {
        if (defined $$entry) {
            if (_isNumeric($$entry)) {
                $$entry += $increment;
            }

# DEBUG:
#            else {
#                print STDERR "DEBUG: Attempting to add a numeric value to a non-numeric entry\n";
#            }
# DEBUG:

        } else {
            $$entry = $increment;
        }
    }
    return $$entry;
}

sub query {
    my $self = shift;
    my $entry = $self->_find_element(@_);
    return unless defined $entry;
    return if ref $$entry;
    return $$entry;
}

sub _convert_section_to_hash {
    my ($element) = @_;

    my $hash = {};

    for my $line (@$$element) {
            my ($key, $contents) = ($line =~ m/^(\S+)(?:\s+|\s*=\s*)(.*)$/);
            $hash->{"$key"} = $contents;
    }

    $$element = $hash;
}

sub _consolidate_hashes {
    my ($master, $to_add) = @_;
    return unless ref $to_add eq 'HASH';
    return unless ref $master eq 'HASH';

    for my $key (keys %$to_add) {

        if (!exists $master->{$key}) {
            # If there is no dest entry, don't bother exploding the entry
            $master->{$key} = $to_add->{$key};
            next;
        }

        # ASSERT master->{$key} exists

        if (ref $to_add->{$key} eq 'ARRAY' && ref $master->{$key}) {
            #print STDERR "DEBUG: converting sections to hashes\n";
            _convert_section_to_hash(\$to_add->{$key});
        }

        if (!ref $to_add->{$key} && ($to_add->{$key} || '') =~ m/^\S+=/) {
            _breakup_stats(\$to_add->{$key});
        }

        if(ref $to_add->{$key} eq 'HASH') {
            _convert_section_to_hash(\$master->{$key}) if (ref $master->{$key} eq 'ARRAY');
            if (!exists $master->{$key}) {
                $master->{$key} = {};
            } elsif (!ref $master->{$key}) {
                _breakup_stats(\$master->{$key});
            }
            # If master->{$key} wasn't converted to a hash, the _consolidate_hashes function will early-out
            _consolidate_hashes($master->{$key}, $to_add->{$key});
        } else { # source is scalar
            next if ref $master->{$key};
            $master->{$key} ||= 0;
            if(_isNumeric($to_add->{$key})) {
                $master->{$key} += $to_add->{$key};
#} else {
#                print STDERR "DEBUG: Stats: trying to combine a non-numeric value '$to_add->{$key}'\n";
            }
        }
    }
}

sub merge_sections {
    my ($self, $report) = @_;

    unless (ref $report) {
        my $tmp = $self->new();
        $tmp->parse_report($report);
        $report = $tmp;
    }

    my $oldest = pmx_time2time($self->header('Report-Time'));
    my $candidate = pmx_time2time($report->header('Report-Time'));
    if (defined $oldest) {
        $oldest = $candidate if defined $candidate && $candidate < $oldest;
    } else {
        $oldest = $candidate;
    }

    _consolidate_hashes($self->{'sections'}, $report->{'sections'});

    $self->{headers}{'Report-Time'} = time2pmx_time($oldest) if $oldest;

    return $oldest;
}

sub merge_files_in_dir {
    my ($self, $dir, $dryrun) = @_;

    opendir(my $fh, $dir) or die "Failed to opendir $dir: $!";
    my @files = grep { ! /^\./ && -f "$dir/$_" } readdir($fh);
    closedir $fh;

    my $oldest;

    foreach ( @files ) {
        die "File '$dir/$_' does not exist" unless -f "$dir/$_";
        my $candidate = $self->merge_file ( "$dir/$_", $dryrun );
        if (defined $oldest) {
            $oldest = $candidate if defined($candidate) && $candidate < $oldest;
        } else {
            $oldest = $candidate;
        }
    }

    $self->{headers}{'Report-Time'} = time2pmx_time($oldest) if $oldest;

    return $oldest;
}

sub merge_file {
    my ($self, $file, $dryrun) = @_;

    my $bytes;

    my $stats = stat($file);
    my $size = $stats->[7];

    if ($^O =~ /Win32/) {
        # On Windows, the attempt to open the file for reading will fail if the
        # milter has it open for writing.  That's ok - we'll pick this file up
        # next time round.

        # Win32: Rename, open, read, close, delete
        return if $file =~ m/\.tmp$/;
        my $fpin;
        my $tmpfile = "$file.tmp";

        if ($dryrun) {
            $tmpfile = $file;
        } else {
            rename $file, $tmpfile or return;
        }
        if (! open $fpin, "<$tmpfile") {
            unlink $tmpfile unless $dryrun;
            return;
        }
        binmode($fpin);
        sysread($fpin, $bytes, $size);
        close $fpin;

        unlink $tmpfile unless $dryrun;

    } else {
        # On POSIX systems, we are allowed to open a file even if another
        # process has it opened.  We therefore use mandatory locking here and
        # in the engine stats writing functions to ensure that we are not going
        # to be reading the file while it is being simultaneously updated.

        # Linux: open, flock, read, unlink, close
        return if (not open my $fpin, "<$file");

        use Fcntl ':flock';
        # Wait for the file to be unlocked...
        flock $fpin, LOCK_EX;

        binmode($fpin);
        sysread($fpin, $bytes, $size);

        unlink $file unless $dryrun;
        close $fpin;
    }
    my $contents = eval { Storable::thaw($bytes) };
    return if $@;

    return $self->merge_sections($contents);
}

1;
