package Sophos::SXL;

use strict;
use Sophos::Debug qw(_debug DEBUG);

# include ActiveState::SpamEngine to gain access to '_get1()' - however we
# can't import the _get1() function into the local namespace, which is why when
# we actually require it, we fully qualify it.
use ActiveState::SpamEngine;
use ActiveState::Message;

use Sophos::SXL::Limit;

use ActiveState::DNS::RequestManager;
use ActiveState::DNS::Request qw/:status/;
use ActiveState::DNS::Util qw/:types/;

use Digest::MD5 qw(md5_hex);

use constant ROUNDROBIN => 0;
use constant SINGLE => 1;

use constant STATS_QUERIES => 0;
use constant STATS_ELAPSED => 1;
use constant STATS_MAX => 2;
use constant STATS_TIMEOUTS => 3;
use constant STATS_PING_TIMEOUTS => 4;

our $VERSION = '0.000';

sub new {
    my ($this, $eng) = @_;
    $eng or die("Sophos::SXL->new(eng): eng is undefined!\n");

    my $self = bless {}, ref $this || $this;

    $self->{'asdns'} = undef;
    $self->{'queries_a'} = [];
    $self->{'queries_ns'} = [];
    $self->{'limit'} = undef;
    $self->{'count'} = 0;
    $self->{sxl_requests} = {};
    $self->{'statistics'} = $eng->statistics();
    $self->{'state'} = ROUNDROBIN;
    $self->{'timeout_handicap'} = 1; # Default handicap for timeouts is 1 second

    $self->{'sleep_time'} = 300;
    $self->{'min_server_responses'} = 2;
    $self->{'max_server_iterations'} = 2;
    $self->{'always_round_robin'} = 0;

    if (my $server = $ENV{'SXLSERVER'}) {
        $self->{servers} = [split(/:/, $server)];
    } else {
        $self->{servers} = ['sophosxl.com'];
    }

    return $self;
}

sub set_dns_manager {
    my ($self, $manager) = @_;
    $self->{'asdns'} = $manager;
}

sub fisher_yates_shuffle {
    my $array = shift;
    my $i;
    for ($i = @$array; --$i; ) {
        my $j = int rand ($i+1);
        @$array[$i,$j] = @$array[$j,$i];
    }
}

sub compile {
    my ($self, $engine) = @_;

    $self->{'uri_dns_limit_db'} = eval {$engine->nested_db("uri_dns.limits")};
    # uri_dns_limit_db may not exist if no limits were enforced in the DFT
    if ($@) {
        $self->{'uri_dns_limit_db'} = undef;
        $engine->log('WARNING', "no uri_dns limit database found");
    }

    $self->{'db'} = eval { $engine->nested_db('sxl.match')};
    $self->{'sxl_limit_db'} = eval {$engine->nested_db("sxl.limits")};
    # sxl_limit_db may not exist if no limits were enforced in the DFT
    if ($@) {
        $self->{'sxl_limit_db'} = undef;
        $engine->log('WARNING', "no sxl limit database found");
    }
    my $globals = eval {$engine->nested_db("sxl.globals")};

    my $enable_sxl = 0;
    {
        my ($env_enable, $attr_enable);
        $env_enable = $ENV{'SXL_ENABLE'};
        return 0 if defined $env_enable && $env_enable eq '0';
        if (defined $env_enable) {
            $enable_sxl = 1;
        } else {
            $engine->enumerate_attributes('plugin.sxl.enable', \&ActiveState::SpamEngine::_get1, \$attr_enable);
            if (defined $attr_enable) {
                return 0 if $attr_enable eq '0';
                $enable_sxl = 1;
            }
        }
    }

    if ($globals) {
        if (!$ENV{'SXLSERVER'}) {
            my $servers = $globals->get('servers');
            if (defined $servers && $servers) {
                $self->{'servers'} = [split /\0/, $servers];
            }
        }

        $self->{'limit'} = $globals->get('global_limit');
        $self->{'maxlength'} = $globals->get('maxlength');
        $self->{'hashed_uris'} = 0;
        $self->{'max_response_time'} = 0;

        if (my $handicap = $globals->get('timeout_handicap')) {
            # handicap was specified in milliseconds, internal units are seconds
            $self->{'timeout_handicap'} = $handicap / 1000;
        }
        if ($globals->get('disable')) {
            # SPOPS wants SXL lookups disabled.
            return 0 unless $enable_sxl;
        }
        my $sleep_time = $globals->get('sleep_time');
        if (defined $sleep_time) {
            # sleep time is measured in seconds
            $self->{'sleep_time'} = $sleep_time;
        }
        my $min_server_responses = $globals->get('min_server_responses');
        if (defined $min_server_responses) {
            # this represtents the percentage of servers that need to respond to stay out of sleep mode
            $self->{'min_server_responses'} = $min_server_responses;
        }
        my $max_server_iterations = $globals->get('max_server_iterations');
        if (defined $max_server_iterations) {
            # this is the maximum number of times that SXL can go through the
            # server list in SINGLE mode before it goes into sleep mode
            $self->{'max_server_iterations'} = $max_server_iterations;
        }
        if (my $always_round_robin = $globals->get('always_round_robin')) {
            $self->{'always_round_robin'} = $always_round_robin;
        }

        # Give customers the opportunity to send all uris as a hash.
        {
            my $enable;
            $engine->enumerate_attributes('plugin.sxl.hashed-uris', \&ActiveState::SpamEngine::_get1, \$enable);
            #print STDERR "DEBUG: SXL: engine=$engine, hashed-uris attribute = $enable\n";
            $enable = 1 if $ENV{'SXL_HASHURIS'};
            $self->{'hashed_uris'} = 1 if $enable;
        }

        # Give customers the ability to set a maximum response time for server selection
        {
            my $enable;
            $engine->enumerate_attributes('plugin.sxl.max-response-time', \&ActiveState::SpamEngine::_get1, \$enable);
            # max-response-time is specified in milliseconds, so convert to internal units of seconds.
            $self->{'max_response_time'} = $enable / 1000 if $enable;
            _debug("max-response-time attribute = %s", $self->{'max_response_time'}) if DEBUG;
        }
    }

    # randomize the server list
    fisher_yates_shuffle($self->{servers});

    push @{$self->{'original_servers'}}, @{$self->{'servers'}};

    return 1;
}

sub maxlength {
    my $self = shift;

    return $self->{'maxlength'};
}

sub hash_uris {
    my $self = shift;
    #print STDERR "DEBUG: hash_uris() = $self->{'hashed_uris'}\n";
    return $self->{'hashed_uris'};
}

# Any of the following are possible
#  - no limits in dft.  $self->{*_limit_db} are undefined
#  - limits were set in a DFT for the type - db returns a numeric value
#  - limits were set in a DFT, but not for this type - db returns undef
#  - limits were set in a DFT, and count for this type was set to 0 - db returns 0

# The return value for this function on the other hand is:
#  - numeric 0 or more - number of SXL lookups to make on this type
#  - undef - no limits for lookups on this type

sub limit {
    my ($self, $type) = @_;

    my @databases = qw(sxl_limit_db uri_dns_limit_db);
    for my $db (@databases) {
        my $value = $self->{$db}->get($type) if (defined $self->{$db} && $self->{$db}->exists($type));
        if (defined $value) {
            _debug("{%s}{%s} = %s", $db, $type, $value) if DEBUG;
            return $value;
        }
    }
    return undef if not defined($self->{sxl_limit_db});
    _debug("examined dbs @databases; can't find type='%s'", $type) if DEBUG;
    return 0; # If an ASDB limit isn't available to provide limits, then assume a limit of 0 per type.
}

sub reset_count {
    my ($self) = @_;

    # An early out may have left pending queries unsent.  Make sure that these are not accidentally sent.
    $self->{queries_a} = [];
    $self->{queries_ns} = [];
    $self->{sxl_requests} = {};

    # Reset the limit count.
    $self->{'count'} = 0;
}

# Prevent any outstanding queries from being sent.
sub clear {
    my ($self) = @_;
    #print STDERR "DEBUG: SXL::clear()\n";
    my $count = @{$self->{'queries_ns'}};
    _debug(qq[%s NS queries to clear\n], $count) if DEBUG;
    for my $query (@{$self->{'queries_ns'}}) {
        next if defined $query->[0];
        $query->[0] = '';
        $count--;
    }
    _debug(qq[%s NS queries still outstanding after clearing\n], $count) if DEBUG;
    # If we neutered all NS queries, then remove the array.
    $self->{queries_ns} = [] unless $count;

    $count = @{$self->{'queries_a'}};
    _debug(qq[%s A queries to clear\n], $count) if (DEBUG);
    for my $query (@{$self->{'queries_a'}}) {
        next if defined $query->[0];
        $query->[0] = '';
        $count--;
    }
    _debug(qq[%s A queries still outstanding after clearing\n], $count) if (DEBUG);
    # If we neutered all A queries, then remove the array.
    $self->{queries_a} = [] unless $count;
    $self->{sxl_requests} = {};
}

sub select_dns_server {
    my ($self) = @_;
    _debug("always_round_robin='%s' state='%s' server_index='%s'", 
        $self->{always_round_robin}, $self->{state}, $self->{server_index}) if DEBUG;

    if (!defined $self->{server_index} || ($self->{server_index} > $#{$self->{servers}})) {
        _debug("set server index to 0") if DEBUG;
        $self->{server_index} = 0;
    }

    if ($self->{always_round_robin} || $self->{state} == SINGLE) {
        _debug("always_round_robin or state == SINGLE") if DEBUG;
        _debug("select server '%d'", $self->{server_index}) if DEBUG;
        return ($self->{server_index}) if wantarray;
        return $self->{server_index};
    }
    elsif ($self->{state} == ROUNDROBIN) {
        _debug("state == ROUNDROBIN") if DEBUG;
        if (scalar(@{$self->{servers}}) != scalar(@{$self->{original_servers}})) {
            _debug("server list has changed(%d); clear the server list then re-add original_servers(%d)",
                scalar(@{$self->{servers}}), scalar(@{$self->{original_servers}})) if DEBUG;
            $self->{servers} = [];
            push @{$self->{servers}}, @{$self->{original_servers}};
        }

        return (0 .. scalar $#{$self->{servers}}) if wantarray;
        return $self->{server_index}++; 

    }
}

# Sends any outstanding SXL queries.  May be called multiple times.  While
# available as an external API function, its use is optional because it will be
# called from collect_results.
sub dispatch {
    my ($self, $engine) = @_;
    my $asdns = $self->{'asdns'};
    return 0 unless $asdns && $asdns->available();

    for my $query (@{$self->{'queries_ns'}}) {
        next if (defined $query->[0]);
        my (undef, $q, $cb, $data, undef) = @$query;
        _debug(qq[query: ns: %s\n], $q) if (DEBUG);
        push @{$query->[0]}, $asdns->ns_lookup($q);
    }

    my $stats_reporter = $self->{statistics};
    my $count = 1;

    if ($self->{sleep_timestamp}) {
        my $time = time;
        my $time_since_last_message = $time - $self->{last_message_time};
        my $time_slept = $time - $self->{sleep_timestamp};
        $stats_reporter->add('engine-stats', 'sleep-time', $time_since_last_message);

        if ($time_slept < $self->{sleep_time}) {
            $stats_reporter->add('engine-stats', 'sleep-query-count', scalar(@{$self->{queries_a}}));

            $self->{last_message_time} = $time; 
            $self->{queries_a} = [];
            return 0;
        }
        else {
            $self->{sleep_timestamp} = 0;
            $self->{last_message_time} = 0;
            $engine->log('DEBUG', 'sxl: Ending SXL sleep mode');
        }
    }


    if (!$self->{always_round_robin} && $self->{state} == ROUNDROBIN) {
        if (@{$self->{queries_a}} && !defined $self->{'queries_a'}[0][0]) {
            my $thread_id = eval { ActiveState::Message::ThreadId() } || '';
            my $ping = md5_hex(time . '.' . rand(1000000) . ".$$.$thread_id");
            unshift @{$self->{'queries_a'}}, [undef, "$ping.ping", undef, undef, undef];
            $count = 0;
        }
    }

    for my $query (@{$self->{'queries_a'}}) {
        unless (defined $query->[0]) {
            my (undef, $q, $cb, $data, undef) = @$query;
            if (ref $q && ref $q eq 'ARRAY' && @$q >= 2) {
                # $q->[0] is the original query
                # $q->[1] is the DNS query
                $q = $q->[1];
            }

            my @indexes;

            if ($count == 0) {
                _debug("Ping queries are sent to all servers") if DEBUG;
                @indexes = $self->select_dns_server();
                $self->{sxl_requests}{ping} = scalar(@indexes);
            } else {
                _debug("original behaviour, poll one server at a time") if DEBUG;
                @indexes = (scalar $self->select_dns_server()); 
            }
            $count++;
            _debug("indexes = %d, count = %d", scalar @indexes, $count) if DEBUG;

            my @ids = ();

            for my $index (@indexes) {
                _debug("server index = %d", $index) if DEBUG;
                my $dnsserver = $self->{servers}[$index];
                _debug("querying $q.$dnsserver\n") if DEBUG;
                $stats_reporter->add('sxl-servers', $dnsserver, 'upload', 22 + length($q) + length($dnsserver));
                _debug(qq[query: a : %s.%s\n], $q, $dnsserver) if DEBUG;
                push @ids, $asdns->lookup("$q.$dnsserver"); 
                $engine->log('DEBUG', "sxl: lookup $q.$dnsserver");
            }
            $query->[0] = \@ids;
            $query->[4] = \@indexes;
        }
    }

    # Now add to the statistics
    for my $key (keys %{$self->{sxl_requests}}) {
        $stats_reporter->add('sxl-requests', $key, $self->{sxl_requests}{$key});
    }
     
    $self->{sxl_requests} = {};
    return 1;
}

sub get_ns {
    my ($self, $query, $cb, $data) = @_;

    _debug(qq[query='%s'], $query) if DEBUG;

    # If no request manager has been set, then NS lookups can't take place.
    # This may happen if local-checks-only is enabled, in which case this
    # method returns '0' to indicate that no NS lookup has been attempted.
    return 0 unless $self->{'asdns'};
    return 0 if $query =~ m/^\.|\.$|\.\.|^$/;

    # Do the lookup, and store the id we'll need to get the result
    my $asdns = $self->{'asdns'};
    return 0 unless $asdns->available();

    _debug("(query='%s'): pushing query\n", $query) if (DEBUG);
    push @{$self->{'queries_ns'}}, [
        undef,
        $query,
        $cb,
        $data,
        undef,
    ];

    return 1;
}

sub get {
    my ($self, $query, $cb, $data) = @_;

    _debug("query='%s'", ref $query eq 'ARRAY'? join(', ', @{$query}) : $query) if DEBUG;

    # If no request manager has been set, then SXL lookups can't take place.
    # This may happen if local-checks-only is enabled, in which case this
    # method returns '0' to indicate that no SXL lookup has been attempted.
    return 0 unless $self->{'asdns'};
    return 0 unless defined $query;
    return 0 if (ref $query ? $query->[1] : $query) =~ m/^\.|\.$|\.\.|^$/;

    $self->{'count'}++;
    if (defined $self->{'limit'} && ($self->{'count'} > $self->{'limit'}))
    {
        # The limit has been exceeded, so we will not perform an SXL lookup.
        # Indicate this to the caller (assuming they're interested) by
        # returning FALSE.
        _debug("The limit has been exceeded so we will not perform an SXL lookup.") if DEBUG;
        return 0;
    }

    # Do the lookup, and store the id we'll need to get the result
    my $asdns = $self->{'asdns'};
    return 0 unless $asdns->available();

    _debug("query='%s'", ref $query eq 'ARRAY'? join(', ', @{$query}) : $query) if DEBUG;

    push @{$self->{'queries_a'}}, [
        undef,
        $query,
        $cb,
        $data,
        undef,
        ];

    return 1;
}

# potential refactor, merge with read_result
sub read_result_ns {
    my ($self, $id, $engine) = @_;

    _debug("id=@$id\n") if DEBUG;
    my $asdns = $self->{'asdns'};
    my %result = ();

    my $request = $asdns->request($id->[0]);
    _debug("sxl: ns: requested domain: '%s'", $request->requested_domain()) if DEBUG;
    my $time = $request->elapsed_time();

    if (($request->status == REQUEST_OK)) {
        _debug("sxl: ns: REQUEST_OK") if DEBUG;
        my $value = $request->returned_domains;

        if ($value && ref $value eq 'ARRAY' && @$value) {
            $result{domains} = $value;
            $engine->log('DEBUG', sprintf("sxl: ns: [%s] from %s times: r=%.3fs",
                join(', ', @$value), $request->requested_domain, $time));
        }
        else {
            _debug("sxl: ns: error: received from asdns: empty dns answer section") if DEBUG;
        }

        if ($result{glue} = $request->returned_glue()) {
            $engine->log('DEBUG', sprintf("sxl: ns-glue: [%s] from %s times: r=%.3fs",
                                          join(",", keys(%{$result{glue}})), $request->requested_domain, $time));
        } elsif (DEBUG) { # could be else, but this hides from Devel::Cover.
            _debug("sxl: ns: info: received from asdns: no glue in additional section") if DEBUG;
        }
    }
    elsif ($request->status == REQUEST_ERROR) {
        _debug("sxl: ns: REQUEST_ERROR") if DEBUG;
        my $error = $request->error();
        $engine->log('DEBUG', sprintf('sxl: ns: error (%s) from %s times r=%.3fs',
            $error, $request->requested_domain, $time));

        if ($error =~ m/^RCODE_(\S+)/) {
            my $rcode_text = $1;
            _debug("sxl: ns: error: received from asdns: rcode: '%s'", $rcode_text) if DEBUG;
            $result{request_error_rcode} = $rcode_text;
        }
        else {
            _debug("sxl: ns: error: received from asdns: error string: '%s'", $error) if DEBUG;
            $engine->log('DEBUG', sprintf "received from asdns: error string: '%s' (non-rcode)", $error);
            $result{request_error_rcode} = 'ERROR_MISSING_RCODE'; # fake error only used in uri.pm
        }
    }
    else {
        _debug("sxl: ns: TIMEOUT") if DEBUG;
        $engine->log('DEBUG', sprintf('sxl: ns: [] from %s (timeout) times r=%.3fs', $request->requested_domain, $time));
        $result{timeout} = 1;
    }
    return \%result;
}

sub read_result {
    my ($self, $stats, $id, $server_index, $engine) = @_;

    _debug("id=@$id, server_index=@$server_index\n") if DEBUG;
    my $asdns = $self->{'asdns'};
    my $stats_reporter = $self->{statistics};
    my $index = 0;
    my %result = ();
    foreach (ref $id ? @$id : ($id)) {

        my $si = ref($server_index) ? $server_index->[$index++] : $server_index;

        my $request = $asdns->request($_);
        _debug("requested domain: '%s'", $request->requested_domain()) if DEBUG;
        my $time = $request->elapsed_time();
        $stats->[$si] = [0, 0, 0, 0, 0] unless defined $stats->[$si];
        my $sxlserver = $self->{servers}[$si];

        if (exists($request->{'reply'})) {
            $stats_reporter->add('sxl-servers', $sxlserver, 'download', length($request->{'reply'}));
        }

        my $ping_query = ($request->requested_domain() =~ m/[0-9a-f]+\.ping\.$sxlserver/);

        # if no error or nxdomain error (nxdomain isn't REALLY an error)
        if (($request->status == REQUEST_OK) || ($request->status == REQUEST_ERROR && $request->error eq 'RCODE_NXDOMAIN')) {
            # This *must* be a REQUEST_TYPE_LOOKUP - SXL will only make this type of query.
            _debug("REQUEST_TYPE_LOOKUP or NXDOMAIN") if DEBUG;
            my $value = $request->returned_ips;

            if ($value && ref $value eq 'ARRAY' && @$value) {
                $result{$si} = $value;

                $engine->log('DEBUG', do {
                    local $" = ", ";
                    sprintf("sxl: [@$value] from %s times: r=%.3fs", $request->requested_domain, $time);
                });
            } else {

                $engine->log('DEBUG', do {
                        local $" = ", ";
                        sprintf("sxl: NXDOMAIN from %s times: r=%.3fs", $request->requested_domain, $time);
                    });
            }

            if ($ping_query) {
                $stats_reporter->add('sxl-servers', $sxlserver, 'ping_latency', $time * 1000,);
                $stats_reporter->increment('sxl-servers', $sxlserver, 'pings');
            }
            else {
                $stats_reporter->add('sxl-servers', $sxlserver, 'latency', $time * 1000,);
                $stats_reporter->increment('sxl-servers', $sxlserver, 'count');
            }
        }
        elsif ($request->status == REQUEST_ERROR) {
            _debug("REQUEST_ERROR") if DEBUG;
            $stats_reporter->increment('sxl-servers', $sxlserver, 'errors');
            my $error = $request->error();
            $engine->log('DEBUG', sprintf('sxl: error (%s) from %s times r=%.3fs', $error, $request->requested_domain, $time));

            if ($error =~ m/^RCODE_(\S+)/) {
                my $rcode_error = $1;
                my $zone;
                if($request->requested_domain() =~ m/\.([^\.]+)\.$sxlserver$/) {
                    $zone = uc $1;
                } else {
                    $zone = '';
                }
                my $rule_hit = "__SXL_${zone}_ERROR_${rcode_error}";
                $result{$si} = [ $rule_hit ];
            }

            # For an error, we make no changes to the elapsed time and the
            # timeout count.

        } else {
            _debug("Request timeout") if DEBUG;
            $stats_reporter->increment('sxl-servers', $sxlserver, 'timeouts');
            $engine->log('DEBUG', sprintf('sxl: [] from %s (timeout) times r=%.3fs', $request->requested_domain, $time));

            my $zone;
            if($request->requested_domain() =~ m/\.([^\.]+)\.$sxlserver$/) {
                $zone = uc $1;
            } else {
                $zone = '';
            }
            my $rule_hit = "__SXL_${zone}_TIMEOUT";
            $result{$si} = [ $rule_hit ];
            _debug("timeout rule hit: '%s'", $rule_hit) if DEBUG;

            if ($ping_query) {
                $stats->[$si][STATS_PING_TIMEOUTS]++;
            }
            else {
                $stats->[$si][STATS_TIMEOUTS]++;
            }
            $time = $self->{timeout_handicap};
        }

        $stats->[$si][STATS_QUERIES]++;
        $stats->[$si][STATS_ELAPSED] += $time;
        $stats->[$si][STATS_MAX] = $time if $time > ($stats->[$si][STATS_MAX] || 0);
    }
    return \%result;
}



# called by the sxl plugin during the net.finish callback.
sub collect_results {
    my ($self, $engine) = @_;

    if (scalar(@{$self->{'queries_a'}}) == 0 && scalar(@{$self->{'queries_ns'}}) == 0) {
        _debug("short circuit for when there are no queries at all") if DEBUG;
        return;
    }

    my $asdns = $self->{'asdns'};
    # each stats entry contains [queries, accumulated time, timeouts]
    my @stats = ();
    my $stats_reporter = $self->{statistics};

    # If the SXL plugin has been disabled, then compile() was never called and
    # therefore there is no 'asdns' attribute.  In this case we do nothing.
    return unless $asdns && $asdns->available();

    return unless $self->dispatch($engine);

    $asdns->join();

    _debug("sxl: ns: looking at %s queries", scalar(@{$self->{'queries_ns'}})) if DEBUG;
    for my $query_ns (@{$self->{'queries_ns'}}) {
        my ($id, $q, $cb, $data) = @$query_ns;
        next unless defined $id;
        next if !ref($id) && $id eq ''; # ignore queries we never sent.
        my $ns_results = read_result_ns($self, $id, $engine);
        _debug("call callback: data='%s', q='%s', ns_results='%s'", $data, $q, $ns_results) if DEBUG;
        $cb->($data, $q, $ns_results);
        _debug("returned from callback") if DEBUG;
    }
    $self->{queries_ns} = [];

    if (scalar(@{$self->{'queries_a'}}) == 0) {
        _debug("short circuit for when there are no A queries") if DEBUG;
        return;
    }
    for my $query (@{$self->{'queries_a'}}) {
        my ($id, $q, $cb, $data, $server_index) = @$query;
        next unless defined $id;
        next if !ref($id) && $id eq ''; # ignore queries we never sent.
        my $values = read_result($self, \@stats, $id, $server_index, $engine);

        if (ref $q && ref $q eq 'ARRAY' && @$q >= 1) {
            # $q->[0] is the original query, and that's what we want to show in the callback
            $q = $q->[0];
        }

        if (defined $cb) {
            my %seen = ();
            while (my ($si, $value) = each %$values) {
                if ($value) {
                    for my $val (@$value) {
                        my $rule;
                        if($val =~ m/^__SXL_.*_(:?TIMEOUT|ERROR_[^.\s]+)$/) {
                            $rule = $val;
                            $q .= '.' . $self->{servers}[$si];
                            _debug("Timeout or error: '%s', '%s'", $rule, $q) if DEBUG;
                        } else {
                            _debug("Rule hit is: '%s'", $val) if DEBUG;
                            $rule = $self->{db}->get($val) if exists $self->{db};
                        }

                        if ($rule) {
                            # The same rule could be returned by multiple servers.
                            if ($seen{$rule}) {
                                _debug("Skipping - this rule '%s' has already been seen", $rule) if DEBUG;
                                next;
                            }
                            $seen{$rule} = 1;

                            if (defined $data) {
                                _debug("call callback with data: data='%s', q='%s', rule='%s'", $data, $q, $rule) if DEBUG;
                                $cb->($data, $q, $rule);
                            } else {
                                _debug("call callback with NO data: q='%s', rule='%s'", $q, $rule) if DEBUG;
                                $cb->($q, $rule);
                            }
                        } else {
                            _debug("An invalid response does nothing - though we track it against the DNS server.") if DEBUG;
                            $stats_reporter->increment('unknown-responses', $self->{servers}[$si], $value->[0]);
                        }
                    }
                }
            }
        }
    }
    $self->{queries_a} = [];

    #
    # server selection
    #

    if ($self->{always_round_robin}) {
        if (++$self->{server_index} > $#{$self->{servers}}) {
            $self->{server_index} = 0;
        }
    }
    elsif ($self->{state} == SINGLE) {
        _debug("state is SINGLE") if DEBUG;
        my $i = $self->{server_index};
        _debug("server_index=$self->{server_index}, servers=%d", scalar @{$self->{servers}}) if DEBUG;
        _debug("max_response_time=$self->{'max_response_time'}, average=%f", ($stats[$i][STATS_ELAPSED] || 0) / ($stats[$i][STATS_QUERIES] || 1)) if DEBUG;
        _debug("stats_timeouts=%d", $stats[$i][STATS_TIMEOUTS]) if DEBUG;
        if (
                ($self->{'max_response_time'} and $stats[$i][STATS_QUERIES] > 0 and ($stats[$i][STATS_ELAPSED] / $stats[$i][STATS_QUERIES]) > $self->{'max_response_time'}) or
                (defined $stats[$i][STATS_TIMEOUTS]) and ($stats[$i][STATS_TIMEOUTS] > 0)
           ) {

            _debug("Server $i is too slow; advancing to the next one (%d)", $self->{server_index} + 1) if DEBUG;
            if (++$self->{server_index} > $#{$self->{servers}}) {
                _debug("ran out of servers; set server index to 0") if DEBUG;
                $self->{server_index} = 0;

                if (++$self->{server_iteration_count} >= $self->{max_server_iterations}) {
                    $self->{sleep_timestamp} = $self->{last_message_time} = time;
                    $self->{state} = ROUNDROBIN;
                    $self->{servers} = [];
                    $engine->log('DEBUG', 'sxl: Starting SXL sleep mode (too many server list iterations in single mode)');
                }
            }
        }
        elsif(DEBUG) {
            _debug("Server stats look OK, server $i will continue to be used");
        }
    }
    else {
        _debug("state is ROUNDROBIN") if DEBUG;
        # hash linking server name to its weighted response time
        my %weight;
        # hash linking server name to its average response time
        my %average;
        # hash linking server name to whether there was a timeout
        my %timeout;

        _debug("Extract the servers we wanted to use from the server list") if DEBUG;
        my @servers;

        $self->{server_iteration_count} = 0;

        @servers = @{$self->{servers}};
        @{$self->{servers}} = ();

        # Pre-seed weight with maximum timeout, so servers which aren't
        # queried go after those which do (though servers with timeouts will
        # always go at the very end)
        $weight{$_} = $self->{timeout_handicap} foreach @servers;

        for (my $i = 0; $i <= $#servers; $i++) {
            # in ROUNDROBIN mode, servers are queried in
            # order.  If we come across a server to which we didn't send a
            # query, no subsequent server will be sent one either, so we 
            # use 'last' instead of 'next'
            last unless defined $stats[$i] && ref $stats[$i] && $stats[$i][STATS_QUERIES];

            $average{$servers[$i]} = $stats[$i][STATS_ELAPSED] / $stats[$i][STATS_QUERIES];
            $weight{$servers[$i]} = $stats[$i][STATS_MAX];
            $weight{$servers[$i]} += $average{$servers[$i]} if $stats[$i][STATS_MAX] == $self->{timeout_handicap};
            $timeout{$servers[$i]} = $stats[$i][STATS_PING_TIMEOUTS];
            _debug("DEBUG: server index=$i, name=$servers[$i], weight=$weight{$servers[$i]}, average=$average{$servers[$i]}\n") if DEBUG;
        }

        # Sort based on max/average times.  Note that this isn't the final order,
        # because we later separately strip out servers with timeouts from
        # those without.
        @servers = sort {$weight{$a} <=> $weight{$b}} @servers;

        _debug("Separate 'good' from 'bad' servers") if DEBUG;
        my @goodservers;
        my @badservers;
        foreach my $server (@servers) {
            if (
                    ($self->{'max_response_time'} and exists($average{$server}) and $average{$server} > $self->{'max_response_time'}) or
                    ((defined $timeout{$server}) and ($timeout{$server} > 0))
               ) {
                _debug("Adding server to the bad list '%s'", $server);
                push @badservers, $server;
            } else {
                _debug("Adding server to the good list '%s'", $server);
                push @goodservers, $server;
            }
        }

        # Only servers that responded put in the server list
        push @{$self->{servers}}, @goodservers;

        if (scalar(@goodservers) >=  $self->{min_server_responses}) { 
            _debug("Enough servers responded that we can continue into SINGLE mode.") if DEBUG;
            $self->{state} = SINGLE;
        }
        else {
            _debug("Not enough servers responded, so we go into sleep mode.") if DEBUG;
            $self->{sleep_timestamp} = $self->{last_message_time} = time;
            $engine->log('DEBUG', 'sxl: Starting SXL sleep mode (not enough servers responded in round robin mode)');
        }

        _debug("server list sorted - " . join(',', @{$self->{servers}})) if DEBUG;

        $self->{server_index} = 0;
    }
}

sub statistics {
    my $self = shift;
    return $self->{statistics};
}

1;
