
package Sophos::SXL::Limit;

use constant (DNS_MAX_QUERY_SIZE => 300);

use Digest::MD5;
use Sophos::Debug qw(_debug DEBUG);

sub new {
    my $pkg = shift;
    $class = ref $pkg || $pkg;
    my ($sxl, $type) = @_;
    my $self = bless {
        type => $type,
        sxl => $sxl,
        count => 0,
        unique_queries_a => {},
        unique_queries_ns => {},
        }, $class;

    $self->{limit} = $sxl->limit($type);

    return $self;
}

sub maxlength {
    my $self = shift;
    unless (exists $self->{maxlength}) {
        $self->{maxlength} = $self->{sxl}->maxlength;
    }
    return $self->{maxlength};
}

sub hash_uris {
    my $self = shift;
    return $self->{'sxl'}->hash_uris;
}

sub select_limit {
    my ($sxl, $type) = @_;
    my $result = Sophos::SXL::Limit->new($sxl, $type);
    return $result;
}

sub reset_count {
    my ($self) = @_;

    $self->{'count'} = 0;
    $self->{'unique_queries_a'} = {};
    $self->{'unique_queries_ns'} = {};
}

sub toAscii {
    my ($label) = @_;
    # Not nameprep and punycode - because they're silly!
    # Ok, so they're not silly, but they are for application on unicode text,
    # and our input is a byte sequence and not unicode.
    # 
    # All non-alphanum characters are output as '-NN', where NN is the hex encoding of the byte.
    # As an exception, the first non-'x' non-alphanum character is output as 'x-NN'.
    # For example:
    #    abc-def  => abc-2ddef
    #    abc-     => abc-2d
    #    -abc     => x-2dabc
    #    xyz      => xyz
    #    xx-abc   => xxx-2dabc
    #    1234     => x-31234

    my ($xxxx, $remaining) = ($label =~ m/^(x*)(.*)$/s);

    if ($xxxx) {
        # An existing x means no special case for first character
        $xxxx .= 'x' if $remaining =~ m/^[^a-zA-Z0-9]/s;
    } else {
        if ($remaining =~ m/^([^a-zA-Z0-9])(.*)$/s) {
            # Encode non-alpha first character, and preceed with an x.
            $remaining = $2;
            $xxxx .= sprintf("x-%02x", ord($1));
        }
    }

    $remaining =~ s/([^a-zA-Z0-9])/sprintf("-%02x", ord($1))/gse;

    return $xxxx . $remaining;
}

sub encode_domain {
    my $domain = shift;
    my $maxlength = shift || DNS_MAX_QUERY_SIZE;

    # without the '-1', we would discard empty trailing elements.  We want to
    # see them so that we can return the query as a hash.
    my @labels = split(/\./, $domain, -1);

    for my $label (@labels) {
        $label = toAscii($label);
        
        return Digest::MD5::md5_hex($domain) if length($label) == 0 || length($label) > 63;
    }

    my $result = join('.', @labels);

    return Digest::MD5::md5_hex($domain) if length($result) > $maxlength;
    return $result;
}

# Common function which concentrates on whether the query should take place,
# and incrementing statistics.
sub _common_get {
    my ($self, $q, $query, $cb, $data, $type) = @_;

    return $self->{"unique_queries_".$type}{$query} if defined $self->{"unique_queries_".$type}{$query};

    my $result;

    $self->{'count'}++;

    if (defined $self->{'limit'} && ($self->{'count'} > $self->{'limit'}))
    {
        # The limit has been exceeded, so we will not perform an SXL lookup.
        # Indicate this to the caller (assuming they're interested) by
        # returning FALSE.
        _debug("limit reached: type=%s val=%s", $self->{type}, $self->{limit}) if (DEBUG);
        $result = 0;
    } else {
        if ($type eq "ns") {
            $result = $self->{sxl}->get_ns($q, $cb, $data);
            # don't add this stat to the feedback counters (not really sxl anyway)
        }
        else {
            $result = $self->{sxl}->get($q, $cb, $data);
            # add this stat to the feedback counters
            if ($result != 0) {
                $self->{sxl}{sxl_requests}{$self->{'type'}} ||= 0;
                $self->{sxl}{sxl_requests}{$self->{'type'}}++;
            }
        }

        $self->{"unique_queries_".$type}{$query} = $result;
    }
    return $result;
}

sub get_uri_ns {
    my ($self, $query, $cb, $data) = @_;
    _debug("query='%s'", ($query || 'undef')) if DEBUG;

    return 0 if (!defined($query) || ($query eq ''));

    return $self->_common_get($query, $query, $cb, $data, "ns");
}

sub get_uri {
    my ($self, $query, $cb, $data) = @_;
    _debug("query='%s'", ($query || 'undef')) if DEBUG;

    return $self->get($query, $cb, $data) unless $self->hash_uris;

    my $encoded = Digest::MD5::md5_hex($query);
    my $q = ["$query.$self->{'type'}", "$encoded.$self->{'type'}"];
    return $self->_common_get($q, $query, $cb, $data, "a");
}

sub get {
    my ($self, $query, $cb, $data) = @_;
    #print STDERR "DEBUG: Sophos::SXL::Limit->get(query=$query)\n";
    return 0 if (!defined($query) || ($query eq ''));

    my $q = "$query.$self->{'type'}";
    my $encoded = encode_domain($query, $self->maxlength);

    if ($encoded ne $query) {
        $q = [$q, "$encoded.$self->{'type'}"];
    }
    return $self->_common_get($q, $query, $cb, $data, "a");
}
1;

