package Sophos::Statistics;

our $VERSION = '0.000';

use warnings;
use strict;

use ActiveState::Message;
use Sophos::Feedback::Report;
use File::Spec;
use File::Path;
use Storable;

use constant (STATS_DIR_SIZE_LIMIT => 1048576);  # 1MB

sub new {
    my ($this, $stats_dir, $stats_dir_limit) = @_;
    my $self = bless {}, ref $this || $this;

    # No stats dir passed in is specifically means statistics
    #   are disabled.
    return $self unless $stats_dir;

    # Always create the stats directory
    eval {
        mkpath($stats_dir) unless -d $stats_dir;
    };

    # The directory must now exist.
    return $self unless -d $stats_dir;

    $self->{dir} = $stats_dir;
    # Hardwired default limit is 100KB
    $self->{dir_limit} = $stats_dir_limit || STATS_DIR_SIZE_LIMIT;
    $self->{cur_time} = time;

    #reduce the size of stats dir if required
    if($self->reduce_dir_size()) { #error
        $self->{dir} = '';
    }

    $self->{report} = Sophos::Feedback::Report->new();
    $self->{accumulated} = Sophos::Feedback::Report->new();

    return $self;
}

sub increment {
    my $self = shift;
    return unless $self->{report};
    return unless @_;
    unshift @_, 'engine-stats' unless @_ > 1;
    $self->{accumulated}->increment(@_);
    return $self->{report}->increment(@_);
}

#category parameter is optional
sub add {
    my $self = shift;
    return unless $self->{report};
    return unless @_ > 1;
    unshift @_, 'engine-stats' unless @_ > 2;
    $self->{accumulated}->add(@_);
    return $self->{report}->add(@_);
}

sub query {
    my $self = shift;
    return unless $self->{report};
    return unless @_;
    unshift @_, 'engine-stats' unless @_ > 1;
    return $self->{report}->query(@_);
}

sub query_accumulated {
    my $self = shift;
    return unless $self->{accumulated};
    return unless @_;
    unshift @_, 'engine-stats' unless @_ > 1;
    return $self->{accumulated}->query(@_);
}

sub write {
    my ($self, $file, $data) = @_;

    use Fcntl qw(:flock);

    open (my $fh, ">$file") or return;
    if ($^O !~ /Win32/ && !flock($fh, LOCK_EX | LOCK_NB)) {
        close $fh;
        return;
    }
    binmode($fh);
    syswrite $fh, $data;
    close $fh;
    return 1;
}

sub store_stats {
    my ($self) = @_;
    return 1 unless $self->{dir};

    # XXX Only necessary for Win32 and be zero for others
    my $thread_id = eval { ActiveState::Message::ThreadId() };
    $thread_id = 0 if $@;

    my $file = File::Spec->catfile($self->{dir}, "$$.$thread_id.$self->{cur_time}");

    $self->{accumulated} = $self->{report} if(not -e $file); 
    $self->{report} = Sophos::Feedback::Report->new();

    my $content = Storable::freeze($self->{accumulated});

    if(!$self->write($file, $content)) {
        my $newtime = time;
        $newtime = $newtime + 1 if $newtime == $self->{cur_time};
        $self->{cur_time} = $newtime;

        eval {
            mkpath($self->{dir}) unless -d $self->{dir};
        };

        $file = File::Spec->catfile($self->{dir}, "$$.$thread_id.$self->{cur_time}");
        if (!$self->write($file, $content)) {
            # An error after the second bite means that this is no simple locking issue.
            $self->{dir} = '';
            return 1;
        }
    }

    return 0;
}

sub reduce_dir_size {
    my ($self) = @_;

    return 1 unless $self->{dir};

    opendir(my $DIR, "$self->{dir}") or return 1;
    my @files = grep $_ =~ /^\d+\.\d+\.\d+$/, readdir($DIR);
    closedir($DIR) or return 1;


    @files = sort {
        my ($a1) = ($a =~ m/\.(\d+)$/);
        my ($b1) = ($b =~ m/\.(\d+)$/);
        $b1 <=> $a1; } @files;

    my $accum = 0;
    for my $existing (@files) {
        $accum += (stat("$self->{'dir'}/$existing"))[7];

        if ($accum > $self->{dir_limit}) {
            $self->remove_stat_file("$self->{'dir'}/$existing") and return 1;
        }
    }
    return 0;
}

sub remove_stat_file {
    my ($self, $file) = @_;
    unlink $file or return 1;
    return 0;
}

1;
