use strict;
use warnings;

=head1 NAME

Sophos::TLD - Reliable TLD recognition using Mozilla's database

=cut

package Sophos::TLD;

our $VERSION = "0.0.2";

#original defaults my @default_data = qw(
#original defaults ph
#original defaults mil.ph
#original defaults *.uk
#original defaults *.sch.uk
#original defaults !parliament.uk
#original defaults *.tr
#original defaults );

=head1 SYNOPSIS

    open my $fh, "<", "effective_tld_names.dat";
    my $tld = Sophos::TLD->new($fh);

    if ($tld->is_tld("co.uk")) {
        # ...
    }

    my $tld_len_in_chars = $tld->tld_length("news.bbc.co.uk");
    # Value is 5 (length of "co.uk")

    my $prefix_len_in_chars = $tld->prefix_length("news.bbc.co.uk");
    # Value is 9 (length of "news.bbc.")

    my ($prefix, $tld) = $tld->parts("news.bbc.co.uk");
    # $prefix is "news.bbc", $tld is "co.uk"

=head1 DESCRIPTION

This module reads rules specified in Mozilla's "effective TLD names" format and
can tell you whether or not a string is a valid TLD, can split a domain name
into its prefix and its TLD, etc.

At the time of writing the current "official" Mozilla rules file is available
from http://mxr.mozilla.org/mozilla-central/source/netwerk/dns/src/effective_tld_names.dat?raw=1

=head1 METHODS

=head2 new($fh)

Instantiates a new Sophos::TLD object. The method will read its rules from
the supplied filehandle. (To pass in the rules as a string, use something like
L<IO::String>.)

=cut

sub new {
    my $class = shift;
    my $self = bless {}, ref($class) || $class;
    $self->{tree} = {};
    my $fh = shift;
    if (!$fh) {
        my $default_data_ref = _return_tld_list_defaults();
        for my $l (@{ $default_data_ref }) {
             $self->_parse_line($l);
        }
    }
    else {
        while (defined (my $l = readline $fh)) {
            $self->_parse_line($l);
        }
    }
    return $self;
}

sub _parse_line {
    my ($self, $line) = @_;
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    if ($line !~ m{^(?://|$)}) {
        if ($line =~ s/^!//) {
            $self->_tree_insert($line, 0); # exception
        } else {
            $self->_tree_insert($line, 1);
        }
    }
}

sub _tree_insert {
    my ($self, $tld, $val) = @_;
    my $cur = $self->{tree};
    foreach my $label (reverse(split(/\./, lc $tld))) {
        if (my $l = $cur->{$label}) {
            $cur = $l;
        } else {
            $cur = $cur->{$label} = {};
        }
    }
    $cur->{" "} = $val;
}

=head2 is_tld($str)

Returns true if the supplied string is a TLD, false otherwise.

=cut

sub is_tld {
    my ($self, $str) = @_;
    return 0 if not defined $str;
    $str =~ s/\.$//;
    return 0 if $str =~ m/^\s*$/;
    $str = lc $str;
    my $cur = $self->{tree};
    foreach my $label (reverse(split(/\./, $str))) {
        my $next;
        if ($next = $cur->{$label}) {
            $cur = $next;
        }
        elsif ($next = $cur->{"*"}) {
            return 0 if !_valid_label($label);
            $cur = $next;
        }
        else {
            return 0;
        }
    }
    return $cur->{" "} ? 1 : 0;
}

=head2 tld_length($name)

Returns the length in characters of the TLD part of the supplied domain name.
The TLD length of both "bob.co.uk" and "bob.co.uk." is 5 (i.e., the length of
"co.uk").

=cut

sub tld_length {
    my ($self, $str) = @_;
    return 0 if not defined $str;
    $str =~ s/\.$//;
    return 0 if $str =~ m/\.$/;
    my @labels = split(/\./, lc($str));

    for (my $i = 0; $i <= $#labels; ++$i) {
        my $suffix = join ".", @labels[$i..$#labels];
        return length $suffix if $self->is_tld($suffix);
    }
    return 0; # even last suffix (shortest) was not a valid tld, so tld length is 0
}

=head2 prefix_length($name)

Returns the length in characters of the non-TLD part of the supplied domain
name. The following is true:

    prefix_length($name) + tld_length($name) = length($name)

For example, the prefix length of "bob.co.uk" is 4 (i.e., the length of
"bob.").

=cut

sub prefix_length {
    my ($self, $str) = @_;
    return 0 if not defined $str;
    $str =~ s/\.$//;
    return length($str) - $self->tld_length($str);
}

=head2 parts($name)

Returns a two-element array, the first element being the prefix of the supplied
domain name and the second element being the TLD.

For example, the prefix of "search.cpan.org" is "search.cpan.". The TLD is, of
course, "org".

=cut

sub parts {
    my ($self, $str) = @_;

    $str =~ s/\.$//;
    my $prefix_len = $self->prefix_length($str);
    my $prefix = substr $str, 0, $prefix_len;
    my $tld = substr $str, $prefix_len;
    return ($prefix, $tld);
}

=head2 single_label_plus_tld($uri)

Returns the domain name of a URI.  E.g. sophos.com for www.sophos.com.  

Returns undef if the URI doesn't end with a domain name.

=cut

sub single_label_plus_tld {
    my ($self, $uri) = @_;

    (my $prefix, my $tld) = $self->parts($uri);

    return undef if !$tld;
    
    if ($prefix =~ m/\.?([^.]+\.)$/) {
        return $1.$tld;
    }

    return undef;
}

=head1 AUTHOR

Ross Thomas E<lt>ross.thomas@sophos.comE<gt>

=cut

# local helper functions:

sub _valid_label {
    my $label = shift;
    return 0 if $label !~ m/^[a-z]/;
    return 0 if $label !~ m/[a-z0-9]$/;
    return 0 if $label !~ m/^[a-z0-9\-]*$/;
    return 0 if length($label) > 63;
    return 1;
}

sub _return_tld_list_defaults {
    my @default_data = qw(
    a.sad   
    ab.ca   
    ac      
    ac.ac   
    ac.at   
    ac.be   
    ac.cn   
    ac.cr   
    ac.cy   
    ac.fj   
    ac.id   
    ac.il   
    ac.im   
    ac.in   
    ac.je   
    ac.jp   
    ac.kr   
    ac.nz   
    ac.pa   
    ac.th   
    ac.ug   
    ac.uk   
    ac.yu   
    ac.za   
    ad      
    ad.jp   
    adm.br  
    adv.br  
    ae      
    aero    
    af
    ag      
    ah.cn   
    ai      
    al      
    alt.za  
    am      
    am.br   
    an      
    ao      
    aq      
    ar      
    arpa    
    arq.br  
    art.br
    arts.ro 
    as      
    asdsad  
    asia    
    asn.au  
    asso.fr 
    asso.mc 
    at      
    au      
    aw      
    ax      
    az      
    ba      
    bb      
    bbs.tr  
    bc.ca   
    bd      
    be      
    bf      
    bg      
    bh      
    bi      
    bio.br  
    biz     
    bj      
    bj.cn   
    bm      
    bn      
    bo      
    br      
    bs      
    bt      
    bv      
    bw      
    by      
    bz      
    ca      
    cat     
    cc      
    cd      
    cf      
    cg      
    ch      
    ci      
    ck      
    cl      
    club.tw 
    cm      
    cn      
    cng.br  
    cnt.br  
    co      
    co.ac   
    co.at   
    co.cc   
    co.ck   
    co.cr   
    co.hu   
    co.id   
    co.il   
    co.im   
    co.in   
    co.je   
    co.jp   
    co.kr   
    co.nr   
    co.nz   
    co.sr   
    co.sy   
    co.th   
    co.to   
    co.tt   
    co.tz   
    co.ug   
    co.uk   
    co.uz   
    co.ve   
    co.vi   
    co.yu   
    co.za   
    co.zw   
    com     
    com.ag  
    com.al  
    com.ar  
    com.au  
    com.az  
    com.bb  
    com.br  
    com.bs  
    com.cn  
    com.co  
    com.cu  
    com.cy  
    com.do  
    com.ec  
    com.eg  
    com.es  
    com.fj  
    com.fr  
    com.ge  
    com.gt  
    com.gu  
    com.gy  
    com.hk  
    com.hr  
    com.jo  
    com.kh  
    com.kw  
    com.kz  
    com.la  
    com.lb  
    com.lv  
    com.ly  
    com.mm  
    com.mo  
    com.mt  
    com.mx  
    com.my  
    com.na  
    com.nc  
    com.nf  
    com.ng  
    com.ni  
    com.np  
    com.pa  
    com.pe  
    com.ph  
    com.pl  
    com.pn  
    com.pr  
    com.py  
    com.ro  
    com.ru  
    com.sa  
    com.sg  
    com.sy  
    com.tn  
    com.tr  
    com.tw  
    com.ua  
    com.uy  
    com.ve  
    com.vn  
    coop    
    cq.cn   
    cr      
    cu      
    cv      
    cx      
    cy      
    cz      
    de      
    dj      
    dk      
    dm      
    do      
    dsa     
    dz      
    ec      
    ecn.br  
    edu     
    edu.au  
    edu.bo  
    edu.cn  
    edu.hk  
    edu.mm  
    edu.mx  
    edu.pt  
    edu.tr  
    edu.za  
    ee      
    eg      
    eh      
    eng.br  
    er      
    ernet.in        
    es      
    esp.br  
    et      
    etc.br  
    eti.br  
    eu      
    fi      
    fi.cr   
    fin.ec  
    firm.ro 
    fj      
    fk      
    fm      
    fm.br   
    fo      
    fot.br  
    fr      
    fst.br  
    ga      
    game.tw 
    gb      
    gb.com  
    gb.net  
    gc.ca   
    gd      
    gd.cn   
    ge      
    gen.nz  
    gf      
    gg      
    gh      
    gi      
    gl      
    gm      
    gn      
    go.jp   
    go.kr   
    go.th   
    gov     
    gov.au  
    gov.br  
    gov.cn  
    gov.ec  
    gov.hk  
    gov.il  
    gov.in  
    gov.me  
    gov.mm  
    gov.mx  
    gov.sg  
    gov.tr  
    gov.uk  
    gov.za  
    gp      
    gq      
    gr      
    gs      
    gs.cn   
    gt      
    gu      
    gv.ac   
    gv.at   
    gw      
    gx.cn   
    gy      
    gz.cn   
    hb.cn   
    he.cn   
    hi.cn   
    hk      
    hk.cn   
    hl.cn   
    hm      
    hn      
    hn.cn   
    hr      
    ht      
    hu      
    id      
    id.lv   
    idv.tw  
    ie      
    il      
    im      
    in      
    ind.br  
    inf.br  
    info    
    info.hu 
    info.ro 
    info.tn 
    info.ve 
    int     
    io      
    iq      
    ir      
    is      
    it      
    it.ao   
    iz.hr   
    je      
    jl.cn   
    jm      
    jo      
    jobs    
    jor.br  
    jp      
    js.cn   
    ke      
    kg      
    kh      
    ki      
    km      
    kn      
    kp      
    kr      
    kw      
    ky      
    kz      
    la      
    lb      
    lc      
    lel.br  
    li      
    lk      
    ln.cn   
    lr      
    ls      
    lt      
    ltd.uk  
    lu      
    lv      
    ly      
    ma      
    mb.ca   
    mc      
    md      
    me      
    me.uk   
    med.br  
    med.ec  
    mg      
    mh      
    mi.th   
    mil     
    mil.br  
    mil.ec  
    mil.tr  
    mil.za  
    mk      
    ml      
    mm      
    mn      
    mo      
    mo.cn   
    mobi    
    mod.uk  
    mp      
    mq      
    mr      
    ms      
    mt      
    mu      
    muni.il 
    museum  
    mv      
    mw      
    mx      
    my      
    mz      
    na      
    name    
    nb.ca   
    nc      
    ne      
    ne.jp   
    ne.kr   
    net     
    net.ar  
    net.au  
    net.az  
    net.bb  
    net.br  
    net.bs  
    net.cn  
    net.co  
    net.cu  
    net.cy  
    net.do  
    net.ec  
    net.eg  
    net.ge  
    net.gr  
    net.hk  
    net.id  
    net.il  
    net.im  
    net.in  
    net.je  
    net.jo  
    net.kh  
    net.la  
    net.lb  
    net.lv  
    net.mm  
    net.mt  
    net.mx  
    net.my  
    net.np  
    net.nz  
    net.pa  
    net.pe  
    net.ph  
    net.pl  
    net.py  
    net.ru  
    net.sg  
    net.sy  
    net.th  
    net.tn  
    net.tr  
    net.tw  
    net.ua  
    net.uk  
    net.uy  
    net.ve  
    net.vi  
    net.yu  
    net.za  
    nf      
    nf.ca   
    ng      
    ngo.za  
    nhs.uk  
    ni      
    nic.uk  
    nl      
    nm.cn   
    nm.kr   
    no      
    nom.br  
    nom.co  
    nom.ro  
    nom.ve  
    nom.za  
    np      
    nr      
    ns.ca   
    nt.ca   
    nt.ro   
    ntr.br  
    nu      
    nx.cn   
    nz      
    odo.br  
    off.ai  
    om      
    on.ca   
    or.ac   
    or.at   
    or.cr   
    or.id   
    or.jp   
    or.kr   
    or.th   
    or.ug   
    org     
    org.ar  
    org.au  
    org.az  
    org.bb  
    org.bd  
    org.br  
    org.bs  
    org.cn  
    org.co  
    org.cu  
    org.cy  
    org.do  
    org.ec  
    org.eg  
    org.fj  
    org.ge  
    org.gu  
    org.hk  
    org.hu  
    org.il  
    org.im  
    org.je  
    org.jo  
    org.kh  
    org.la  
    org.lb  
    org.lc  
    org.lv  
    org.ly  
    org.mm  
    org.mo  
    org.mt  
    org.mx  
    org.my  
    org.na  
    org.nc  
    org.nz  
    org.pa  
    org.pe  
    org.ph  
    org.pl  
    org.py  
    org.ro  
    org.ru  
    org.sg  
    org.sy  
    org.tn  
    org.tr  
    org.tw  
    org.ua  
    org.uk  
    org.uy  
    org.ve  
    org.vi  
    org.yu  
    org.za  
    pa      
    pe      
    pe.ca   
    per.sg  
    pf      
    pg      
    ph      
    pk      
    pl      
    plc.uk  
    pm      
    pn      
    police.uk       
    ppg.br  
    pr      
    presse.fr       
    priv.hu 
    pro     
    pro.br  
    ps      
    psc.br  
    psi.br  
    pt      
    pvt.ge  
    pw      
    py      
    qa      
    qc.ca   
    qh.cn   
    re      
    re.kr   
    rec.br  
    rec.ro  
    res.in  
    ro      
    rs      
    ru      
    rw      
    sa      
    sa.cr   
    sb      
    sc      
    sc.cn   
    sch.uk  
    school.za       
    sd      
    se      
    sg      
    sh      
    sh.cn   
    si      
    sj      
    sk      
    sk.ca   
    sl      
    slg.br  
    sm      
    sn      
    sn.cn   
    so      
    sr      
    st      
    store.ro        
    su      
    sv      
    sy      
    sz      
    tc      
    td      
    tec.ve  
    tel     
    tf      
    tg      
    th      
    tj      
    tj.cn   
    tk      
    tl      
    tm      
    tm.fr   
    tm.mc   
    tm.ro   
    tm.za   
    tmp.br  
    tn      
    to      
    tp      
    tr      
    travel  
    tt      
    tur.br  
    tv      
    tv.br   
    tw      
    tw.cn   
    tx.us   
    txt     
    tz      
    ua      
    ug      
    uk      
    uk.co   
    uk.com  
    uk.net  
    um      
    us      
    uy      
    uz      
    va      
    vc      
    ve      
    vet.br  
    vg      
    vi      
    vn      
    vu      
    web.do  
    web.id  
    web.ve  
    web.za  
    wf      
    ws      
    www.ro  
    xj.cn   
    xn--0zwm56d     
    xn--11b5bs3a9aj6g       
    xn--80akhbyknj4f        
    xn--9t4b11yi5a  
    xn--deba0ad     
    xn--g6w251d     
    xn--hgbk6aj7f53bba      
    xn--hlcj6aya9esc7a      
    xn--jxalpdlp    
    xn--kgbechtv    
    xn--zckzah      
    xz.cn   
    ye      
    yk.ca   
    yn.cn   
    yt      
    yu      
    za      
    zj.cn   
    zlg.br  
    zm      
    zw      
    );
    return \@default_data;
}

1;

# vim: set sts=4 sw=4 et ft=perl:

